//----------------------------------------------------------------------------
// Copyright (c) 2015-2016 Fabrice HARROUET (ENIB)
//
// This file is part of GpuElint.
//   http://www.enib.fr/~harrouet/gpuelint.html
//
// GpuElint is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GpuElint is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with GpuElint.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "gpuElint.h"

#include <hdf5.h>

typedef struct _PulseRecord
{
real64_t toa;
real32_t di;
real32_t fn;
real32_t niv;
real32_t pri;
real32_t azimut;
real32_t site;
int64_t rang;
} PulseRecord;

void
loadHdf5PulseFile(const char *fileName,
                  int *out_pulseCount,
                  PulseRecord **out_pulses,
                  int32_t **out_inter,
                  int32_t **out_expInter)
{
printf("~~~~~~~~ loading %s HDF5 file ~~~~~~~~\n",
       fileName);
int pulseCount=-1;
herr_t err;
hid_t pulseFile=H5Fopen(fileName,H5F_ACC_RDONLY,H5P_DEFAULT);
ON_ERROR(pulseFile<0,"H5Fopen() failure",{ exit(1); });
const char *groupName="data";
hid_t group=H5Gopen(pulseFile,groupName,H5P_DEFAULT);
ON_ERROR(group<0,"H5Gopen() failure",{ exit(1); });
hsize_t objCount;
err=H5Gget_num_objs(group,&objCount);
ON_ERROR(err<0,"H5Gget_num_objs() failure",{ exit(1); });
for(hsize_t o=0;o<objCount;++o)
  {
  int objType=H5Gget_objtype_by_idx(group,o);
  ON_ERROR(objType<0,"H5Gget_objtype_by_idx() failure",{ exit(1); });
  if(objType!=H5G_DATASET) { continue; }
  char objName[0x100];
  ssize_t nameLen=H5Gget_objname_by_idx(group,o,objName,sizeof(objName));
  ON_ERROR(nameLen<0,"H5Gget_objname_by_idx() failure",{ exit(1); });
  int kind=-1;
  if(!strcmp(objName,"impulsions")) kind=0;
  else if(!strcmp(objName,"interception")) kind=1;
  else if(!strcmp(objName,"interception_expected")) kind=2;
  hid_t dataSet=H5Dopen2(group,objName,H5P_DEFAULT);
  ON_ERROR(dataSet<0,"H5Dopen2() failure",{ exit(1); });
  if(kind>=0)
    {
    hid_t space=H5Dget_space(dataSet);
    ON_ERROR(space<0,"H5Dget_space() failure",{ exit(1); });
    hsize_t dims[1];
    int ndims=H5Sget_simple_extent_dims(space,dims,NULL);
    ON_ERROR(ndims<0,"H5Sget_simple_extent_dims() failure",{ exit(1); });
    H5Sclose(space);
    printf("found [%s/%s] dataset with %d pulses\n",
           groupName,objName,(int)dims[0]);
    if(pulseCount==-1)
      {
      pulseCount=(int)dims[0];
      *out_pulseCount=pulseCount;
      }
    else if(pulseCount!=(int)dims[0])
      {
      printf("pulse count mismatch: %d != %d\n",
             pulseCount,(int)dims[0]);
      kind=-1;
      }
    }
  switch(kind)
    {
    case 0:
      {
      if(out_pulses)
        {
        PulseRecord *pulses=Memory_alloc(pulseCount,PulseRecord);
        hid_t memPulseType=H5Tcreate(H5T_COMPOUND,sizeof(PulseRecord));
        ON_ERROR(memPulseType<0,"H5Tcreate() failure",{ exit(1); });
        err=H5Tinsert(memPulseType,"toa",
                      HOFFSET(PulseRecord,toa),H5T_NATIVE_DOUBLE);
        ON_ERROR(err<0,"H5Tinsert() failure",{ exit(1); });
        err=H5Tinsert(memPulseType,"di",
                      HOFFSET(PulseRecord,di),H5T_NATIVE_FLOAT);
        ON_ERROR(err<0,"H5Tinsert() failure",{ exit(1); });
        err=H5Tinsert(memPulseType,"fn",
                      HOFFSET(PulseRecord,fn),H5T_NATIVE_FLOAT);
        ON_ERROR(err<0,"H5Tinsert() failure",{ exit(1); });
        err=H5Tinsert(memPulseType,"niv",
                      HOFFSET(PulseRecord,niv),H5T_NATIVE_FLOAT);
        ON_ERROR(err<0,"H5Tinsert() failure",{ exit(1); });
        err=H5Tinsert(memPulseType,"pri",
                      HOFFSET(PulseRecord,pri),H5T_NATIVE_FLOAT);
        ON_ERROR(err<0,"H5Tinsert() failure",{ exit(1); });
        err=H5Tinsert(memPulseType,"azimut",
                      HOFFSET(PulseRecord,azimut),H5T_NATIVE_FLOAT);
        ON_ERROR(err<0,"H5Tinsert() failure",{ exit(1); });
        err=H5Tinsert(memPulseType,"site",
                      HOFFSET(PulseRecord,site),H5T_NATIVE_FLOAT);
        ON_ERROR(err<0,"H5Tinsert() failure",{ exit(1); });
        err=H5Tinsert(memPulseType,"rang desc intra",
                      HOFFSET(PulseRecord,rang),H5T_NATIVE_INT64);
        ON_ERROR(err<0,"H5Tinsert() failure",{ exit(1); });
        err=H5Dread(dataSet,memPulseType,H5S_ALL,H5S_ALL,H5P_DEFAULT,pulses);
        ON_ERROR(err<0,"H5Dread() failure",{ exit(1); });
        H5Tclose(memPulseType);
        *out_pulses=pulses;
        }
      break;
      }
    case 1:
    case 2:
      {
      if(((kind==1)&&out_inter)||
         ((kind==2)&&out_expInter))
        {
        int32_t *inter=Memory_alloc(pulseCount,int32_t);
        hid_t memInterType=H5Tcreate(H5T_COMPOUND,sizeof(int));
        ON_ERROR(memInterType<0,"H5Tcreate() failure",{ exit(1); });
        err=H5Tinsert(memInterType,"group_id",0,H5T_NATIVE_INT);
        ON_ERROR(err<0,"H5Tinsert() failure",{ exit(1); });
        err=H5Dread(dataSet,memInterType,H5S_ALL,H5S_ALL,H5P_DEFAULT,inter);
        ON_ERROR(err<0,"H5Dread() failure",{ exit(1); });
        H5Tclose(memInterType);
        if(kind==1) *out_inter=inter;
        else *out_expInter=inter;
        }
      break;
      }
    }
  H5Dclose(dataSet);
  }
H5Gclose(group);
H5Fclose(pulseFile);
}

void
writeHdf5PulseFile(const char *fileName,
                   int pulseCount,
                   const PulseRecord *pulses,
                   const int *inter,
                   const int *expInter)
{
printf("~~~~~~~~ writing %d pulses to %s HDF5 file ~~~~~~~~\n",
       pulseCount,fileName);
herr_t err;
hid_t pulseFile=H5Fcreate(fileName,H5F_ACC_TRUNC,H5P_DEFAULT,H5P_DEFAULT);
ON_ERROR(pulseFile<0,"H5Fcreate() failure",{ exit(1); });
hsize_t dims[1]={ (hsize_t)pulseCount };
hid_t dataSpace=H5Screate_simple(1,dims,NULL);
ON_ERROR(dataSpace<0,"H5Screate_simple() failure",{ exit(1); });
hid_t memPulseType=H5Tcreate(H5T_COMPOUND,sizeof(PulseRecord));
ON_ERROR(memPulseType<0,"H5Tcreate() failure",{ exit(1); });
err=H5Tinsert(memPulseType,"toa",
              HOFFSET(PulseRecord,toa),H5T_NATIVE_DOUBLE);
ON_ERROR(err<0,"H5Tinsert() failure",{ exit(1); });
err=H5Tinsert(memPulseType,"di",
              HOFFSET(PulseRecord,di),H5T_NATIVE_FLOAT);
ON_ERROR(err<0,"H5Tinsert() failure",{ exit(1); });
err=H5Tinsert(memPulseType,"fn",
              HOFFSET(PulseRecord,fn),H5T_NATIVE_FLOAT);
ON_ERROR(err<0,"H5Tinsert() failure",{ exit(1); });
err=H5Tinsert(memPulseType,"niv",
              HOFFSET(PulseRecord,niv),H5T_NATIVE_FLOAT);
ON_ERROR(err<0,"H5Tinsert() failure",{ exit(1); });
err=H5Tinsert(memPulseType,"pri",
              HOFFSET(PulseRecord,pri),H5T_NATIVE_FLOAT);
ON_ERROR(err<0,"H5Tinsert() failure",{ exit(1); });
err=H5Tinsert(memPulseType,"azimut",
              HOFFSET(PulseRecord,azimut),H5T_NATIVE_FLOAT);
ON_ERROR(err<0,"H5Tinsert() failure",{ exit(1); });
err=H5Tinsert(memPulseType,"site",
              HOFFSET(PulseRecord,site),H5T_NATIVE_FLOAT);
ON_ERROR(err<0,"H5Tinsert() failure",{ exit(1); });
err=H5Tinsert(memPulseType,"rang desc intra",
              HOFFSET(PulseRecord,rang),H5T_NATIVE_INT64);
ON_ERROR(err<0,"H5Tinsert() failure",{ exit(1); });
hid_t filePulseType=H5Tcreate(H5T_COMPOUND,(1*8)+(6*4)+(1*8));
ON_ERROR(filePulseType<0,"H5Tcreate() failure",{ exit(1); });
err=H5Tinsert(filePulseType,"toa",
              (0*8)+(0*4)+(0*8),H5T_IEEE_F64LE);
ON_ERROR(err<0,"H5Tinsert() failure",{ exit(1); });
err=H5Tinsert(filePulseType,"di",
              (1*8)+(0*4)+(0*8),H5T_IEEE_F32LE);
ON_ERROR(err<0,"H5Tinsert() failure",{ exit(1); });
err=H5Tinsert(filePulseType,"fn",
              (1*8)+(1*4)+(0*8),H5T_IEEE_F32LE);
ON_ERROR(err<0,"H5Tinsert() failure",{ exit(1); });
err=H5Tinsert(filePulseType,"niv",
              (1*8)+(2*4)+(0*8),H5T_IEEE_F32LE);
ON_ERROR(err<0,"H5Tinsert() failure",{ exit(1); });
err=H5Tinsert(filePulseType,"pri",
              (1*8)+(3*4)+(0*8),H5T_IEEE_F32LE);
ON_ERROR(err<0,"H5Tinsert() failure",{ exit(1); });
err=H5Tinsert(filePulseType,"azimut",
              (1*8)+(4*4)+(0*8),H5T_IEEE_F32LE);
ON_ERROR(err<0,"H5Tinsert() failure",{ exit(1); });
err=H5Tinsert(filePulseType,"site",
              (1*8)+(5*4)+(0*8),H5T_IEEE_F32LE);
ON_ERROR(err<0,"H5Tinsert() failure",{ exit(1); });
err=H5Tinsert(filePulseType,"rang desc intra",
              (1*8)+(6*4)+(0*8),H5T_STD_I64LE);
ON_ERROR(err<0,"H5Tinsert() failure",{ exit(1); });
hid_t memInterType=H5Tcreate(H5T_COMPOUND,sizeof(int32_t));
ON_ERROR(memInterType<0,"H5Tcreate() failure",{ exit(1); });
err=H5Tinsert(memInterType,"group_id",0,H5T_NATIVE_INT);
ON_ERROR(err<0,"H5Tinsert() failure",{ exit(1); });
hid_t fileInterType=H5Tcreate(H5T_COMPOUND,4);
ON_ERROR(fileInterType<0,"H5Tcreate() failure",{ exit(1); });
err=H5Tinsert(fileInterType,"group_id",0,H5T_STD_I32LE);
ON_ERROR(err<0,"H5Tinsert() failure",{ exit(1); });
hid_t group=H5Gcreate2(pulseFile,"data",H5P_DEFAULT,H5P_DEFAULT,H5P_DEFAULT);
ON_ERROR(group<0,"H5Gcreate2() failure",{ exit(1); });
if(pulses)
  {
  hid_t dataSet=H5Dcreate2(group,"impulsions",
                           filePulseType,dataSpace,
                           H5P_DEFAULT,H5P_DEFAULT,H5P_DEFAULT);
  ON_ERROR(dataSet<0,"H5Dcreate2() failure",{ exit(1); });
  err=H5Dwrite(dataSet,memPulseType,
               H5S_ALL,H5S_ALL,H5P_DEFAULT,pulses);
  ON_ERROR(err<0,"H5Dwrite() failure",{ exit(1); });
  H5Dclose(dataSet);
  }
if(inter)
  {
  hid_t dataSet=H5Dcreate2(group,"interception",
                           fileInterType,dataSpace,
                           H5P_DEFAULT,H5P_DEFAULT,H5P_DEFAULT);
  ON_ERROR(dataSet<0,"H5Dcreate2() failure",{ exit(1); });
  err=H5Dwrite(dataSet,memInterType,
               H5S_ALL,H5S_ALL,H5P_DEFAULT,inter);
  ON_ERROR(err<0,"H5Dwrite() failure",{ exit(1); });
  H5Dclose(dataSet);
  }
if(expInter)
  {
  hid_t dataSet=H5Dcreate2(group,"interception_expected",
                           fileInterType,dataSpace,
                           H5P_DEFAULT,H5P_DEFAULT,H5P_DEFAULT);
  ON_ERROR(dataSet<0,"H5Dcreate2() failure",{ exit(1); });
  err=H5Dwrite(dataSet,memInterType,
               H5S_ALL,H5S_ALL,H5P_DEFAULT,expInter);
  ON_ERROR(err<0,"H5Dwrite() failure",{ exit(1); });
  H5Dclose(dataSet);
  }
H5Gclose(group);
H5Tclose(filePulseType);
H5Tclose(memPulseType);
H5Sclose(dataSpace);
H5Fclose(pulseFile);
}

void
loadRawPulseFile(const char *fileName,
                 int *out_pulseCount,
                 PulseRecord **out_pulses)
{
printf("~~~~~~~~ loading %s raw file ~~~~~~~~\n",
       fileName);
FILE *pulseFile=fopen(fileName,"rb");
if(!pulseFile)
  { printf("cannot open %s\n",fileName); exit(1); }
#define LOAD_RAW_FIELD(f)                      \
  do                                           \
    {                                          \
    if(fread(&(f),sizeof((f)),1,pulseFile)!=1) \
      { printf("fread failure\n"); exit(1); }  \
    } while(0)
int32_t count;
LOAD_RAW_FIELD(count);
int pulseCount=(int)count;
PulseRecord *pulses=Memory_alloc(pulseCount,PulseRecord);
for(int i=0;i<pulseCount;++i) { LOAD_RAW_FIELD(pulses[i].toa);    }
for(int i=0;i<pulseCount;++i) { LOAD_RAW_FIELD(pulses[i].di);     }
for(int i=0;i<pulseCount;++i) { LOAD_RAW_FIELD(pulses[i].fn);     }
for(int i=0;i<pulseCount;++i) { LOAD_RAW_FIELD(pulses[i].niv);    }
for(int i=0;i<pulseCount;++i) { LOAD_RAW_FIELD(pulses[i].azimut); }
fclose(pulseFile);
for(int i=0;i<pulseCount;++i)
  {
  pulses[i].pri=-1.0f;
  pulses[i].site=-1.0f;
  pulses[i].rang=-1;
  }
printf("%d pulses loaded\n",pulseCount);
*out_pulseCount=pulseCount;
*out_pulses=pulses;
}

void
writeRawPulseFile(const char *fileName,
                  int pulseCount,
                  const PulseRecord *pulses)
{
printf("~~~~~~~~ writing %d pulses to %s raw file ~~~~~~~~\n",
       pulseCount,fileName);
FILE *pulseFile=fopen(fileName,"wb");
if(!pulseFile)
  { printf("cannot open %s\n",fileName); exit(1); }
#define WRITE_RAW_FIELD(f)                      \
  do                                            \
    {                                           \
    if(fwrite(&(f),sizeof((f)),1,pulseFile)!=1) \
      { printf("fwrite failure\n"); exit(1); }  \
    } while(0)
int32_t count=(int32_t)pulseCount;
WRITE_RAW_FIELD(count);
for(int i=0;i<pulseCount;++i) { WRITE_RAW_FIELD(pulses[i].toa);    }
for(int i=0;i<pulseCount;++i) { WRITE_RAW_FIELD(pulses[i].di);     }
for(int i=0;i<pulseCount;++i) { WRITE_RAW_FIELD(pulses[i].fn);     }
for(int i=0;i<pulseCount;++i) { WRITE_RAW_FIELD(pulses[i].niv);    }
for(int i=0;i<pulseCount;++i) { WRITE_RAW_FIELD(pulses[i].azimut); }
fclose(pulseFile);
}

int
main(int argc,
     char **argv)
{
const char *inputFile=NULL;
const char *outputFile=NULL;
int repeat=1;
bool showUsage=false;
if(argc>1) inputFile=argv[1];
if(argc>2) outputFile=argv[2];
if((argc>3)&&((sscanf(argv[3]," %d ",&repeat)!=1)||(repeat<1)))
  { showUsage=true; }
if((argc>4)||!inputFile) { showUsage=true; }
if(showUsage)
  {
  fprintf(stderr,"usage: %s input_file [output_file] [repeat]\n",argv[0]);
  exit(1);
  }
const char *inputExt=strrchr(inputFile,'.');
const char *outputExt=NULL;
if(outputFile) { outputExt=strrchr(outputFile,'.'); }
int pulseCount=-1;
PulseRecord *pulses=NULL;
int *inter=NULL;
int *expInter=NULL;
if(inputExt&&!strcmp(inputExt,".h5"))
  { loadHdf5PulseFile(inputFile,&pulseCount,&pulses,&inter,&expInter); }
else
  { loadRawPulseFile(inputFile,&pulseCount,&pulses); }
PulseRecord minRecord,maxRecord;
memset(&minRecord,0,sizeof(minRecord));
memset(&maxRecord,0,sizeof(maxRecord));
for(int i=0;i<pulseCount;++i)
  {
  PulseRecord *r=&pulses[i];
#define TEST_FIELD(f) \
        if((!i)||(r->f<minRecord.f)) minRecord.f=r->f; \
        if((!i)||(r->f>maxRecord.f)) maxRecord.f=r->f;
  TEST_FIELD(toa)
  TEST_FIELD(di)
  TEST_FIELD(fn)
  TEST_FIELD(niv)
  TEST_FIELD(pri)
  TEST_FIELD(azimut)
  TEST_FIELD(site)
  TEST_FIELD(rang)
  }
printf("toa: min=%.12f max=%.12f\n",minRecord.toa,maxRecord.toa);
printf("di: min=%.12f max=%.12f\n",minRecord.di,maxRecord.di);
printf("fn: min=%.12f max=%.12f\n",minRecord.fn,maxRecord.fn);
printf("niv: min=%.12f max=%.12f\n",minRecord.niv,maxRecord.niv);
printf("pri: min=%.12f max=%.12f\n",minRecord.pri,maxRecord.pri);
printf("azimut: min=%.12f max=%.12f\n",minRecord.azimut,maxRecord.azimut);
printf("site: min=%.12f max=%.12f\n",minRecord.site,maxRecord.site);
printf("rang: min=%" F64 "d max=%" F64 "d\n",minRecord.rang,maxRecord.rang);
if(expInter)
  {
  int minExp=expInter[0],maxExp=expInter[0];
  for(int i=0;i<pulseCount;++i)
    {
    if(minExp>expInter[i]) minExp=expInter[i];
    if(maxExp<expInter[i]) maxExp=expInter[i];
    }
  int expRange=maxExp-minExp+1;
  int *expHisto=Memory_initAlloc(expRange,int);
  for(int i=0;i<pulseCount;++i)
    {
    ++expHisto[expInter[i]-minExp];
    }
  for(int i=0;i<expRange;++i)
    {
    int meanCount=0;
    real64_t lastToa=-1.0,meanDtoa=0.0,meanDtoa2=0.0;
    real64_t meanDi=0.0,meanFn=0.0,meanNiv=0.0,meanAzimut=0.0;
    real64_t meanDi2=0.0,meanFn2=0.0,meanNiv2=0.0,meanAzimut2=0.0;
    for(int pass=0;pass<2;++pass)
      {
      for(int j=0;j<pulseCount;++j)
        {
        if(expInter[j]==i+minExp)
          {
          real64_t toa=pulses[j].toa;
          real64_t di=pulses[j].di;
          real64_t fn=pulses[j].fn;
          real64_t niv=pulses[j].niv;
          real64_t azimut=pulses[j].azimut;
          real64_t dtoa=meanCount ? fabs(toa-lastToa) : 0.0;
          if(pass)
            {
            if(meanCount) { dtoa-=meanDtoa; meanDtoa2+=dtoa*dtoa; }
            di-=meanDi; meanDi2+=di*di;
            fn-=meanFn; meanFn2+=fn*fn;
            niv-=meanNiv; meanNiv2+=niv*niv;
            azimut-=meanAzimut; meanAzimut2+=azimut*azimut;
            }
          else
            {
            meanDtoa+=dtoa;
            meanDi+=di;
            meanFn+=fn;
            meanNiv+=niv;
            meanAzimut+=azimut;
            }
          lastToa=toa;
          ++meanCount;
          }
        }
      if(meanCount!=expHisto[i]) { printf("COUNT MISMATCH\n"); }
      if(pass)
        {
        meanDtoa2/=meanCount-1;
        meanDi2/=meanCount;
        meanFn2/=meanCount;
        meanNiv2/=meanCount;
        meanAzimut2/=meanCount;
        }
      else
        {
        meanDtoa/=meanCount-1;
        meanDi/=meanCount;
        meanFn/=meanCount;
        meanNiv/=meanCount;
        meanAzimut/=meanCount;
        meanCount=0;
        lastToa=-1.0;
        }
      }
    if(meanCount>1)
      {
      printf("expected=%3d --> %5d: "
             "dtoa=%g(%g)  di=%g(%g)  fn=%g(%g)  niv=%g(%g)  azimut=%g(%g)\n",
             i+minExp,expHisto[i],
             meanDtoa,sqrt(meanDtoa2),
             meanDi,sqrt(meanDi2),
             meanFn,sqrt(meanFn2),
             meanNiv,sqrt(meanNiv2),
             meanAzimut,sqrt(meanAzimut2));
      }
    }
  Memory_free(expHisto);
  }
#if 0 // filter some specific pulses
  {
  int newPulseCount=0;
  for(int i=0;i<pulseCount;++i)
    {
    if(expInter&&(expInter[i]==1))
    // if(fabsf(pulses[i].fn-4152068.5f)<0.01f)
      {
      pulses[newPulseCount]=pulses[i];
      if(inter) { inter[newPulseCount]=inter[i]; }
      if(expInter) { expInter[newPulseCount]=expInter[i]; }
      ++newPulseCount;
      }
    }
  pulseCount=newPulseCount;
  }
#endif
if(repeat>1)
  {
  real64_t minToa=DBL_MAX,maxToa=-DBL_MAX;
  for(int i=0;i<pulseCount;++i)
    {
    const real64_t toa=pulses[i].toa;
    minToa=fmin(minToa,toa);
    maxToa=fmax(maxToa,toa);
    }
  pulses=Memory_reAlloc(pulses,repeat*pulseCount,PulseRecord);
  if(inter)
    { inter=Memory_reAlloc(inter,repeat*pulseCount,int32_t); }
  if(expInter)
    { expInter=Memory_reAlloc(expInter,repeat*pulseCount,int32_t); }
  double toaOffset=0.0;
  int j=pulseCount;
  for(int r=1;r<repeat;++r)
    {
    toaOffset+=maxToa-minToa;
    for(int i=0;i<pulseCount;++i,++j)
      {
      pulses[j]=pulses[i];
      pulses[j].toa+=toaOffset;
      if(inter) { inter[j]=inter[i]; }
      if(expInter) { expInter[j]=expInter[i]; }
      }
    }
  pulseCount*=repeat;
  }
if(outputFile)
  {
  if(outputExt&&!strcmp(outputExt,".h5"))
    { writeHdf5PulseFile(outputFile,pulseCount,pulses,inter,expInter); }
  else
    { writeRawPulseFile(outputFile,pulseCount,pulses); }
  }
if(expInter) { Memory_free(expInter); }
if(inter) { Memory_free(inter); }
Memory_free(pulses);
return 0;
}

//----------------------------------------------------------------------------
