//----------------------------------------------------------------------------
// Copyright (c) 2015-2016 Fabrice HARROUET (ENIB)
//
// This file is part of GpuElint.
//   http://www.enib.fr/~harrouet/gpuelint.html
//
// GpuElint is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GpuElint is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with GpuElint.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

using System;
using System.Runtime.InteropServices;

public class GpuElint
{

public class Error : Exception
  {
  public Error() : base() { }
  public Error(string message) : base(message) { }
  }

[StructLayout(LayoutKind.Sequential)]
public sealed class Properties
  {
  public int DTOA_SKIP_COUNT;
  public int MAX_MATCH_COUNT;
  public int glVersionMajor;
  public int glVersionMinor;
  public int glslVersion;
  public int maxWorkGroupCountX;
  public int maxWorkGroupCountY;
  public int maxWorkGroupCountZ;
  public int maxWorkGroupSizeX;
  public int maxWorkGroupSizeY;
  public int maxWorkGroupSizeZ;
  public int maxWorkGroupInvocations;
  public int maxUniformBlocks;
  public int maxTextureImageUnits;
  public int maxAtomicCounterBuffers;
  public int maxAtomicCounters;
  public int maxSharedMemorySize;
  public int maxUniformComponents;
  public int maxImageUniforms;
  public int maxCombinedUniformComponents;
  public int maxShaderStorageBlocks;
  public int warpSize;
  public int warpsPerStreamingMultiprocessor;
  public int streamingMultiprocessorCount;
  [MarshalAs(UnmanagedType.I1)] public bool hasBallot;
  [MarshalAs(UnmanagedType.I1)] public bool hasShuffle;
  }

public
GpuElint(GpuElint shared,
         bool errorify)
  {
  _errorify=errorify;
  _ntv=GpuElint_new(shared!=null ? shared._ntv : IntPtr.Zero);
  if(_errorify&&(_ntv==IntPtr.Zero))
    {
    throw new Error();
    }
  }

~GpuElint()
  {
  GpuElint_delete(_ntv);
  }

public
Properties
properties()
  {
  IntPtr result=GpuElint_properties(_ntv);
  if(_errorify&&result==IntPtr.Zero)
    {
    throw new Error();
    }
  Properties properties=null;
  if(result!=IntPtr.Zero)
    {
    properties=(Properties)Marshal.PtrToStructure(result,typeof(Properties));
    }
  return properties;
  }

public
float
elapsed()
  {
  float result=GpuElint_elapsed(_ntv);
  if(_errorify&&(result<0.0f))
    {
    throw new Error();
    }
  return result;
  }

public
bool
flush(bool wait)
  {
  bool success=GpuElint_flush(_ntv,wait);
  if(_errorify&&!success)
    {
    throw new Error();
    }
  return success;
  }

public
int
gpuAlloc(IntPtr capacity)
  {
  int id=GpuElint_gpuAlloc(_ntv,capacity);
  if(_errorify&&(id==0))
    {
    throw new Error();
    }
  return id;
  }

public
bool
gpuFree(int gpuBuffer)
  {
  bool success=GpuElint_gpuFree(_ntv,gpuBuffer);
  if(_errorify&&!success)
    {
    throw new Error();
    }
  return success;
  }

//
// FIXME: missing GpuElint_beginMapping()
//

//
// FIXME: missing GpuElint_endMapping()
//

public
bool
hostToGpu(int gpuDst,
          Array hostSrc,
          IntPtr dstOffset,
          IntPtr srcOffset,
          IntPtr size)
  {
  GCHandle pinnedHostSrc=GCHandle.Alloc(hostSrc,GCHandleType.Pinned);
  IntPtr hostSrcPointer=pinnedHostSrc.AddrOfPinnedObject();
  bool success=GpuElint_hostToGpu(_ntv,gpuDst,hostSrcPointer,
                                  dstOffset,srcOffset,size);
  pinnedHostSrc.Free();
  if(_errorify&&!success)
    {
    throw new Error();
    }
  return success;
  }

public
bool
gpuToHost(Array hostDst,
          int gpuSrc,
          IntPtr dstOffset,
          IntPtr srcOffset,
          IntPtr size)
  {
  GCHandle pinnedHostDst=GCHandle.Alloc(hostDst,GCHandleType.Pinned);
  IntPtr hostDstPointer=pinnedHostDst.AddrOfPinnedObject();
  bool success=GpuElint_gpuToHost(_ntv,hostDstPointer,gpuSrc,
                                  dstOffset,srcOffset,size);
  pinnedHostDst.Free();
  if(_errorify&&!success)
    {
    throw new Error();
    }
  return success;
  }

public
bool
gpuToGpu(int gpuDst,
         int gpuSrc,
         IntPtr dstOffset,
         IntPtr srcOffset,
         IntPtr size)
  {
  bool success=GpuElint_gpuToGpu(_ntv,gpuDst,gpuSrc,
                                 dstOffset,srcOffset,size);
  if(_errorify&&!success)
    {
    throw new Error();
    }
  return success;
  }

public
int
createMarker()
  {
  int marker=GpuElint_createMarker(_ntv);
  if(_errorify&&(marker==0))
    {
    throw new Error();
    }
  return marker;
  }

public
bool
destroyMarker(int marker)
  {
  bool success=GpuElint_destroyMarker(_ntv,marker);
  if(_errorify&&!success)
    {
    throw new Error();
    }
  return success;
  }

public
bool
setMarker(int marker)
  {
  bool success=GpuElint_setMarker(_ntv,marker);
  if(_errorify&&!success)
    {
    throw new Error();
    }
  return success;
  }

public
int
testMarker(int marker)
  {
  int result=GpuElint_testMarker(_ntv,marker);
  if(_errorify&&(result==-1))
    {
    throw new Error();
    }
  return result;
  }

public
float
markerInterval(int firstMarker,
               int secondMarker)
  {
  float interval=GpuElint_markerInterval(_ntv,firstMarker,secondMarker);
  if(_errorify&&(interval<0.0f))
    {
    throw new Error();
    }
  return interval;
  }

public
bool
zeroFill(int gpuDst,
         int begin,
         int end)
  {
  bool success=GpuElint_zeroFill(_ntv,gpuDst,begin,end);
  if(_errorify&&!success)
    {
    throw new Error();
    }
  return success;
  }

public
bool
countFill(int gpuDst,
          int begin,
          int end,
          int countStart,
          int countIncrement)
  {
  bool success=GpuElint_countFill(_ntv,gpuDst,begin,end,
                                  countStart,countIncrement);
  if(_errorify&&!success)
    {
    throw new Error();
    }
  return success;
  }

[StructLayout(LayoutKind.Sequential)]
public struct HistoMode1D
  {
  public int maxCount;
  public int weight;
  public int firstBin;
  public int lastBin;
  }

public
int
modes1D(HistoMode1D[] hostDst_modes,
        int gpuSrc_counts,
        float cutoffRatio,
        int binCount,
        bool cyclic)
  {
  GCHandle pinnedModes=GCHandle.Alloc(hostDst_modes,GCHandleType.Pinned);
  IntPtr modesPointer=pinnedModes.AddrOfPinnedObject();
  int modeCount=GpuElint_modes1D(_ntv,modesPointer,hostDst_modes.Length,
                                 gpuSrc_counts,cutoffRatio,
                                 binCount,cyclic);
  pinnedModes.Free();
  if(_errorify&&(modeCount==-1))
    {
    throw new Error();
    }
  return modeCount;
  }

[StructLayout(LayoutKind.Sequential)]
public struct HistoMode2D
  {
  public int maxCount;
  public int weight;
  public int firstBinA;
  public int lastBinA;
  public int firstBinB;
  public int lastBinB;
  }

public
int
modes2D(HistoMode2D[] hostDst_modes,
        int gpuSrc_counts,
        float cutoffRatio,
        int binCountA,
        bool cyclicA,
        int binCountB,
        bool cyclicB)
  {
  GCHandle pinnedModes=GCHandle.Alloc(hostDst_modes,GCHandleType.Pinned);
  IntPtr modesPointer=pinnedModes.AddrOfPinnedObject();
  int modeCount=GpuElint_modes2D(_ntv,modesPointer,hostDst_modes.Length,
                                 gpuSrc_counts,cutoffRatio,
                                 binCountA,cyclicA,
                                 binCountB,cyclicB);
  pinnedModes.Free();
  if(_errorify&&(modeCount==-1))
    {
    throw new Error();
    }
  return modeCount;
  }

[StructLayout(LayoutKind.Sequential)]
public struct HistoMode3D
  {
  public int maxCount;
  public int weight;
  public int firstBinA;
  public int lastBinA;
  public int firstBinB;
  public int lastBinB;
  public int firstBinC;
  public int lastBinC;
  }

public
int
modes3D(HistoMode3D[] hostDst_modes,
        int gpuSrc_counts,
        float cutoffRatio,
        int binCountA,
        bool cyclicA,
        int binCountB,
        bool cyclicB,
        int binCountC,
        bool cyclicC)
  {
  GCHandle pinnedModes=GCHandle.Alloc(hostDst_modes,GCHandleType.Pinned);
  IntPtr modesPointer=pinnedModes.AddrOfPinnedObject();
  int modeCount=GpuElint_modes3D(_ntv,modesPointer,hostDst_modes.Length,
                                 gpuSrc_counts,cutoffRatio,
                                 binCountA,cyclicA,
                                 binCountB,cyclicB,
                                 binCountC,cyclicC);
  pinnedModes.Free();
  if(_errorify&&(modeCount==-1))
    {
    throw new Error();
    }
  return modeCount;
  }

[StructLayout(LayoutKind.Sequential)]
public struct HistoMode4D
  {
  public int maxCount;
  public int weight;
  public int firstBinA;
  public int lastBinA;
  public int firstBinB;
  public int lastBinB;
  public int firstBinC;
  public int lastBinC;
  public int firstBinD;
  public int lastBinD;
  }

public
int
modes4D(HistoMode4D[] hostDst_modes,
        int gpuSrc_counts,
        float cutoffRatio,
        int binCountA,
        bool cyclicA,
        int binCountB,
        bool cyclicB,
        int binCountC,
        bool cyclicC,
        int binCountD,
        bool cyclicD)
  {
  GCHandle pinnedModes=GCHandle.Alloc(hostDst_modes,GCHandleType.Pinned);
  IntPtr modesPointer=pinnedModes.AddrOfPinnedObject();
  int modeCount=GpuElint_modes4D(_ntv,modesPointer,hostDst_modes.Length,
                                 gpuSrc_counts,cutoffRatio,
                                 binCountA,cyclicA,
                                 binCountB,cyclicB,
                                 binCountC,cyclicC,
                                 binCountD,cyclicD);
  pinnedModes.Free();
  if(_errorify&&(modeCount==-1))
    {
    throw new Error();
    }
  return modeCount;
  }

public
float
estimateMajorPri(int[] hostSrc_counts,
                 float[] hostSrc_sums,
                 int binCount,
                 float minDtoa,
                 float dtoaRatio,
                 float exclusivePriLowerLimit,
                 float minDtoaSum)
  {
  GCHandle pinnedCounts=GCHandle.Alloc(hostSrc_counts,GCHandleType.Pinned);
  GCHandle pinnedSums=GCHandle.Alloc(hostSrc_sums,GCHandleType.Pinned);
  IntPtr countsPointer=pinnedCounts.AddrOfPinnedObject();
  IntPtr sumsPointer=pinnedSums.AddrOfPinnedObject();
  float pri=GpuElint_estimateMajorPri(_ntv,countsPointer,sumsPointer,
                                      binCount,minDtoa,dtoaRatio,
                                      exclusivePriLowerLimit,minDtoaSum);
  pinnedSums.Free();
  pinnedCounts.Free();
  return pri;
  }

public
bool
minMax(int gpuSrc_indices,
       int pulseCount,
       int gpuSrc_values,
       out float minValue,
       out int minId,
       out float maxValue,
       out int maxId)
  {
  bool success=GpuElint_ID_minMax(_ntv,gpuSrc_indices,pulseCount,
                                  gpuSrc_values,
                                  out minValue,out minId,
                                  out maxValue,out maxId);
  if(_errorify&&!success)
    {
    throw new Error();
    }
  return success;
  }

public
bool
histogram1D(int gpuSrc_indices,
            int pulseCount,
            int gpuDst_counts,
            int gpuSrc_values,
            float minValue,
            float maxValue,
            int binCount)
  {
  bool success=GpuElint_ID_histogram1D(_ntv,gpuSrc_indices,pulseCount,
                                       gpuDst_counts,
                                       gpuSrc_values,
                                       minValue,maxValue,binCount);
  if(_errorify&&!success)
    {
    throw new Error();
    }
  return success;
  }

public
bool
histogram2D(int gpuSrc_indices,
            int pulseCount,
            int gpuDst_counts,
            int gpuSrc_valuesA,
            float minValueA,
            float maxValueA,
            int binCountA,
            int gpuSrc_valuesB,
            float minValueB,
            float maxValueB,
            int binCountB)
  {
  bool success=GpuElint_ID_histogram2D(_ntv,gpuSrc_indices,pulseCount,
                                       gpuDst_counts,
                                       gpuSrc_valuesA,
                                       minValueA,maxValueA,binCountA,
                                       gpuSrc_valuesB,
                                       minValueB,maxValueB,binCountB);
  if(_errorify&&!success)
    {
    throw new Error();
    }
  return success;
  }

public
bool
histogram3D(int gpuSrc_indices,
            int pulseCount,
            int gpuDst_counts,
            int gpuSrc_valuesA,
            float minValueA,
            float maxValueA,
            int binCountA,
            int gpuSrc_valuesB,
            float minValueB,
            float maxValueB,
            int binCountB,
            int gpuSrc_valuesC,
            float minValueC,
            float maxValueC,
            int binCountC)
  {
  bool success=GpuElint_ID_histogram3D(_ntv,gpuSrc_indices,pulseCount,
                                       gpuDst_counts,
                                       gpuSrc_valuesA,
                                       minValueA,maxValueA,binCountA,
                                       gpuSrc_valuesB,
                                       minValueB,maxValueB,binCountB,
                                       gpuSrc_valuesC,
                                       minValueC,maxValueC,binCountC);
  if(_errorify&&!success)
    {
    throw new Error();
    }
  return success;
  }

public
bool
histogram4D(int gpuSrc_indices,
            int pulseCount,
            int gpuDst_counts,
            int gpuSrc_valuesA,
            float minValueA,
            float maxValueA,
            int binCountA,
            int gpuSrc_valuesB,
            float minValueB,
            float maxValueB,
            int binCountB,
            int gpuSrc_valuesC,
            float minValueC,
            float maxValueC,
            int binCountC,
            int gpuSrc_valuesD,
            float minValueD,
            float maxValueD,
            int binCountD)
  {
  bool success=GpuElint_ID_histogram4D(_ntv,gpuSrc_indices,pulseCount,
                                       gpuDst_counts,
                                       gpuSrc_valuesA,
                                       minValueA,maxValueA,binCountA,
                                       gpuSrc_valuesB,
                                       minValueB,maxValueB,binCountB,
                                       gpuSrc_valuesC,
                                       minValueC,maxValueC,binCountC,
                                       gpuSrc_valuesD,
                                       minValueD,maxValueD,binCountD);
  if(_errorify&&!success)
    {
    throw new Error();
    }
  return success;
  }

public
int
filter1D(int gpuSrc_indices,
         int pulseCount,
         int gpuDst_indices,
         int gpuSrc_values,
         float minValue,
         float maxValue)
  {
  int remaining=GpuElint_ID_filter1D(_ntv,gpuSrc_indices,pulseCount,
                                     gpuDst_indices,
                                     gpuSrc_values,minValue,maxValue);
  if(_errorify&&(remaining==-1))
    {
    throw new Error();
    }
  return remaining;
  }

public
int
filter2D(int gpuSrc_indices,
         int pulseCount,
         int gpuDst_indices,
         int gpuSrc_valuesA,
         float minValueA,
         float maxValueA,
         int gpuSrc_valuesB,
         float minValueB,
         float maxValueB)
  {
  int remaining=GpuElint_ID_filter2D(_ntv,gpuSrc_indices,pulseCount,
                                     gpuDst_indices,
                                     gpuSrc_valuesA,minValueA,maxValueA,
                                     gpuSrc_valuesB,minValueB,maxValueB);
  if(_errorify&&(remaining==-1))
    {
    throw new Error();
    }
  return remaining;
  }

public
int
filter3D(int gpuSrc_indices,
         int pulseCount,
         int gpuDst_indices,
         int gpuSrc_valuesA,
         float minValueA,
         float maxValueA,
         int gpuSrc_valuesB,
         float minValueB,
         float maxValueB,
         int gpuSrc_valuesC,
         float minValueC,
         float maxValueC)
  {
  int remaining=GpuElint_ID_filter3D(_ntv,gpuSrc_indices,pulseCount,
                                     gpuDst_indices,
                                     gpuSrc_valuesA,minValueA,maxValueA,
                                     gpuSrc_valuesB,minValueB,maxValueB,
                                     gpuSrc_valuesC,minValueC,maxValueC);
  if(_errorify&&(remaining==-1))
    {
    throw new Error();
    }
  return remaining;
  }

public
int
filter4D(int gpuSrc_indices,
         int pulseCount,
         int gpuDst_indices,
         int gpuSrc_valuesA,
         float minValueA,
         float maxValueA,
         int gpuSrc_valuesB,
         float minValueB,
         float maxValueB,
         int gpuSrc_valuesC,
         float minValueC,
         float maxValueC,
         int gpuSrc_valuesD,
         float minValueD,
         float maxValueD)
  {
  int remaining=GpuElint_ID_filter4D(_ntv,gpuSrc_indices,pulseCount,
                                     gpuDst_indices,
                                     gpuSrc_valuesA,minValueA,maxValueA,
                                     gpuSrc_valuesB,minValueB,maxValueB,
                                     gpuSrc_valuesC,minValueC,maxValueC,
                                     gpuSrc_valuesD,minValueD,maxValueD);
  if(_errorify&&(remaining==-1))
    {
    throw new Error();
    }
  return remaining;
  }

public
bool
dtoaHistogram(int gpuSrc_indices,
              int pulseCount,
              int gpuSrc_toa64,
              int[] hostDst_counts,
              float[] hostDst_sums,
              int binCount,
              float minDtoa,
              float dtoaRatio)
  {
  GCHandle pinnedCounts=GCHandle.Alloc(hostDst_counts,GCHandleType.Pinned);
  GCHandle pinnedSums=GCHandle.Alloc(hostDst_sums,GCHandleType.Pinned);
  IntPtr countsPointer=pinnedCounts.AddrOfPinnedObject();
  IntPtr sumsPointer=pinnedSums.AddrOfPinnedObject();
  bool success=GpuElint_ID_dtoaHistogram(_ntv,gpuSrc_indices,pulseCount,
                                         gpuSrc_toa64,
                                         countsPointer,sumsPointer,
                                         binCount,minDtoa,dtoaRatio);
  pinnedSums.Free();
  pinnedCounts.Free();
  if(_errorify&&!success)
    {
    throw new Error();
    }
  return success;
  }

public
float
markPulses(int gpuSrc_indices,
           int pulseCount,
           int gpuSrc_toa64,
           int gpu_marks,
           int mark,
           float estimatedPri,
           float priRatioThreshold,
           int minMatchCount)
  {
  float pri=GpuElint_ID_markPulses(_ntv,gpuSrc_indices,pulseCount,
                                   gpuSrc_toa64,gpu_marks,mark,estimatedPri,
                                   priRatioThreshold,minMatchCount);
  if(_errorify&&(pri<0.0f))
    {
    throw new Error();
    }
  return pri;
  }

public
int
filterUnmarked(int gpuSrc_indices,
               int pulseCount,
               int gpuDst_indices,
               int gpuSrc_marks)
  {
  int remaining=GpuElint_ID_filterUnmarked(_ntv,gpuSrc_indices,pulseCount,
                                           gpuDst_indices,gpuSrc_marks);
  if(_errorify&&(remaining==-1))
    {
    throw new Error();
    }
  return remaining;
  }

//----------------------------------------------------------------------------

protected bool _errorify;
protected IntPtr _ntv;

protected const string _ntvLibName="GpuElint";

[DllImport(_ntvLibName)]
protected static extern
IntPtr
GpuElint_new(IntPtr shared);

[DllImport(_ntvLibName)]
protected static extern
void
GpuElint_delete(IntPtr ntv);

[DllImport(_ntvLibName)]
protected static extern
IntPtr
GpuElint_properties(IntPtr ntv);

[DllImport(_ntvLibName)]
protected static extern
float
GpuElint_elapsed(IntPtr ntv);

[DllImport(_ntvLibName)]
protected static extern
bool
GpuElint_flush(IntPtr ntv,
               bool wait);

[DllImport(_ntvLibName)]
protected static extern
int
GpuElint_gpuAlloc(IntPtr ntv,
                  IntPtr capacity);

[DllImport(_ntvLibName)]
protected static extern
bool
GpuElint_gpuFree(IntPtr ntv,
                 int gpuBuffer);

[DllImport(_ntvLibName)]
protected static extern
IntPtr
GpuElint_beginMapping(IntPtr ntv,
                      int gpuBuffer,
                      IntPtr offset,
                      IntPtr size,
                      bool readAccess,
                      bool writeAccess);

[DllImport(_ntvLibName)]
protected static extern
bool
GpuElint_endMapping(IntPtr ntv,
                    int gpuBuffer);

[DllImport(_ntvLibName)]
protected static extern
bool
GpuElint_hostToGpu(IntPtr ntv,
                   int gpuDst,
                   IntPtr hostSrc,
                   IntPtr dstOffset,
                   IntPtr srcOffset,
                   IntPtr size);

[DllImport(_ntvLibName)]
protected static extern
bool
GpuElint_gpuToHost(IntPtr ntv,
                   IntPtr hostDst,
                   int gpuSrc,
                   IntPtr dstOffset,
                   IntPtr srcOffset,
                   IntPtr size);

[DllImport(_ntvLibName)]
protected static extern
bool
GpuElint_gpuToGpu(IntPtr ntv,
                  int gpuDst,
                  int gpuSrc,
                  IntPtr dstOffset,
                  IntPtr srcOffset,
                  IntPtr size);

[DllImport(_ntvLibName)]
protected static extern
int
GpuElint_createMarker(IntPtr ntv);

[DllImport(_ntvLibName)]
protected static extern
bool
GpuElint_destroyMarker(IntPtr ntv,
                       int marker);

[DllImport(_ntvLibName)]
protected static extern
bool
GpuElint_setMarker(IntPtr ntv,
                   int marker);

[DllImport(_ntvLibName)]
protected static extern
int
GpuElint_testMarker(IntPtr ntv,
                    int marker);

[DllImport(_ntvLibName)]
protected static extern
float
GpuElint_markerInterval(IntPtr ntv,
                        int firstMarker,
                        int secondMarker);

[DllImport(_ntvLibName)]
protected static extern
bool
GpuElint_zeroFill(IntPtr ntv,
                  int gpuDst,
                  int begin,
                  int end);

[DllImport(_ntvLibName)]
protected static extern
bool
GpuElint_countFill(IntPtr ntv,
                   int gpuDst,
                   int begin,
                   int end,
                   int countStart,
                   int countIncrement);

[DllImport(_ntvLibName)]
protected static extern
int
GpuElint_modes1D(IntPtr ntv,
                 IntPtr hostDst_modes,
                 int modeCapacity,
                 int gpuSrc_counts,
                 float cutoffRatio,
                 int binCount,
                 bool cyclic);

[DllImport(_ntvLibName)]
protected static extern
int
GpuElint_modes2D(IntPtr ntv,
                 IntPtr hostDst_modes,
                 int modeCapacity,
                 int gpuSrc_counts,
                 float cutoffRatio,
                 int binCountA,
                 bool cyclicA,
                 int binCountB,
                 bool cyclicB);

[DllImport(_ntvLibName)]
protected static extern
int
GpuElint_modes3D(IntPtr ntv,
                 IntPtr hostDst_modes,
                 int modeCapacity,
                 int gpuSrc_counts,
                 float cutoffRatio,
                 int binCountA,
                 bool cyclicA,
                 int binCountB,
                 bool cyclicB,
                 int binCountC,
                 bool cyclicC);

[DllImport(_ntvLibName)]
protected static extern
int
GpuElint_modes4D(IntPtr ntv,
                 IntPtr hostDst_modes,
                 int modeCapacity,
                 int gpuSrc_counts,
                 float cutoffRatio,
                 int binCountA,
                 bool cyclicA,
                 int binCountB,
                 bool cyclicB,
                 int binCountC,
                 bool cyclicC,
                 int binCountD,
                 bool cyclicD);

[DllImport(_ntvLibName)]
protected static extern
float
GpuElint_estimateMajorPri(IntPtr ntv,
                          IntPtr hostSrc_counts,
                          IntPtr hostSrc_sums,
                          int binCount,
                          float minDtoa,
                          float dtoaRatio,
                          float exclusivePriLowerLimit,
                          float minDtoaSum);

[DllImport(_ntvLibName)]
protected static extern
bool
GpuElint_ID_minMax(IntPtr ntv,
                   int gpuSrc_indices,
                   int pulseCount,
                   int gpuSrc_values,
                   out float minValue,
                   out int minId,
                   out float maxValue,
                   out int maxId);

[DllImport(_ntvLibName)]
protected static extern
bool
GpuElint_ID_histogram1D(IntPtr ntv,
                        int gpuSrc_indices,
                        int pulseCount,
                        int gpuDst_counts,
                        int gpuSrc_values,
                        float minValue,
                        float maxValue,
                        int binCount);

[DllImport(_ntvLibName)]
protected static extern
bool
GpuElint_ID_histogram2D(IntPtr ntv,
                        int gpuSrc_indices,
                        int pulseCount,
                        int gpuDst_counts,
                        int gpuSrc_valuesA,
                        float minValueA,
                        float maxValueA,
                        int binCountA,
                        int gpuSrc_valuesB,
                        float minValueB,
                        float maxValueB,
                        int binCountB);

[DllImport(_ntvLibName)]
protected static extern
bool
GpuElint_ID_histogram3D(IntPtr ntv,
                        int gpuSrc_indices,
                        int pulseCount,
                        int gpuDst_counts,
                        int gpuSrc_valuesA,
                        float minValueA,
                        float maxValueA,
                        int binCountA,
                        int gpuSrc_valuesB,
                        float minValueB,
                        float maxValueB,
                        int binCountB,
                        int gpuSrc_valuesC,
                        float minValueC,
                        float maxValueC,
                        int binCountC);

[DllImport(_ntvLibName)]
protected static extern
bool
GpuElint_ID_histogram4D(IntPtr ntv,
                        int gpuSrc_indices,
                        int pulseCount,
                        int gpuDst_counts,
                        int gpuSrc_valuesA,
                        float minValueA,
                        float maxValueA,
                        int binCountA,
                        int gpuSrc_valuesB,
                        float minValueB,
                        float maxValueB,
                        int binCountB,
                        int gpuSrc_valuesC,
                        float minValueC,
                        float maxValueC,
                        int binCountC,
                        int gpuSrc_valuesD,
                        float minValueD,
                        float maxValueD,
                        int binCountD);

[DllImport(_ntvLibName)]
protected static extern
int
GpuElint_ID_filter1D(IntPtr ntv,
                     int gpuSrc_indices,
                     int pulseCount,
                     int gpuDst_indices,
                     int gpuSrc_values,
                     float minValue,
                     float maxValue);

[DllImport(_ntvLibName)]
protected static extern
int
GpuElint_ID_filter2D(IntPtr ntv,
                     int gpuSrc_indices,
                     int pulseCount,
                     int gpuDst_indices,
                     int gpuSrc_valuesA,
                     float minValueA,
                     float maxValueA,
                     int gpuSrc_valuesB,
                     float minValueB,
                     float maxValueB);

[DllImport(_ntvLibName)]
protected static extern
int
GpuElint_ID_filter3D(IntPtr ntv,
                     int gpuSrc_indices,
                     int pulseCount,
                     int gpuDst_indices,
                     int gpuSrc_valuesA,
                     float minValueA,
                     float maxValueA,
                     int gpuSrc_valuesB,
                     float minValueB,
                     float maxValueB,
                     int gpuSrc_valuesC,
                     float minValueC,
                     float maxValueC);

[DllImport(_ntvLibName)]
protected static extern
int
GpuElint_ID_filter4D(IntPtr ntv,
                     int gpuSrc_indices,
                     int pulseCount,
                     int gpuDst_indices,
                     int gpuSrc_valuesA,
                     float minValueA,
                     float maxValueA,
                     int gpuSrc_valuesB,
                     float minValueB,
                     float maxValueB,
                     int gpuSrc_valuesC,
                     float minValueC,
                     float maxValueC,
                     int gpuSrc_valuesD,
                     float minValueD,
                     float maxValueD);

[DllImport(_ntvLibName)]
protected static extern
bool
GpuElint_ID_dtoaHistogram(IntPtr ntv,
                          int gpuSrc_indices,
                          int pulseCount,
                          int gpuSrc_toa64,
                          IntPtr hostDst_counts,
                          IntPtr hostDst_sums,
                          int binCount,
                          float minDtoa,
                          float dtoaRatio);

[DllImport(_ntvLibName)]
protected static extern
float
GpuElint_ID_markPulses(IntPtr ntv,
                       int gpuSrc_indices,
                       int pulseCount,
                       int gpuSrc_toa64,
                       int gpu_marks,
                       int mark,
                       float estimatedPri,
                       float priRatioThreshold,
                       int minMatchCount);

[DllImport(_ntvLibName)]
protected static extern
int
GpuElint_ID_filterUnmarked(IntPtr ntv,
                           int gpuSrc_indices,
                           int pulseCount,
                           int gpuDst_indices,
                           int gpuSrc_marks);
}

//----------------------------------------------------------------------------
