//----------------------------------------------------------------------------
// Copyright (c) 2015-2016 Fabrice HARROUET (ENIB)
//
// This file is part of GpuElint.
//   http://www.enib.fr/~harrouet/gpuelint.html
//
// GpuElint is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GpuElint is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with GpuElint.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

import java.nio.Buffer;
import java.nio.ByteOrder;
import java.nio.ByteBuffer;
import java.nio.IntBuffer;
import java.nio.FloatBuffer;
import java.nio.DoubleBuffer;

/**
 * A GPU processing environment.
 * It provides means to create and transfer buffers between host and GPU
 * as well as predefined GPU processing algorithms.
 * Prior to any invocation, the current thread must hold an active OpenGL
 * context which must remain valid during the whole usage of this class.
 * If computation has to take place in another thread, this latter must
 * create its own processing environment which refers to the initial one;
 * their ressources will be shared into the initial OpenGL context.
 */
public class JavaGpuElint
{

/**
 * Errors thrown by JavaGpuElint.
 */
public class Error extends java.lang.Error
  {
  // FIXME: several error codes may have been provided here, just for a
  // ``nice looking'' purpose, even though it wouldn't actually change
  // anything for anyone...
  public Error() { super(); }
  public Error(String message) { super(message); }
  }

/**
 * GPU computing properties chosen or detected at initialisation.
 */
public final class Properties
  {
  public int DTOA_SKIP_COUNT;
  public int MAX_MATCH_COUNT;
  public int glVersionMajor;
  public int glVersionMinor;
  public int glslVersion;
  public int maxWorkGroupCountX;
  public int maxWorkGroupCountY;
  public int maxWorkGroupCountZ;
  public int maxWorkGroupSizeX;
  public int maxWorkGroupSizeY;
  public int maxWorkGroupSizeZ;
  public int maxWorkGroupInvocations;
  public int maxUniformBlocks;
  public int maxTextureImageUnits;
  public int maxAtomicCounterBuffers;
  public int maxAtomicCounters;
  public int maxSharedMemorySize;
  public int maxUniformComponents;
  public int maxImageUniforms;
  public int maxCombinedUniformComponents;
  public int maxShaderStorageBlocks;
  public int warpSize;
  public int warpsPerStreamingMultiprocessor;
  public int streamingMultiprocessorCount;
  public boolean hasBallot;
  public boolean hasShuffle;
  }

/**
 * Constructs a new GPU processing environment.
 * @param  shared
 *         A shared GPU processing environment used within another thread
 *         or null.
 * @param  errorify
 *         Tells whether to throw errors on failure or not.
 */
public
JavaGpuElint(JavaGpuElint shared,
             boolean errorify)
  {
  _errorify=errorify;
  _ntv=_new(shared!=null ? shared._ntv : 0L);
  if(_errorify&&(_ntv==0))
    {
    throw new Error();
    }
  }

/**
 * Provides a copy of the GPU computing properties that were chosen or
 * detected at initialisation.
 * @return A set of properties (null on failure).
 */
public
Properties
properties()
  {
  Properties properties=new Properties();
  boolean success=_properties(_ntv,properties);
  if(_errorify&&!success)
    {
    throw new Error();
    }
  return success ? properties : null;
  }

/**
 * Gives the time elapsed since the initial GpuElint creation.
 * @return The time in seconds (-1.0f on failure).
 */
public
float
elapsed()
  {
  float result=_elapsed(_ntv);
  if(_errorify&&(result<0.0f))
    {
    throw new Error();
    }
  return result;
  }

/**
 * Ensures all the preceding commands have been sent to the GPU.
 * @param  wait
 *         Tells whether to wait for these commands to complete or not.
 * @return Success.
 */
public
boolean
flush(boolean wait)
  {
  boolean success=_flush(_ntv,wait);
  if(_errorify&&!success)
    {
    throw new Error();
    }
  return success;
  }

/**
 * Allocates a new host-buffer suitable for GPU transfers
 * (direct-allocation, native-order).
 * @param  capacity
 *         The capacity of the new host-buffer, in bytes.
 * @return The new host-buffer.
 */
public static
ByteBuffer
hostAlloc(long capacity)
  {
  // JAVA WEIRDNESS: ByteBuffer.allocateDirect() expects capacity to be an
  // int although GetDirectBufferCapacity() JNI function returns a jlong!
  // (Wouldn't it be foreseeable, in a very far future, that computers could
  // handle more than 2GB of memory?)
  // By the way, I wish I could have provided a class derived from ByteBuffer
  // to enforce direct-allocation and native-order, but it seems that,
  // according to Java, this concept is out of scope...
  // (Wasn't Java supposed to support object-oriented design?)
  return ByteBuffer.allocateDirect((int)capacity)
                   .order(ByteOrder.nativeOrder());
  }

/**
 * Allocates a new GPU-buffer.
 * @param  capacity
 *         The capacity of the new GPU-buffer, in bytes.
 * @return The identifier of the new GPU-buffer (0 on failure).
 */
public
int
gpuAlloc(long capacity)
  {
  int id=_gpuAlloc(_ntv,capacity);
  if(_errorify&&(id==0))
    {
    throw new Error();
    }
  return id;
  }

/**
 * Frees a GPU-buffer that will not be used anymore.
 * @param  gpuBuffer
 *         The identifier of the GPU-buffer to be freed.
 * @return Success.
 */
public
boolean
gpuFree(int gpuBuffer)
  {
  boolean success=_gpuFree(_ntv,gpuBuffer);
  if(_errorify&&!success)
    {
    throw new Error();
    }
  return success;
  }

//
// FIXME: missing GpuElint_beginMapping()
//

//
// FIXME: missing GpuElint_endMapping()
//

/**
 * Copies some data from a host-buffer to a GPU-buffer.
 * @param  gpuDst
 *         The identifier of the destination GPU-buffer.
 * @param  hostSrc
 *         The source host-buffer.
 * @param  dstOffset
 *         The offset inside the destination GPU-buffer, in bytes.
 * @param  srcOffset
 *         The offset inside the source host-buffer, in bytes.
 * @param  size
 *         The amount of data to be copied, in bytes.
 * @return Success.
 */
public
boolean
hostToGpu(int gpuDst,
          Buffer hostSrc,
          long dstOffset,
          long srcOffset,
          long size)
  {
  boolean success=_hostToGpu(_ntv,gpuDst,hostSrc,dstOffset,srcOffset,size);
  if(_errorify&&!success)
    {
    throw new Error();
    }
  return success;
  }

/**
 * Copies some data from a GPU-buffer to a host-buffer.
 * @param  hostDst
 *         The destination host-buffer.
 * @param  gpuSrc
 *         The identifier of the source GPU-buffer.
 * @param  dstOffset
 *         The offset inside the destination host-buffer, in bytes.
 * @param  srcOffset
 *         The offset inside the source GPU-buffer, in bytes.
 * @param  size
 *         The amount of data to be copied, in bytes.
 * @return Success.
 */
public
boolean
gpuToHost(Buffer hostDst,
          int gpuSrc,
          long dstOffset,
          long srcOffset,
          long size)
  {
  boolean success=_gpuToHost(_ntv,hostDst,gpuSrc,dstOffset,srcOffset,size);
  if(_errorify&&!success)
    {
    throw new Error();
    }
  return success;
  }

/**
 * Copies some data from a GPU-buffer to another.
 * @param  gpuDst
 *         The identifier of the destination GPU-buffer.
 * @param  gpuSrc
 *         The identifier of the source GPU-buffer.
 * @param  dstOffset
 *         The offset inside the destination GPU-buffer, in bytes.
 * @param  srcOffset
 *         The offset inside the source GPU-buffer, in bytes.
 * @param  size
 *         The amount of data to be copied, in bytes.
 * @return Success.
 */
public
boolean
gpuToGpu(int gpuDst,
         int gpuSrc,
         long dstOffset,
         long srcOffset,
         long size)
  {
  boolean success=_gpuToGpu(_ntv,gpuDst,gpuSrc,dstOffset,srcOffset,size);
  if(_errorify&&!success)
    {
    throw new Error();
    }
  return success;
  }

/**
 * Creates a new GPU-time-marker to measure GPU processing durations.
 * @return The identifier of the new GPU-time-marker (0 on failure).
 */
public
int
createMarker()
  {
  int marker=_createMarker(_ntv);
  if(_errorify&&(marker==0))
    {
    throw new Error();
    }
  return marker;
  }

/**
 * Destroys a GPU-time-marker that will not be used anymore.
 * @param  marker
 *         The identifier of the GPU-time-marker to be destroyed.
 * @return Success.
 */
public
boolean
destroyMarker(int marker)
  {
  boolean success=_destroyMarker(_ntv,marker);
  if(_errorify&&!success)
    {
    throw new Error();
    }
  return success;
  }

/**
 * Inserts a GPU-time-marker in the processing stream of the GPU.
 * @param  marker
 *         The identifier of the GPU-time-marker to be inserted.
 * @return Success.
 */
public
boolean
setMarker(int marker)
  {
  boolean success=_setMarker(_ntv,marker);
  if(_errorify&&!success)
    {
    throw new Error();
    }
  return success;
  }

/**
 * Tests whether a GPU-time-marker has been reached by the GPU processing
 * stream or not.
 * @param  marker
 *         The identifier of the GPU-time-marker to be tested.
 * @return 0 if the marker is not reached, 1 if it is (-1 on failure).
 */
public
int
testMarker(int marker)
  {
  int result=_testMarker(_ntv,marker);
  if(_errorify&&(result==-1))
    {
    throw new Error();
    }
  return result;
  }

/**
 * Gives the time elapsed between two GPU-time-markers.
 * @param  firstMarker
 *         The identifier of the first GPU-time-marker to reach.
 * @param  secondMarker
 *         The identifier of the second GPU-time-marker to reach.
 * @return The duration in seconds or (-1.0f on failure).
 */
public
float
markerInterval(int firstMarker,
               int secondMarker)
  {
  float interval=_markerInterval(_ntv,firstMarker,secondMarker);
  if(_errorify&&(interval<0.0f))
    {
    throw new Error();
    }
  return interval;
  }

/**
 * Sets to zero the float or integer value of each element.
 * @param  gpuDst
 *         The identifier of the destination GPU-buffer.
 * @param  begin
 *         The index of the first element to consider.
 * @param  end
 *         The index after the last element to consider.
 * @return Success.
 */
public
boolean
zeroFill(int gpuDst,
         int begin,
         int end)
  {
  boolean success=_zeroFill(_ntv,gpuDst,begin,end);
  if(_errorify&&!success)
    {
    throw new Error();
    }
  return success;
  }

/**
 * Initialises the integer value of each element by counting them
 * as follows:
 *   value of element at (begin+i) = countStart + i * countIncrement
 * @param  gpuDst
 *         The identifier of the destination GPU-buffer.
 * @param  begin
 *         The index of the first element to consider.
 * @param  end
 *         The index after the last element to consider.
 * @param  countStart
 *         The starting value to count from.
 * @param  countIncrement
 *         The increment between each consecutive value.
 * @return Success.
 */
public
boolean
countFill(int gpuDst,
          int begin,
          int end,
          int countStart,
          int countIncrement)
  {
  boolean success=_countFill(_ntv,gpuDst,begin,end,
                             countStart,countIncrement);
  if(_errorify&&!success)
    {
    throw new Error();
    }
  return success;
  }

/**
 * Description of a mode detected in a 1D histogram.
 */
public final class HistoMode1D
  {
  public int maxCount;
  public int weight;
  public int firstBin;
  public int lastBin;
  }

/**
 * Detects the modes of a 1D histogram.
 * @param  hostDst_modes
 *         The destination array to store the detected modes
 *         (must be preallocated to the maximal mode capacity).
 * @param  gpuSrc_counts
 *         The identifier of the source GPU-buffer containing the
 *         counts of each bin.
 * @param  cutoffRatio
 *         The cutoff level of a mode relative its maximum.
 * @param  binCount
 *         The number of bins in the histogram.
 * @param  cyclic
 *         Tells if detection should cycle through histogram boundaries.
 * @return The number of detected modes (-1 on failure).
 */
public
int
modes1D(HistoMode1D[] hostDst_modes,
        int gpuSrc_counts,
        float cutoffRatio,
        int binCount,
        boolean cyclic)
  {
  // JAVA WEIRDNESS: it's a shame that Java does not provide a contiguous
  //                 storage for an array of structures; we thus have to
  //                 repeatedly allocate/release such a storage
  //                 and create/copy/destroy loads of small objects.
  int sizeof_int=Integer.SIZE/8;
  int fieldCount=2+2*1;
  IntBuffer modes=hostAlloc(hostDst_modes.length*fieldCount*sizeof_int)
                  .asIntBuffer();
  int modeCount=_modes1D(_ntv,modes,gpuSrc_counts,cutoffRatio,
                         binCount,cyclic);
  if(_errorify&&(modeCount==-1))
    {
    throw new Error();
    }
  for(int i=0;i<modeCount;++i)
    {
    HistoMode1D m=new HistoMode1D();
    m.maxCount=modes.get(fieldCount*i+0);
    m.weight  =modes.get(fieldCount*i+1);
    m.firstBin=modes.get(fieldCount*i+2);
    m.lastBin =modes.get(fieldCount*i+3);
    hostDst_modes[i]=m;
    }
  return modeCount;
  }

/**
 * Description of a mode detected in a 2D histogram.
 */
public final class HistoMode2D
  {
  public int maxCount;
  public int weight;
  public int firstBinA;
  public int lastBinA;
  public int firstBinB;
  public int lastBinB;
  }

/**
 * Detects the modes of a 2D histogram.
 * The histogram is firstly ordered along A values then B.
 * @param  hostDst_modes
 *         The destination array to store the detected modes
 *         (must be preallocated to the maximal mode capacity).
 * @param  gpuSrc_counts
 *         The identifier of the source GPU-buffer containing the
 *         counts of each bin.
 * @param  cutoffRatio
 *         The cutoff level of a mode relative its maximum.
 * @param  binCountA
 *         The number of bins for A values in the histogram.
 * @param  cyclicA
 *         Tells if detection should cycle through histogram boundaries
 *         for A values.
 * @param  binCountB
 *         The number of bins for B values in the histogram.
 * @param  cyclicB
 *         Tells if detection should cycle through histogram boundaries
 *         for B values.
 * @return The number of detected modes (-1 on failure).
 */
public
int
modes2D(HistoMode2D[] hostDst_modes,
        int gpuSrc_counts,
        float cutoffRatio,
        int binCountA,
        boolean cyclicA,
        int binCountB,
        boolean cyclicB)
  {
  // JAVA WEIRDNESS: it's a shame that Java does not provide a contiguous
  //                 storage for an array of structures; we thus have to
  //                 repeatedly allocate/release such a storage
  //                 and create/copy/destroy loads of small objects.
  int sizeof_int=Integer.SIZE/8;
  int fieldCount=2+2*2;
  IntBuffer modes=hostAlloc(hostDst_modes.length*fieldCount*sizeof_int)
                  .asIntBuffer();
  int modeCount=_modes2D(_ntv,modes,gpuSrc_counts,cutoffRatio,
                         binCountA,cyclicA,
                         binCountB,cyclicB);
  if(_errorify&&(modeCount==-1))
    {
    throw new Error();
    }
  for(int i=0;i<modeCount;++i)
    {
    HistoMode2D m=new HistoMode2D();
    m.maxCount =modes.get(fieldCount*i+0);
    m.weight   =modes.get(fieldCount*i+1);
    m.firstBinA=modes.get(fieldCount*i+2);
    m.lastBinA =modes.get(fieldCount*i+3);
    m.firstBinB=modes.get(fieldCount*i+4);
    m.lastBinB =modes.get(fieldCount*i+5);
    hostDst_modes[i]=m;
    }
  return modeCount;
  }

/**
 * Description of a mode detected in a 3D histogram.
 */
public final class HistoMode3D
  {
  public int maxCount;
  public int weight;
  public int firstBinA;
  public int lastBinA;
  public int firstBinB;
  public int lastBinB;
  public int firstBinC;
  public int lastBinC;
  }

/**
 * Detects the modes of a 3D histogram.
 * The histogram is firstly ordered along A values then B and C.
 * @param  hostDst_modes
 *         The destination array to store the detected modes
 *         (must be preallocated to the maximal mode capacity).
 * @param  gpuSrc_counts
 *         The identifier of the source GPU-buffer containing the
 *         counts of each bin.
 * @param  cutoffRatio
 *         The cutoff level of a mode relative its maximum.
 * @param  binCountA
 *         The number of bins for A values in the histogram.
 * @param  cyclicA
 *         Tells if detection should cycle through histogram boundaries
 *         for A values.
 * @param  binCountB
 *         The number of bins for B values in the histogram.
 * @param  cyclicB
 *         Tells if detection should cycle through histogram boundaries
 *         for B values.
 * @param  binCountC
 *         The number of bins for C values in the histogram.
 * @param  cyclicC
 *         Tells if detection should cycle through histogram boundaries
 *         for C values.
 * @return The number of detected modes (-1 on failure).
 */
public
int
modes3D(HistoMode3D[] hostDst_modes,
        int gpuSrc_counts,
        float cutoffRatio,
        int binCountA,
        boolean cyclicA,
        int binCountB,
        boolean cyclicB,
        int binCountC,
        boolean cyclicC)
  {
  // JAVA WEIRDNESS: it's a shame that Java does not provide a contiguous
  //                 storage for an array of structures; we thus have to
  //                 repeatedly allocate/release such a storage
  //                 and create/copy/destroy loads of small objects.
  int sizeof_int=Integer.SIZE/8;
  int fieldCount=2+2*3;
  IntBuffer modes=hostAlloc(hostDst_modes.length*fieldCount*sizeof_int)
                  .asIntBuffer();
  int modeCount=_modes3D(_ntv,modes,gpuSrc_counts,cutoffRatio,
                         binCountA,cyclicA,
                         binCountB,cyclicB,
                         binCountC,cyclicC);
  if(_errorify&&(modeCount==-1))
    {
    throw new Error();
    }
  for(int i=0;i<modeCount;++i)
    {
    HistoMode3D m=new HistoMode3D();
    m.maxCount =modes.get(fieldCount*i+0);
    m.weight   =modes.get(fieldCount*i+1);
    m.firstBinA=modes.get(fieldCount*i+2);
    m.lastBinA =modes.get(fieldCount*i+3);
    m.firstBinB=modes.get(fieldCount*i+4);
    m.lastBinB =modes.get(fieldCount*i+5);
    m.firstBinC=modes.get(fieldCount*i+6);
    m.lastBinC =modes.get(fieldCount*i+7);
    hostDst_modes[i]=m;
    }
  return modeCount;
  }

/**
 * Description of a mode detected in a 4D histogram.
 */
public final class HistoMode4D
  {
  public int maxCount;
  public int weight;
  public int firstBinA;
  public int lastBinA;
  public int firstBinB;
  public int lastBinB;
  public int firstBinC;
  public int lastBinC;
  public int firstBinD;
  public int lastBinD;
  }

/**
 * Detects the modes of a 4D histogram.
 * The histogram is firstly ordered along A values then B, C and D.
 * @param  hostDst_modes
 *         The destination array to store the detected modes
 *         (must be preallocated to the maximal mode capacity).
 * @param  gpuSrc_counts
 *         The identifier of the source GPU-buffer containing the
 *         counts of each bin.
 * @param  cutoffRatio
 *         The cutoff level of a mode relative its maximum.
 * @param  binCountA
 *         The number of bins for A values in the histogram.
 * @param  cyclicA
 *         Tells if detection should cycle through histogram boundaries
 *         for A values.
 * @param  binCountB
 *         The number of bins for B values in the histogram.
 * @param  cyclicB
 *         Tells if detection should cycle through histogram boundaries
 *         for B values.
 * @param  binCountC
 *         The number of bins for C values in the histogram.
 * @param  cyclicC
 *         Tells if detection should cycle through histogram boundaries
 *         for C values.
 * @param  binCountD
 *         The number of bins for D values in the histogram.
 * @param  cyclicD
 *         Tells if detection should cycle through histogram boundaries
 *         for D values.
 * @return The number of detected modes (-1 on failure).
 */
public
int
modes4D(HistoMode4D[] hostDst_modes,
        int gpuSrc_counts,
        float cutoffRatio,
        int binCountA,
        boolean cyclicA,
        int binCountB,
        boolean cyclicB,
        int binCountC,
        boolean cyclicC,
        int binCountD,
        boolean cyclicD)
  {
  // JAVA WEIRDNESS: it's a shame that Java does not provide a contiguous
  //                 storage for an array of structures; we thus have to
  //                 repeatedly allocate/release such a storage
  //                 and create/copy/destroy loads of small objects.
  int sizeof_int=Integer.SIZE/8;
  int fieldCount=2+2*4;
  IntBuffer modes=hostAlloc(hostDst_modes.length*fieldCount*sizeof_int)
                  .asIntBuffer();
  int modeCount=_modes4D(_ntv,modes,gpuSrc_counts,cutoffRatio,
                         binCountA,cyclicA,
                         binCountB,cyclicB,
                         binCountC,cyclicC,
                         binCountD,cyclicD);
  if(_errorify&&(modeCount==-1))
    {
    throw new Error();
    }
  for(int i=0;i<modeCount;++i)
    {
    HistoMode4D m=new HistoMode4D();
    m.maxCount =modes.get(fieldCount*i+0);
    m.weight   =modes.get(fieldCount*i+1);
    m.firstBinA=modes.get(fieldCount*i+2);
    m.lastBinA =modes.get(fieldCount*i+3);
    m.firstBinB=modes.get(fieldCount*i+4);
    m.lastBinB =modes.get(fieldCount*i+5);
    m.firstBinC=modes.get(fieldCount*i+6);
    m.lastBinC =modes.get(fieldCount*i+7);
    m.firstBinD=modes.get(fieldCount*i+8);
    m.lastBinD =modes.get(fieldCount*i+9);
    hostDst_modes[i]=m;
    }
  return modeCount;
  }

/**
 * Attempts to estimate the pulse-repetition-interval (PRI) which is the
 * most significant in a DTOA histogram.
 * @param  hostSrc_counts
 *         The source host-buffer containing the DTOA counts of each bin.
 * @param  hostSrc_sums
 *         The source host-buffer containing the DTOA sums of each bin.
 * @param  binCount
 *         The number of bins of the histogram.
 * @param  minDtoa
 *         The lower bound (strictly greater than 0.0f) for a DTOA to be
 *         inserted in the histogram.
 * @param  dtoaRatio
 *         The multiplicative factor (strictly greater than 1.0f) which
 *         determines the logarithmic width of each bin.
 * @param  exclusivePriLowerLimit
 *         The strict lower limit (possibly 0.0f) above which a PRI will be
 *         considered in the estimation.
 * @param  minDtoaSum
 *         The minimal DTOA sum for a bin to be considered.
 * @return The most significant pulse-repetition-interval or 0.0f
 */
public
float
estimateMajorPri(IntBuffer hostSrc_counts,
                 FloatBuffer hostSrc_sums,
                 int binCount,
                 float minDtoa,
                 float dtoaRatio,
                 float exclusivePriLowerLimit,
                 float minDtoaSum)
  {
  return _estimateMajorPri(_ntv,hostSrc_counts,hostSrc_sums,
                           binCount,minDtoa,dtoaRatio,
                           exclusivePriLowerLimit,minDtoaSum);
  }

/**
 * Result of the minMax() operation.
 */
public final class MinMaxResult
  {
  public float minValue;
  public int minId;
  public float maxValue;
  public int maxId;
  }

/**
 * Determines the min and max of the float value of each pulse,
 * as well as their respective indices.
 * @param  gpuSrc_indices
 *         The identifier of the source GPU-buffer containing the indices of the
 *         pulses to analyse.
 * @param  pulseCount
 *         The number of pulses to consider.
 * @param  gpuSrc_values
 *         The identifier of the source GPU-buffer containing the values
 *         to analyse.
 * @return The values and indices of the min/max elements (null on failure).
 */
public
MinMaxResult
minMax(int gpuSrc_indices,
       int pulseCount,
       int gpuSrc_values)
  {
  MinMaxResult result=new MinMaxResult();
  boolean success=_minMax(_ntv,gpuSrc_indices,pulseCount,
                          gpuSrc_values,result);
  if(_errorify&&!success)
    {
    throw new Error();
    }
  return success ? result : null;
  }

/**
 * Builds up an histogram of the float value of each pulse.
 * The destination GPU-buffer is initially reset to zero.
 * @param  gpuSrc_indices
 *         The identifier of the source GPU-buffer containing the indices
 *         of the pulses to analyse.
 * @param  pulseCount
 *         The number of pulses to consider.
 * @param  gpuDst_counts
 *         The identifier of the destination GPU-buffer to store the
 *         counts of each bin.
 * @param  gpuSrc_values
 *         The identifier of the source GPU-buffer containing the values
 *         to analyse.
 * @param  minValue
 *         The lower bound of the range used for counting values.
 * @param  maxValue
 *         The upper bound of the range used for counting values.
 * @param  binCount
 *         The number of bins of the range used for counting values.
 * @return Success.
 */
public
boolean
histogram1D(int gpuSrc_indices,
            int pulseCount,
            int gpuDst_counts,
            int gpuSrc_values,
            float minValue,
            float maxValue,
            int binCount)
  {
  boolean success=_histogram1D(_ntv,gpuSrc_indices,pulseCount,gpuDst_counts,
                               gpuSrc_values,
                               minValue,maxValue,binCount);
  if(_errorify&&!success)
    {
    throw new Error();
    }
  return success;
  }

/**
 * Builds up a 2D histogram of the float (A,B) values of each pulse.
 * The destination GPU-buffer is initially reset to zero and is firstly
 * ordered along A then B.
 * @param  gpuSrc_indices
 *         The identifier of the source GPU-buffer containing the indices
 *         of the pulses to analyse.
 * @param  pulseCount
 *         The number of pulses to consider.
 * @param  gpuDst_counts
 *         The identifier of the destination GPU-buffer to store the
 *         counts of each bin.
 * @param  gpuSrc_valuesA
 *         The identifier of the source GPU-buffer containing the A values
 *         to analyse.
 * @param  minValueA
 *         The lower bound of the range used for counting A values.
 * @param  maxValueA
 *         The upper bound of the range used for counting A values.
 * @param  binCountA
 *         The number of bins of the range used for counting A values.
 * @param  gpuSrc_valuesB
 *         The identifier of the source GPU-buffer containing the B values
 *         to analyse.
 * @param  minValueB
 *         The lower bound of the range used for counting B values.
 * @param  maxValueB
 *         The upper bound of the range used for counting B values.
 * @param  binCountB
 *         The number of bins of the range used for counting B values.
 * @return Success.
 */
public
boolean
histogram2D(int gpuSrc_indices,
            int pulseCount,
            int gpuDst_counts,
            int gpuSrc_valuesA,
            float minValueA,
            float maxValueA,
            int binCountA,
            int gpuSrc_valuesB,
            float minValueB,
            float maxValueB,
            int binCountB)
  {
  boolean success=_histogram2D(_ntv,gpuSrc_indices,pulseCount,gpuDst_counts,
                               gpuSrc_valuesA,
                               minValueA,maxValueA,binCountA,
                               gpuSrc_valuesB,
                               minValueB,maxValueB,binCountB);
  if(_errorify&&!success)
    {
    throw new Error();
    }
  return success;
  }

/**
 * Builds up a 3D histogram of the float (A,B,C) values of each pulse.
 * The destination GPU-buffer is initially reset to zero and is firstly
 * ordered along A then B, and C.
 * @param  gpuSrc_indices
 *         The identifier of the source GPU-buffer containing the indices
 *         of the pulses to analyse.
 * @param  pulseCount
 *         The number of pulses to consider.
 * @param  gpuDst_counts
 *         The identifier of the destination GPU-buffer to store the
 *         counts of each bin.
 * @param  gpuSrc_valuesA
 *         The identifier of the source GPU-buffer containing the A values
 *         to analyse.
 * @param  minValueA
 *         The lower bound of the range used for counting A values.
 * @param  maxValueA
 *         The upper bound of the range used for counting A values.
 * @param  binCountA
 *         The number of bins of the range used for counting A values.
 * @param  gpuSrc_valuesB
 *         The identifier of the source GPU-buffer containing the B values
 *         to analyse.
 * @param  minValueB
 *         The lower bound of the range used for counting B values.
 * @param  maxValueB
 *         The upper bound of the range used for counting B values.
 * @param  binCountB
 *         The number of bins of the range used for counting B values.
 * @param  gpuSrc_valuesC
 *         The identifier of the source GPU-buffer containing the C values
 *         to analyse.
 * @param  minValueC
 *         The lower bound of the range used for counting C values.
 * @param  maxValueC
 *         The upper bound of the range used for counting C values.
 * @param  binCountC
 *         The number of bins of the range used for counting C values.
 * @return Success.
 */
public
boolean
histogram3D(int gpuSrc_indices,
            int pulseCount,
            int gpuDst_counts,
            int gpuSrc_valuesA,
            float minValueA,
            float maxValueA,
            int binCountA,
            int gpuSrc_valuesB,
            float minValueB,
            float maxValueB,
            int binCountB,
            int gpuSrc_valuesC,
            float minValueC,
            float maxValueC,
            int binCountC)
  {
  boolean success=_histogram3D(_ntv,gpuSrc_indices,pulseCount,gpuDst_counts,
                               gpuSrc_valuesA,
                               minValueA,maxValueA,binCountA,
                               gpuSrc_valuesB,
                               minValueB,maxValueB,binCountB,
                               gpuSrc_valuesC,
                               minValueC,maxValueC,binCountC);
  if(_errorify&&!success)
    {
    throw new Error();
    }
  return success;
  }

/**
 * Builds up a 4D histogram of the float (A,B,C,D) values of each pulse.
 * The destination GPU-buffer is initially reset to zero and is firstly
 * ordered along A then B, C and D.
 * @param  gpuSrc_indices
 *         The identifier of the source GPU-buffer containing the indices
 *         of the pulses to analyse.
 * @param  pulseCount
 *         The number of pulses to consider.
 * @param  gpuDst_counts
 *         The identifier of the destination GPU-buffer to store the
 *         counts of each bin.
 * @param  gpuSrc_valuesA
 *         The identifier of the source GPU-buffer containing the A values
 *         to analyse.
 * @param  minValueA
 *         The lower bound of the range used for counting A values.
 * @param  maxValueA
 *         The upper bound of the range used for counting A values.
 * @param  binCountA
 *         The number of bins of the range used for counting A values.
 * @param  gpuSrc_valuesB
 *         The identifier of the source GPU-buffer containing the B values
 *         to analyse.
 * @param  minValueB
 *         The lower bound of the range used for counting B values.
 * @param  maxValueB
 *         The upper bound of the range used for counting B values.
 * @param  binCountB
 *         The number of bins of the range used for counting B values.
 * @param  gpuSrc_valuesC
 *         The identifier of the source GPU-buffer containing the C values
 *         to analyse.
 * @param  minValueC
 *         The lower bound of the range used for counting C values.
 * @param  maxValueC
 *         The upper bound of the range used for counting C values.
 * @param  binCountC
 *         The number of bins of the range used for counting C values.
 * @param  gpuSrc_valuesD
 *         The identifier of the source GPU-buffer containing the D values
 *         to analyse.
 * @param  minValueD
 *         The lower bound of the range used for counting D values.
 * @param  maxValueD
 *         The upper bound of the range used for counting D values.
 * @param  binCountD
 *         The number of bins of the range used for counting D values.
 * @return Success.
 */
public
boolean
histogram4D(int gpuSrc_indices,
            int pulseCount,
            int gpuDst_counts,
            int gpuSrc_valuesA,
            float minValueA,
            float maxValueA,
            int binCountA,
            int gpuSrc_valuesB,
            float minValueB,
            float maxValueB,
            int binCountB,
            int gpuSrc_valuesC,
            float minValueC,
            float maxValueC,
            int binCountC,
            int gpuSrc_valuesD,
            float minValueD,
            float maxValueD,
            int binCountD)
  {
  boolean success=_histogram4D(_ntv,gpuSrc_indices,pulseCount,gpuDst_counts,
                               gpuSrc_valuesA,
                               minValueA,maxValueA,binCountA,
                               gpuSrc_valuesB,
                               minValueB,maxValueB,binCountB,
                               gpuSrc_valuesC,
                               minValueC,maxValueC,binCountC,
                               gpuSrc_valuesD,
                               minValueD,maxValueD,binCountD);
  if(_errorify&&!success)
    {
    throw new Error();
    }
  return success;
  }

/**
 * Keep only the pulses which float value is in the specified interval.
 * @param  gpuSrc_indices
 *         The identifier of the source GPU-buffer containing the indices
 *         of the pulses to analyse.
 * @param  pulseCount
 *         The number of pulses to consider.
 * @param  gpuDst_indices
 *         The identifier of the destination GPU-buffer containing the
 *         indices of the remaining pulses.
 * @param  gpuSrc_values
 *         The identifier of the source GPU-buffer containing the values
 *         to analyse.
 * @param  minValue
 *         The minimal value of the interval.
 * @param  minValue
 *         The maximal value of the interval.
 * @return The number of remaining pulses (-1 on failure).
 */
public
int
filter1D(int gpuSrc_indices,
         int pulseCount,
         int gpuDst_indices,
         int gpuSrc_values,
         float minValue,
         float maxValue)
  {
  int remaining=_filter1D(_ntv,gpuSrc_indices,pulseCount,gpuDst_indices,
                          gpuSrc_values,minValue,maxValue);
  if(_errorify&&(remaining==-1))
    {
    throw new Error();
    }
  return remaining;
  }

/**
 * Keep only the pulses which float (A,B) values are in the specified
 * interval.
 * @param  gpuSrc_indices
 *         The identifier of the source GPU-buffer containing the indices
 *         of the pulses to analyse.
 * @param  pulseCount
 *         The number of pulses to consider.
 * @param  gpuDst_indices
 *         The identifier of the destination GPU-buffer containing the
 *         indices of the remaining pulses.
 * @param  gpuSrc_valuesA
 *         The identifier of the source GPU-buffer containing the A values
 *         to analyse.
 * @param  minValueA
 *         The minimal value of the A interval.
 * @param  minValueA
 *         The maximal value of the A interval.
 * @param  gpuSrc_valuesB
 *         The identifier of the source GPU-buffer containing the B values
 *         to analyse.
 * @param  minValueB
 *         The minimal value of the B interval.
 * @param  minValueB
 *         The maximal value of the B interval.
 * @return The number of remaining pulses (-1 on failure).
 */
public
int
filter2D(int gpuSrc_indices,
         int pulseCount,
         int gpuDst_indices,
         int gpuSrc_valuesA,
         float minValueA,
         float maxValueA,
         int gpuSrc_valuesB,
         float minValueB,
         float maxValueB)
  {
  int remaining=_filter2D(_ntv,gpuSrc_indices,pulseCount,gpuDst_indices,
                          gpuSrc_valuesA,minValueA,maxValueA,
                          gpuSrc_valuesB,minValueB,maxValueB);
  if(_errorify&&(remaining==-1))
    {
    throw new Error();
    }
  return remaining;
  }

/**
 * Keep only the pulses which float (A,B,C) values are in the specified
 * interval.
 * @param  gpuSrc_indices
 *         The identifier of the source GPU-buffer containing the indices
 *         of the pulses to analyse.
 * @param  pulseCount
 *         The number of pulses to consider.
 * @param  gpuDst_indices
 *         The identifier of the destination GPU-buffer containing the
 *         indices of the remaining pulses.
 * @param  gpuSrc_valuesA
 *         The identifier of the source GPU-buffer containing the A values
 *         to analyse.
 * @param  minValueA
 *         The minimal value of the A interval.
 * @param  minValueA
 *         The maximal value of the A interval.
 * @param  gpuSrc_valuesB
 *         The identifier of the source GPU-buffer containing the B values
 *         to analyse.
 * @param  minValueB
 *         The minimal value of the B interval.
 * @param  minValueB
 *         The maximal value of the B interval.
 * @param  gpuSrc_valuesC
 *         The identifier of the source GPU-buffer containing the C values
 *         to analyse.
 * @param  minValueC
 *         The minimal value of the C interval.
 * @param  minValueC
 *         The maximal value of the C interval.
 * @return The number of remaining pulses (-1 on failure).
 */
public
int
filter3D(int gpuSrc_indices,
         int pulseCount,
         int gpuDst_indices,
         int gpuSrc_valuesA,
         float minValueA,
         float maxValueA,
         int gpuSrc_valuesB,
         float minValueB,
         float maxValueB,
         int gpuSrc_valuesC,
         float minValueC,
         float maxValueC)
  {
  int remaining=_filter3D(_ntv,gpuSrc_indices,pulseCount,gpuDst_indices,
                          gpuSrc_valuesA,minValueA,maxValueA,
                          gpuSrc_valuesB,minValueB,maxValueB,
                          gpuSrc_valuesC,minValueC,maxValueC);
  if(_errorify&&(remaining==-1))
    {
    throw new Error();
    }
  return remaining;
  }

/**
 * Keep only the pulses which float (A,B,C,D) values are in the specified
 * interval.
 * @param  gpuSrc_indices
 *         The identifier of the source GPU-buffer containing the indices
 *         of the pulses to analyse.
 * @param  pulseCount
 *         The number of pulses to consider.
 * @param  gpuDst_indices
 *         The identifier of the destination GPU-buffer containing the
 *         indices of the remaining pulses.
 * @param  gpuSrc_valuesA
 *         The identifier of the source GPU-buffer containing the A values
 *         to analyse.
 * @param  minValueA
 *         The minimal value of the A interval.
 * @param  minValueA
 *         The maximal value of the A interval.
 * @param  gpuSrc_valuesB
 *         The identifier of the source GPU-buffer containing the B values
 *         to analyse.
 * @param  minValueB
 *         The minimal value of the B interval.
 * @param  minValueB
 *         The maximal value of the B interval.
 * @param  gpuSrc_valuesC
 *         The identifier of the source GPU-buffer containing the C values
 *         to analyse.
 * @param  minValueC
 *         The minimal value of the C interval.
 * @param  minValueC
 *         The maximal value of the C interval.
 * @param  gpuSrc_valuesD
 *         The identifier of the source GPU-buffer containing the D values
 *         to analyse.
 * @param  minValueD
 *         The minimal value of the D interval.
 * @param  minValueD
 *         The maximal value of the D interval.
 * @return The number of remaining pulses (-1 on failure).
 */
public
int
filter4D(int gpuSrc_indices,
         int pulseCount,
         int gpuDst_indices,
         int gpuSrc_valuesA,
         float minValueA,
         float maxValueA,
         int gpuSrc_valuesB,
         float minValueB,
         float maxValueB,
         int gpuSrc_valuesC,
         float minValueC,
         float maxValueC,
         int gpuSrc_valuesD,
         float minValueD,
         float maxValueD)
  {
  int remaining=_filter4D(_ntv,gpuSrc_indices,pulseCount,gpuDst_indices,
                          gpuSrc_valuesA,minValueA,maxValueA,
                          gpuSrc_valuesB,minValueB,maxValueB,
                          gpuSrc_valuesC,minValueC,maxValueC,
                          gpuSrc_valuesD,minValueD,maxValueD);
  if(_errorify&&(remaining==-1))
    {
    throw new Error();
    }
  return remaining;
  }

/**
 * Builds up a histogram of the differences (DTOA) between successives
 * time-of-arrival (TOA) values.
 * These TOA differences are computed up to the DTOA_SKIP_COUNT
 * built-in constant.
 * The destination host-buffers are initialy reset to zero.
 * @param  gpuSrc_indices
 *         The identifier of the source GPU-buffer containing the indices
 *         of the pulses to analyse.
 * @param  pulseCount
 *         The number of pulses to consider.
 * @param  gpuSrc_toa64
 *         The identifier of the source GPU-buffer containing the TOA values
 *         (64-bit real) to analyse.
 * @param  hostDst_counts
 *         The destination host-buffer to store the DTOA counts of each bin.
 * @param  hostDst_sums
 *         The destination host-buffer to store the DTOA sums of each bin.
 * @param  binCount
 *         The number of bins of the histogram.
 * @param  minDtoa
 *         The lower bound (strictly greater than 0.0f) for a DTOA to be
 *         inserted in the histogram.
 * @param  dtoaRatio
 *         The multiplicative factor (strictly greater than 1.0f) which
 *         determines the logarithmic width of each bin.
 * @return Success.
 */
public
boolean
dtoaHistogram(int gpuSrc_indices,
              int pulseCount,
              int gpuSrc_toa64,
              IntBuffer hostDst_counts,
              FloatBuffer hostDst_sums,
              int binCount,
              float minDtoa,
              float dtoaRatio)
  {
  boolean success=_dtoaHistogram(_ntv,gpuSrc_indices,pulseCount,gpuSrc_toa64,
                                 hostDst_counts,hostDst_sums,
                                 binCount,minDtoa,dtoaRatio);
  if(_errorify&&!success)
    {
    throw new Error();
    }
  return success;
  }

/**
 * Find sequences of pulses which DTOA match the given PRI and mark them.
 * @param  gpuSrc_indices
 *         The identifier of the source GPU-buffer containing the indices
 *         of the pulses to analyse.
 * @param  pulseCount
 *         The number of pulses to consider.
 * @param  gpuSrc_toa64
 *         The identifier of the source GPU-buffer containing the TOA values
 *         (64-bit real) to analyse.
 * @param  gpu_marks
 *         The identifier of the GPU-buffer to store the mark of each pulse.
 * @param  mark
 *         The mark to assign to each matching pulse.
 * @param  estimatedPri
 *         The previously estimated PRI to be looked for.
 * @param  priRatioThreshold
 *         The precision factor (greater than 1.0f) used to compare the DTOA
 *         in a sequence with the expected PRI.
 * @param  minMatchCount
 *         The minimal number of matching DTOA to detect a sequence.
 * @return The actual PRI of the marked pulses or 0.0f (-1.0f on failure).
 */
public
float
markPulses(int gpuSrc_indices,
           int pulseCount,
           int gpuSrc_toa64,
           int gpu_marks,
           int mark,
           float estimatedPri,
           float priRatioThreshold,
           int minMatchCount)
  {
  float pri=_markPulses(_ntv,gpuSrc_indices,pulseCount,gpuSrc_toa64,
                        gpu_marks,mark,estimatedPri,
                        priRatioThreshold,minMatchCount);
  if(_errorify&&(pri<0.0f))
    {
    throw new Error();
    }
  return pri;
  }

/**
 * Keep only the pulses with no mark (<0).
 * @param  gpuSrc_indices
 *         The identifier of the source GPU-buffer containing the indices
 *         of the pulses to analyse.
 * @param  pulseCount
 *         The number of pulses to consider.
 * @param  gpuDst_indices
 *         The identifier of the destination GPU-buffer containing the
 *         indices of the remaining pulses.
 * @param  gpuSrc_marks
 *         The identifier of the source GPU-buffer containing the marks
 *         to analyse.
 * @return The number of remaining pulses (-1 on failure).
 */
public
int
filterUnmarked(int gpuSrc_indices,
               int pulseCount,
               int gpuDst_indices,
               int gpuSrc_marks)
  {
  int remaining=_filterUnmarked(_ntv,gpuSrc_indices,pulseCount,
                                gpuDst_indices,gpuSrc_marks);
  if(_errorify&&(remaining==-1))
    {
    throw new Error();
    }
  return remaining;
  }

//----------------------------------------------------------------------------

/**
 * Deletes a GPU processing environment that will not be used anymore.
 */
@Override
protected
void
finalize() throws Throwable
  {
  _delete(_ntv);
  }

protected final boolean _errorify;
protected final long _ntv;

protected static native
long
_new(long sharedNtv);

protected static native
void
_delete(long ntv);

protected static native
boolean
_properties(long ntv,
            Properties out_properties);

protected static native
float
_elapsed(long ntv);

protected static native
boolean
_flush(long ntv,
       boolean wait);

protected static native
int
_gpuAlloc(long ntv,
          long capacity);

protected static native
boolean
_gpuFree(long ntv,
         int gpuBuffer);

//
// FIXME: missing GpuElint_beginMapping()
//

//
// FIXME: missing GpuElint_endMapping()
//

protected static native
boolean
_hostToGpu(long ntv,
           int gpuDst,
           Buffer hostSrc,
           long dstOffset,
           long srcOffset,
           long size);

protected static native
boolean
_gpuToHost(long ntv,
           Buffer hostDst,
           int gpuSrc,
           long dstOffset,
           long srcOffset,
           long size);

protected static native
boolean
_gpuToGpu(long ntv,
          int gpuDst,
          int gpuSrc,
          long dstOffset,
          long srcOffset,
          long size);

protected static native
int
_createMarker(long ntv);

protected static native
boolean
_destroyMarker(long ntv,
               int marker);

protected static native
boolean
_setMarker(long ntv,
           int marker);

protected static native
int
_testMarker(long ntv,
            int marker);

protected static native
float
_markerInterval(long ntv,
                int firstMarker,
                int secondMarker);

protected static native
boolean
_zeroFill(long ntv,
          int gpuDst,
          int begin,
          int end);

protected static native
boolean
_countFill(long ntv,
           int gpuDst,
           int begin,
           int end,
           int countStart,
           int countIncrement);

protected static native
int
_modes1D(long ntv,
         IntBuffer hostDst_modes,
         int gpuSrc_counts,
         float cutoffRatio,
         int binCount,
         boolean cyclic);

protected static native
int
_modes2D(long ntv,
         IntBuffer hostDst_modes,
         int gpuSrc_counts,
         float cutoffRatio,
         int binCountA,
         boolean cyclicA,
         int binCountB,
         boolean cyclicB);

protected static native
int
_modes3D(long ntv,
         IntBuffer hostDst_modes,
         int gpuSrc_counts,
         float cutoffRatio,
         int binCountA,
         boolean cyclicA,
         int binCountB,
         boolean cyclicB,
         int binCountC,
         boolean cyclicC);

protected static native
int
_modes4D(long ntv,
         IntBuffer hostDst_modes,
         int gpuSrc_counts,
         float cutoffRatio,
         int binCountA,
         boolean cyclicA,
         int binCountB,
         boolean cyclicB,
         int binCountC,
         boolean cyclicC,
         int binCountD,
         boolean cyclicD);

protected static native
float
_estimateMajorPri(long ntv,
                  IntBuffer hostSrc_counts,
                  FloatBuffer hostSrc_sums,
                  int binCount,
                  float minDtoa,
                  float dtoaRatio,
                  float exclusivePriLowerLimit,
                  float minDtoaSum);

protected static native
boolean
_minMax(long ntv,
        int gpuSrc_indices,
        int pulseCount,
        int gpuSrc_values,
        MinMaxResult out_result);

protected static native
boolean
_histogram1D(long ntv,
             int gpuSrc_indices,
             int pulseCount,
             int gpuDst_counts,
             int gpuSrc_values,
             float minValue,
             float maxValue,
             int binCount);

protected static native
boolean
_histogram2D(long ntv,
             int gpuSrc_indices,
             int pulseCount,
             int gpuDst_counts,
             int gpuSrc_valuesA,
             float minValueA,
             float maxValueA,
             int binCountA,
             int gpuSrc_valuesB,
             float minValueB,
             float maxValueB,
             int binCountB);

protected static native
boolean
_histogram3D(long ntv,
             int gpuSrc_indices,
             int pulseCount,
             int gpuDst_counts,
             int gpuSrc_valuesA,
             float minValueA,
             float maxValueA,
             int binCountA,
             int gpuSrc_valuesB,
             float minValueB,
             float maxValueB,
             int binCountB,
             int gpuSrc_valuesC,
             float minValueC,
             float maxValueC,
             int binCountC);

protected static native
boolean
_histogram4D(long ntv,
             int gpuSrc_indices,
             int pulseCount,
             int gpuDst_counts,
             int gpuSrc_valuesA,
             float minValueA,
             float maxValueA,
             int binCountA,
             int gpuSrc_valuesB,
             float minValueB,
             float maxValueB,
             int binCountB,
             int gpuSrc_valuesC,
             float minValueC,
             float maxValueC,
             int binCountC,
             int gpuSrc_valuesD,
             float minValueD,
             float maxValueD,
             int binCountD);

protected static native
int
_filter1D(long ntv,
          int gpuSrc_indices,
          int pulseCount,
          int gpuDst_indices,
          int gpuSrc_values,
          float minValue,
          float maxValue);

protected static native
int
_filter2D(long ntv,
          int gpuSrc_indices,
          int pulseCount,
          int gpuDst_indices,
          int gpuSrc_valuesA,
          float minValueA,
          float maxValueA,
          int gpuSrc_valuesB,
          float minValueB,
          float maxValueB);

protected static native
int
_filter3D(long ntv,
          int gpuSrc_indices,
          int pulseCount,
          int gpuDst_indices,
          int gpuSrc_valuesA,
          float minValueA,
          float maxValueA,
          int gpuSrc_valuesB,
          float minValueB,
          float maxValueB,
          int gpuSrc_valuesC,
          float minValueC,
          float maxValueC);

protected static native
int
_filter4D(long ntv,
          int gpuSrc_indices,
          int pulseCount,
          int gpuDst_indices,
          int gpuSrc_valuesA,
          float minValueA,
          float maxValueA,
          int gpuSrc_valuesB,
          float minValueB,
          float maxValueB,
          int gpuSrc_valuesC,
          float minValueC,
          float maxValueC,
          int gpuSrc_valuesD,
          float minValueD,
          float maxValueD);

protected static native
boolean
_dtoaHistogram(long ntv,
               int gpuSrc_indices,
               int pulseCount,
               int gpuSrc_toa64,
               IntBuffer hostDst_counts,
               FloatBuffer hostDst_sums,
               int binCount,
               float minDtoa,
               float dtoaRatio);

protected static native
float
_markPulses(long ntv,
            int gpuSrc_indices,
            int pulseCount,
            int gpuSrc_toa64,
            int gpu_marks,
            int mark,
            float estimatedPri,
            float priRatioThreshold,
            int minMatchCount);

protected static native
int
_filterUnmarked(long ntv,
                int gpuSrc_indices,
                int pulseCount,
                int gpuDst_indices,
                int gpuSrc_marks);

static
  {
  System.loadLibrary("GpuElint");
  }
}

//----------------------------------------------------------------------------
