# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (c) 2015-2016 Fabrice HARROUET (ENIB)
#
# This file is part of GpuElint.
#   http://www.enib.fr/~harrouet/gpuelint.html
#
# GpuElint is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GpuElint is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with GpuElint.  If not, see <http://www.gnu.org/licenses/>.
#-----------------------------------------------------------------------------

import sys
import os
import platform
import traceback
import ctypes

class GpuElint(object):

  class Error(Exception):
    def __init__(self,*args,**kw):
      Exception.__init__(self,*args,**kw)

  class Properties(ctypes.Structure):
    _fields_=[('DTOA_SKIP_COUNT',                 ctypes.c_int),
              ('MAX_MATCH_COUNT',                 ctypes.c_int),
              ('glVersionMajor',                  ctypes.c_int),
              ('glVersionMinor',                  ctypes.c_int),
              ('glslVersion',                     ctypes.c_int),
              ('maxWorkGroupCountX',              ctypes.c_int),
              ('maxWorkGroupCountY',              ctypes.c_int),
              ('maxWorkGroupCountZ',              ctypes.c_int),
              ('maxWorkGroupSizeX',               ctypes.c_int),
              ('maxWorkGroupSizeY',               ctypes.c_int),
              ('maxWorkGroupSizeZ',               ctypes.c_int),
              ('maxWorkGroupInvocations',         ctypes.c_int),
              ('maxUniformBlocks',                ctypes.c_int),
              ('maxTextureImageUnits',            ctypes.c_int),
              ('maxAtomicCounterBuffers',         ctypes.c_int),
              ('maxAtomicCounters',               ctypes.c_int),
              ('maxSharedMemorySize',             ctypes.c_int),
              ('maxUniformComponents',            ctypes.c_int),
              ('maxImageUniforms',                ctypes.c_int),
              ('maxCombinedUniformComponents',    ctypes.c_int),
              ('maxShaderStorageBlocks',          ctypes.c_int),
              ('warpSize',                        ctypes.c_int),
              ('warpsPerStreamingMultiprocessor', ctypes.c_int),
              ('streamingMultiprocessorCount',    ctypes.c_int),
              ('hasBallot',                       ctypes.c_bool),
              ('hasShuffle',                      ctypes.c_bool)]

  def __init__(self,shared=None,errorify=True):
    self._errorify=errorify
    sharedNtv=ctypes.c_void_p(0)
    if shared:
      sharedNtv=shared._ntv
    self._ntv=_GpuElint_new(sharedNtv)
    if self._errorify and not self._ntv:
      raise GpuElint.Error()

  def __del__(self):
    try:
      _GpuElint_delete(self._ntv)
    except:
      if traceback: # traceback may have already been destroyed
        traceback.print_exc()

  def properties(self):
    addr=_GpuElint_properties(self._ntv)
    if addr:
      return GpuElint.Properties.from_address(addr)
    else:
      if self._errorify:
        raise GpuElint.Error()
      else:
        return None

  def elapsed(self):
    result=_GpuElint_elapsed(self._ntv)
    if self._errorify and result<0.0:
      raise GpuElint.Error()
    return result

  def flush(self,wait):
    success=_GpuElint_flush(self._ntv,wait)
    if self._errorify and not success:
      raise GpuElint.Error()
    return success

  def gpuAlloc(self,capacity):
    id=_GpuElint_gpuAlloc(self._ntv,capacity)
    if self._errorify and not id:
      raise GpuElint.Error()
    return id

  def gpuFree(self,gpuBuffer):
    success=_GpuElint_gpuFree(self._ntv,gpuBuffer)
    if self._errorify and not success:
      raise GpuElint.Error()
    return success

  #
  # FIXME: missing GpuElint_beginMapping()
  #

  #
  # FIXME: missing GpuElint_endMapping()
  #

  def hostToGpu(self,gpuDst,hostSrc,dstOffset,srcOffset,size):
    success=_GpuElint_hostToGpu(self._ntv,gpuDst,hostSrc,
                                dstOffset,srcOffset,size)
    if self._errorify and not success:
      raise GpuElint.Error()
    return success

  def gpuToHost(self,hostDst,gpuSrc,dstOffset,srcOffset,size):
    success=_GpuElint_gpuToHost(self._ntv,hostDst,gpuSrc,
                                dstOffset,srcOffset,size)
    if self._errorify and not success:
      raise GpuElint.Error()
    return success

  def gpuToGpu(self,gpuDst,gpuSrc,dstOffset,srcOffset,size):
    success=_GpuElint_gpuToGpu(self._ntv,gpuDst,gpuSrc,
                               dstOffset,srcOffset,size)
    if self._errorify and not success:
      raise GpuElint.Error()
    return success

  def createMarker(self):
    marker=_GpuElint_createMarker(self._ntv)
    if self._errorify and not marker:
      raise GpuElint.Error()
    return marker

  def destroyMarker(self,marker):
    success=_GpuElint_destroyMarker(self._ntv,marker)
    if self._errorify and not success:
      raise GpuElint.Error()
    return success

  def setMarker(self,marker):
    success=_GpuElint_setMarker(self._ntv,marker)
    if self._errorify and not success:
      raise GpuElint.Error()
    return success

  def testMarker(self,marker):
    result=_GpuElint_testMarker(self._ntv,marker)
    if self._errorify and result==-1:
      raise GpuElint.Error()
    return result

  def markerInterval(self,firstMarker,secondMarker):
    interval=_GpuElint_markerInterval(self._ntv,firstMarker,secondMarker)
    if self._errorify and interval<0.0:
      raise GpuElint.Error()
    return interval

  def zeroFill(self,gpuDst,begin,end):
    success=_GpuElint_zeroFill(self._ntv,gpuDst,begin,end)
    if self._errorify and not success:
      raise GpuElint.Error()
    return success

  def countFill(self,gpuDst,begin,end,countStart,countIncrement):
    success=_GpuElint_countFill(self._ntv,gpuDst,begin,end,
                                countStart,countIncrement)
    if self._errorify and not success:
      raise GpuElint.Error()
    return success

  class HistoMode1D(ctypes.Structure):
    _fields_=[('maxCount', ctypes.c_int),
              ('weight',   ctypes.c_int),
              ('firstBin', ctypes.c_int),
              ('lastBin',  ctypes.c_int)]

  def modes1D(self,hostDst_modes,gpuSrc_counts,cutoffRatio,
              binCount,cyclic):
    modeCount=_GpuElint_modes1D(self._ntv,ctypes.byref(hostDst_modes),
                                len(hostDst_modes),gpuSrc_counts,cutoffRatio,
                                binCount,cyclic)
    if self._errorify and modeCount==-1:
      raise GpuElint.Error()
    return modeCount

  class HistoMode2D(ctypes.Structure):
    _fields_=[('maxCount',  ctypes.c_int),
              ('weight',    ctypes.c_int),
              ('firstBinA', ctypes.c_int),
              ('lastBinA',  ctypes.c_int),
              ('firstBinB', ctypes.c_int),
              ('lastBinB',  ctypes.c_int)]

  def modes2D(self,hostDst_modes,gpuSrc_counts,cutoffRatio,
              binCountA,cyclicA,binCountB,cyclicB):
    modeCount=_GpuElint_modes2D(self._ntv,ctypes.byref(hostDst_modes),
                                len(hostDst_modes),gpuSrc_counts,cutoffRatio,
                                binCountA,cyclicA,binCountB,cyclicB)
    if self._errorify and modeCount==-1:
      raise GpuElint.Error()
    return modeCount

  class HistoMode3D(ctypes.Structure):
    _fields_=[('maxCount',  ctypes.c_int),
              ('weight',    ctypes.c_int),
              ('firstBinA', ctypes.c_int),
              ('lastBinA',  ctypes.c_int),
              ('firstBinB', ctypes.c_int),
              ('lastBinB',  ctypes.c_int),
              ('firstBinC', ctypes.c_int),
              ('lastBinC',  ctypes.c_int)]

  def modes3D(self,hostDst_modes,gpuSrc_counts,cutoffRatio,
              binCountA,cyclicA,binCountB,cyclicB,
              binCountC,cyclicC):
    modeCount=_GpuElint_modes3D(self._ntv,ctypes.byref(hostDst_modes),
                                len(hostDst_modes),gpuSrc_counts,cutoffRatio,
                                binCountA,cyclicA,binCountB,cyclicB,
                                binCountC,cyclicC)
    if self._errorify and modeCount==-1:
      raise GpuElint.Error()
    return modeCount

  class HistoMode4D(ctypes.Structure):
    _fields_=[('maxCount',  ctypes.c_int),
              ('weight',    ctypes.c_int),
              ('firstBinA', ctypes.c_int),
              ('lastBinA',  ctypes.c_int),
              ('firstBinB', ctypes.c_int),
              ('lastBinB',  ctypes.c_int),
              ('firstBinC', ctypes.c_int),
              ('lastBinC',  ctypes.c_int),
              ('firstBinD', ctypes.c_int),
              ('lastBinD',  ctypes.c_int)]

  def modes4D(self,hostDst_modes,gpuSrc_counts,cutoffRatio,
              binCountA,cyclicA,binCountB,cyclicB,
              binCountC,cyclicC,binCountD,cyclicD):
    modeCount=_GpuElint_modes4D(self._ntv,ctypes.byref(hostDst_modes),
                                len(hostDst_modes),gpuSrc_counts,cutoffRatio,
                                binCountA,cyclicA,binCountB,cyclicB,
                                binCountC,cyclicC,binCountD,cyclicD)
    if self._errorify and modeCount==-1:
      raise GpuElint.Error()
    return modeCount

  def estimateMajorPri(self,hostSrc_counts,hostSrc_sums,binCount,
                       minDtoa,dtoaRatio,exclusivePriLowerLimit,minDtoaSum):
    return _GpuElint_estimateMajorPri(self._ntv,
                                      hostSrc_counts,hostSrc_sums,
                                      binCount,minDtoa,dtoaRatio,
                                      exclusivePriLowerLimit,minDtoaSum)

  def minMax(self,gpuSrc_indices,pulseCount,gpuSrc_values):
    minValue=ctypes.c_float(sys.float_info.max)
    minId=ctypes.c_int(-1)
    maxValue=ctypes.c_float(-minValue.value)
    maxId=ctypes.c_int(-1)
    success=_GpuElint_ID_minMax(self._ntv,gpuSrc_indices,pulseCount,
                                gpuSrc_values,
                                ctypes.byref(minValue),ctypes.byref(minId),
                                ctypes.byref(maxValue),ctypes.byref(maxId))
    if not success:
      if self._errorify:
        raise GpuElint.Error()
      return None
    return (minValue.value,minId.value,
            maxValue.value,maxId.value)

  def histogram1D(self,gpuSrc_indices,pulseCount,gpuDst_counts,
                  gpuSrc_values,minValue,maxValue,binCount):
    success=_GpuElint_ID_histogram1D(self._ntv,gpuSrc_indices,pulseCount,
                                     gpuDst_counts,
                                     gpuSrc_values,
                                     minValue,maxValue,binCount)
    if not success:
      if self._errorify:
        raise GpuElint.Error()
      return None
    return success

  def histogram2D(self,gpuSrc_indices,pulseCount,gpuDst_counts,
                  gpuSrc_valuesA,minValueA,maxValueA,binCountA,
                  gpuSrc_valuesB,minValueB,maxValueB,binCountB):
    success=_GpuElint_ID_histogram2D(self._ntv,gpuSrc_indices,pulseCount,
                                     gpuDst_counts,
                                     gpuSrc_valuesA,
                                     minValueA,maxValueA,binCountA,
                                     gpuSrc_valuesB,
                                     minValueB,maxValueB,binCountB)
    if self._errorify and not success:
      raise GpuElint.Error()
    return success

  def histogram3D(self,gpuSrc_indices,pulseCount,gpuDst_counts,
                  gpuSrc_valuesA,minValueA,maxValueA,binCountA,
                  gpuSrc_valuesB,minValueB,maxValueB,binCountB,
                  gpuSrc_valuesC,minValueC,maxValueC,binCountC):
    success=_GpuElint_ID_histogram3D(self._ntv,gpuSrc_indices,pulseCount,
                                     gpuDst_counts,
                                     gpuSrc_valuesA,
                                     minValueA,maxValueA,binCountA,
                                     gpuSrc_valuesB,
                                     minValueB,maxValueB,binCountB,
                                     gpuSrc_valuesC,
                                     minValueC,maxValueC,binCountC)
    if self._errorify and not success:
      raise GpuElint.Error()
    return success

  def histogram4D(self,gpuSrc_indices,pulseCount,gpuDst_counts,
                  gpuSrc_valuesA,minValueA,maxValueA,binCountA,
                  gpuSrc_valuesB,minValueB,maxValueB,binCountB,
                  gpuSrc_valuesC,minValueC,maxValueC,binCountC,
                  gpuSrc_valuesD,minValueD,maxValueD,binCountD):
    success=_GpuElint_ID_histogram4D(self._ntv,gpuSrc_indices,pulseCount,
                                     gpuDst_counts,
                                     gpuSrc_valuesA,
                                     minValueA,maxValueA,binCountA,
                                     gpuSrc_valuesB,
                                     minValueB,maxValueB,binCountB,
                                     gpuSrc_valuesC,
                                     minValueC,maxValueC,binCountC,
                                     gpuSrc_valuesD,
                                     minValueD,maxValueD,binCountD)
    if self._errorify and not success:
      raise GpuElint.Error()
    return success

  def filter1D(self,gpuSrc_indices,pulseCount,gpuDst_indices,
               gpuSrc_values,minValue,maxValue):
    remaining=_GpuElint_ID_filter1D(self._ntv,gpuSrc_indices,pulseCount,
                                    gpuDst_indices,
                                    gpuSrc_values,minValue,maxValue)
    if self._errorify and remaining==-1:
      raise GpuElint.Error()
    return remaining

  def filter2D(self,gpuSrc_indices,pulseCount,gpuDst_indices,
               gpuSrc_valuesA,minValueA,maxValueA,
               gpuSrc_valuesB,minValueB,maxValueB):
    remaining=_GpuElint_ID_filter2D(self._ntv,gpuSrc_indices,pulseCount,
                                    gpuDst_indices,
                                    gpuSrc_valuesA,minValueA,maxValueA,
                                    gpuSrc_valuesB,minValueB,maxValueB)
    if self._errorify and remaining==-1:
      raise GpuElint.Error()
    return remaining

  def filter3D(self,gpuSrc_indices,pulseCount,gpuDst_indices,
               gpuSrc_valuesA,minValueA,maxValueA,
               gpuSrc_valuesB,minValueB,maxValueB,
               gpuSrc_valuesC,minValueC,maxValueC):
    remaining=_GpuElint_ID_filter3D(self._ntv,gpuSrc_indices,pulseCount,
                                    gpuDst_indices,
                                    gpuSrc_valuesA,minValueA,maxValueA,
                                    gpuSrc_valuesB,minValueB,maxValueB,
                                    gpuSrc_valuesC,minValueC,maxValueC)
    if self._errorify and remaining==-1:
      raise GpuElint.Error()
    return remaining

  def filter4D(self,gpuSrc_indices,pulseCount,gpuDst_indices,
               gpuSrc_valuesA,minValueA,maxValueA,
               gpuSrc_valuesB,minValueB,maxValueB,
               gpuSrc_valuesC,minValueC,maxValueC,
               gpuSrc_valuesD,minValueD,maxValueD):
    remaining=_GpuElint_ID_filter4D(self._ntv,gpuSrc_indices,pulseCount,
                                    gpuDst_indices,
                                    gpuSrc_valuesA,minValueA,maxValueA,
                                    gpuSrc_valuesB,minValueB,maxValueB,
                                    gpuSrc_valuesC,minValueC,maxValueC,
                                    gpuSrc_valuesD,minValueD,maxValueD)
    if self._errorify and remaining==-1:
      raise GpuElint.Error()
    return remaining

  def dtoaHistogram(self,gpuSrc_indices,pulseCount,gpuSrc_toa64,
                    hostDst_counts,hostDst_sums,binCount,minDtoa,dtoaRatio):
    success=_GpuElint_ID_dtoaHistogram(self._ntv,
                                       gpuSrc_indices,pulseCount,gpuSrc_toa64,
                                       hostDst_counts,hostDst_sums,
                                       binCount,minDtoa,dtoaRatio)
    if self._errorify and not success:
      raise GpuElint.Error()
    return success

  def markPulses(self,gpuSrc_indices,pulseCount,gpuSrc_toa64,gpu_marks,mark,
                 estimatedPri,priRatioThreshold,minMatchCount):
    pri=_GpuElint_ID_markPulses(self._ntv,gpuSrc_indices,pulseCount,
                                gpuSrc_toa64,gpu_marks,mark,
                                estimatedPri,priRatioThreshold,minMatchCount)
    if self._errorify and pri<0.0:
      raise GpuElint.Error()
    return pri

  def filterUnmarked(self,gpuSrc_indices,pulseCount,gpuDst_indices,
                     gpuSrc_marks):
    remaining=_GpuElint_ID_filterUnmarked(self._ntv,gpuSrc_indices,pulseCount,
                                          gpuDst_indices,gpuSrc_marks)
    if self._errorify and remaining==-1:
      raise GpuElint.Error()
    return remaining

try:
  _GpuElintDirName=os.path.dirname(sys.modules[__name__].__file__)
  if platform.system().startswith('Windows'):
    _GpuElintLibName='GpuElint.dll'
  else:
    _GpuElintLibName='libGpuElint.so'
  _GpuElintLib=ctypes.CDLL(os.path.join(_GpuElintDirName,_GpuElintLibName))
  #
  def _loadNativeFunction(name,restype,argtypes):
    name='GpuElint_'+name
    ntv=_GpuElintLib[name]
    ntv.restype=restype
    ntv.argtypes=argtypes
    setattr(sys.modules[__name__],'_'+name,ntv)
  #
  _loadNativeFunction('new',
                      ctypes.c_void_p,
                      [ctypes.c_void_p])
  #
  _loadNativeFunction('delete',
                      None,
                      [ctypes.c_void_p])
  #
  _loadNativeFunction('properties',
                      ctypes.c_void_p,
                      [ctypes.c_void_p])
  #
  _loadNativeFunction('elapsed',
                      ctypes.c_float,
                      [ctypes.c_void_p])
  #
  _loadNativeFunction('flush',
                      ctypes.c_bool,
                      [ctypes.c_void_p,
                       ctypes.c_bool])
  #
  _loadNativeFunction('gpuAlloc',
                      ctypes.c_int,
                      [ctypes.c_void_p,
                       ctypes.c_ssize_t])
  #
  _loadNativeFunction('gpuFree',
                      ctypes.c_bool,
                      [ctypes.c_void_p,
                       ctypes.c_int])
  #
  _loadNativeFunction('beginMapping',
                      ctypes.c_void_p,
                      [ctypes.c_void_p,
                       ctypes.c_int,
                       ctypes.c_ssize_t,
                       ctypes.c_ssize_t,
                       ctypes.c_bool,
                       ctypes.c_bool])
  #
  _loadNativeFunction('endMapping',
                      ctypes.c_bool,
                      [ctypes.c_void_p,
                       ctypes.c_int])
  #
  _loadNativeFunction('hostToGpu',
                      ctypes.c_bool,
                      [ctypes.c_void_p,
                       ctypes.c_int,
                       ctypes.c_void_p,
                       ctypes.c_ssize_t,
                       ctypes.c_ssize_t,
                       ctypes.c_ssize_t])
  #
  _loadNativeFunction('gpuToHost',
                      ctypes.c_bool,
                      [ctypes.c_void_p,
                       ctypes.c_void_p,
                       ctypes.c_int,
                       ctypes.c_ssize_t,
                       ctypes.c_ssize_t,
                       ctypes.c_ssize_t])
  #
  _loadNativeFunction('gpuToGpu',
                      ctypes.c_bool,
                      [ctypes.c_void_p,
                       ctypes.c_int,
                       ctypes.c_int,
                       ctypes.c_ssize_t,
                       ctypes.c_ssize_t,
                       ctypes.c_ssize_t])
  #
  _loadNativeFunction('createMarker',
                      ctypes.c_int,
                      [ctypes.c_void_p])
  #
  _loadNativeFunction('destroyMarker',
                      ctypes.c_bool,
                      [ctypes.c_void_p,
                       ctypes.c_int])
  #
  _loadNativeFunction('setMarker',
                      ctypes.c_bool,
                      [ctypes.c_void_p,
                       ctypes.c_int])
  #
  _loadNativeFunction('testMarker',
                      ctypes.c_int,
                      [ctypes.c_void_p,
                       ctypes.c_int])
  #
  _loadNativeFunction('markerInterval',
                      ctypes.c_float,
                      [ctypes.c_void_p,
                       ctypes.c_int,
                       ctypes.c_int])
  #
  _loadNativeFunction('zeroFill',
                      ctypes.c_bool,
                      [ctypes.c_void_p,
                       ctypes.c_int,
                       ctypes.c_int,
                       ctypes.c_int])
  #
  _loadNativeFunction('countFill',
                      ctypes.c_bool,
                      [ctypes.c_void_p,
                       ctypes.c_int,
                       ctypes.c_int,
                       ctypes.c_int,
                       ctypes.c_int,
                       ctypes.c_int])
  #
  _loadNativeFunction('modes1D',
                      ctypes.c_int,
                      [ctypes.c_void_p,
                       ctypes.c_void_p,
                       ctypes.c_int,
                       ctypes.c_int,
                       ctypes.c_float,
                       ctypes.c_int,
                       ctypes.c_bool])
  #
  _loadNativeFunction('modes2D',
                      ctypes.c_int,
                      [ctypes.c_void_p,
                       ctypes.c_void_p,
                       ctypes.c_int,
                       ctypes.c_int,
                       ctypes.c_float,
                       ctypes.c_int,
                       ctypes.c_bool,
                       ctypes.c_int,
                       ctypes.c_bool])
  #
  _loadNativeFunction('modes3D',
                      ctypes.c_int,
                      [ctypes.c_void_p,
                       ctypes.c_void_p,
                       ctypes.c_int,
                       ctypes.c_int,
                       ctypes.c_float,
                       ctypes.c_int,
                       ctypes.c_bool,
                       ctypes.c_int,
                       ctypes.c_bool,
                       ctypes.c_int,
                       ctypes.c_bool])
  #
  _loadNativeFunction('modes4D',
                      ctypes.c_int,
                      [ctypes.c_void_p,
                       ctypes.c_void_p,
                       ctypes.c_int,
                       ctypes.c_int,
                       ctypes.c_float,
                       ctypes.c_int,
                       ctypes.c_bool,
                       ctypes.c_int,
                       ctypes.c_bool,
                       ctypes.c_int,
                       ctypes.c_bool,
                       ctypes.c_int,
                       ctypes.c_bool])
  #
  _loadNativeFunction('estimateMajorPri',
                      ctypes.c_float,
                      [ctypes.c_void_p,
                       ctypes.c_void_p,
                       ctypes.c_void_p,
                       ctypes.c_int,
                       ctypes.c_float,
                       ctypes.c_float,
                       ctypes.c_float,
                       ctypes.c_float])
  #
  _loadNativeFunction('ID_minMax',
                      ctypes.c_bool,
                      [ctypes.c_void_p,
                       ctypes.c_int,
                       ctypes.c_int,
                       ctypes.c_int,
                       ctypes.c_void_p,
                       ctypes.c_void_p,
                       ctypes.c_void_p,
                       ctypes.c_void_p])
  #
  _loadNativeFunction('ID_histogram1D',
                      ctypes.c_bool,
                      [ctypes.c_void_p,
                       ctypes.c_int,
                       ctypes.c_int,
                       ctypes.c_int,
                       ctypes.c_int,
                       ctypes.c_float,
                       ctypes.c_float,
                       ctypes.c_int])
  #
  _loadNativeFunction('ID_histogram2D',
                      ctypes.c_bool,
                      [ctypes.c_void_p,
                       ctypes.c_int,
                       ctypes.c_int,
                       ctypes.c_int,
                       ctypes.c_int,
                       ctypes.c_float,
                       ctypes.c_float,
                       ctypes.c_int,
                       ctypes.c_int,
                       ctypes.c_float,
                       ctypes.c_float,
                       ctypes.c_int])
  #
  _loadNativeFunction('ID_histogram3D',
                      ctypes.c_bool,
                      [ctypes.c_void_p,
                       ctypes.c_int,
                       ctypes.c_int,
                       ctypes.c_int,
                       ctypes.c_int,
                       ctypes.c_float,
                       ctypes.c_float,
                       ctypes.c_int,
                       ctypes.c_int,
                       ctypes.c_float,
                       ctypes.c_float,
                       ctypes.c_int,
                       ctypes.c_int,
                       ctypes.c_float,
                       ctypes.c_float,
                       ctypes.c_int])
  #
  _loadNativeFunction('ID_histogram4D',
                      ctypes.c_bool,
                      [ctypes.c_void_p,
                       ctypes.c_int,
                       ctypes.c_int,
                       ctypes.c_int,
                       ctypes.c_int,
                       ctypes.c_float,
                       ctypes.c_float,
                       ctypes.c_int,
                       ctypes.c_int,
                       ctypes.c_float,
                       ctypes.c_float,
                       ctypes.c_int,
                       ctypes.c_int,
                       ctypes.c_float,
                       ctypes.c_float,
                       ctypes.c_int,
                       ctypes.c_int,
                       ctypes.c_float,
                       ctypes.c_float,
                       ctypes.c_int])
  #
  _loadNativeFunction('ID_filter1D',
                      ctypes.c_int,
                      [ctypes.c_void_p,
                       ctypes.c_int,
                       ctypes.c_int,
                       ctypes.c_int,
                       ctypes.c_int,
                       ctypes.c_float,
                       ctypes.c_float])
  #
  _loadNativeFunction('ID_filter2D',
                      ctypes.c_int,
                      [ctypes.c_void_p,
                       ctypes.c_int,
                       ctypes.c_int,
                       ctypes.c_int,
                       ctypes.c_int,
                       ctypes.c_float,
                       ctypes.c_float,
                       ctypes.c_int,
                       ctypes.c_float,
                       ctypes.c_float])
  #
  _loadNativeFunction('ID_filter3D',
                      ctypes.c_int,
                      [ctypes.c_void_p,
                       ctypes.c_int,
                       ctypes.c_int,
                       ctypes.c_int,
                       ctypes.c_int,
                       ctypes.c_float,
                       ctypes.c_float,
                       ctypes.c_int,
                       ctypes.c_float,
                       ctypes.c_float,
                       ctypes.c_int,
                       ctypes.c_float,
                       ctypes.c_float])
  #
  _loadNativeFunction('ID_filter4D',
                      ctypes.c_int,
                      [ctypes.c_void_p,
                       ctypes.c_int,
                       ctypes.c_int,
                       ctypes.c_int,
                       ctypes.c_int,
                       ctypes.c_float,
                       ctypes.c_float,
                       ctypes.c_int,
                       ctypes.c_float,
                       ctypes.c_float,
                       ctypes.c_int,
                       ctypes.c_float,
                       ctypes.c_float,
                       ctypes.c_int,
                       ctypes.c_float,
                       ctypes.c_float])
  #
  _loadNativeFunction('ID_dtoaHistogram',
                      ctypes.c_bool,
                      [ctypes.c_void_p,
                       ctypes.c_int,
                       ctypes.c_int,
                       ctypes.c_int,
                       ctypes.c_void_p,
                       ctypes.c_void_p,
                       ctypes.c_int,
                       ctypes.c_float,
                       ctypes.c_float])
  #
  _loadNativeFunction('ID_markPulses',
                      ctypes.c_float,
                      [ctypes.c_void_p,
                       ctypes.c_int,
                       ctypes.c_int,
                       ctypes.c_int,
                       ctypes.c_int,
                       ctypes.c_int,
                       ctypes.c_float,
                       ctypes.c_float,
                       ctypes.c_int])
  #
  _loadNativeFunction('ID_filterUnmarked',
                      ctypes.c_int,
                      [ctypes.c_void_p,
                       ctypes.c_int,
                       ctypes.c_int,
                       ctypes.c_int,
                       ctypes.c_int])
except:
  traceback.print_exc()

#-----------------------------------------------------------------------------
