//----------------------------------------------------------------------------
// Copyright (c) 2015-2016 Fabrice HARROUET (ENIB)
//
// This file is part of GpuElint.
//   http://www.enib.fr/~harrouet/gpuelint.html
//
// GpuElint is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GpuElint is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with GpuElint.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "gpuElint_internals.h"

#define DECLARE_AND_CALL_SERVICE_INIT(name) \
        bool GpuService##name##_init(GpuElint *); \
        if(!GpuService##name##_init(ge)) { return false; }

#define DECLARE_AND_CALL_SERVICE_DESTROY(name) \
        void GpuService##name##_destroy(GpuElint *); \
        GpuService##name##_destroy(ge);

static
int64_t // microseconds since 1970/01/01 UTC
GpuElint_microseconds(void)
{
#if defined _WIN32
  FILETIME ft;
  GetSystemTimeAsFileTime(&ft);
  int64_t t=(((int64_t)ft.dwHighDateTime)<<32)+(int64_t)ft.dwLowDateTime;
  return (t/10LL)-11644473600000000LL; // 100ns to 1us, 1601 to 1970
#else
  struct timeval tv;
  gettimeofday(&tv,(struct timezone *)0);
  return tv.tv_sec*1000000LL+tv.tv_usec;
#endif
}

static
bool // success
GpuElint_startup(GpuElint *ge,
                 const GpuElint *shared)
{
ge->initTime=shared ? shared->initTime : GpuElint_microseconds();
// ensure native types are compatible with GLSL types
ON_ERROR(sizeof(int)!=4,"weird size for int",{ return false; });
ON_ERROR(sizeof(float)!=4,"weird size for float",{ return false; });
ON_ERROR(sizeof(double)!=8,"weird size for double",{ return false; });
if(!GpuElintContext_init(&ge->context,
                         shared ? &shared->context
                         : (const GpuElintContext *)0))
  {
  return false;
  }
ge->properties.DTOA_SKIP_COUNT=GPUELINT_DTOA_SKIP_COUNT;
ge->properties.MAX_MATCH_COUNT=GPUELINT_MAX_MATCH_COUNT;
ge->properties.glVersionMajor=CoreGl_majorVersion();
ge->properties.glVersionMinor=CoreGl_minorVersion();
ge->properties.glslVersion=CoreGl_shadingLanguageVersion();
const CoreGlComputeProperties *prop=CoreGl_computeProperties();
ON_ERROR(!prop,"no compute-shader properties",{ return false; });
#define COPY_COMPUTE_PROPERTY(name) ge->properties.name=prop->name;
COPY_COMPUTE_PROPERTY(maxWorkGroupCount[0])
COPY_COMPUTE_PROPERTY(maxWorkGroupCount[1])
COPY_COMPUTE_PROPERTY(maxWorkGroupCount[2])
COPY_COMPUTE_PROPERTY(maxWorkGroupSize[0])
COPY_COMPUTE_PROPERTY(maxWorkGroupSize[1])
COPY_COMPUTE_PROPERTY(maxWorkGroupSize[2])
COPY_COMPUTE_PROPERTY(maxWorkGroupInvocations)
COPY_COMPUTE_PROPERTY(maxUniformBlocks)
COPY_COMPUTE_PROPERTY(maxTextureImageUnits)
COPY_COMPUTE_PROPERTY(maxAtomicCounterBuffers)
COPY_COMPUTE_PROPERTY(maxAtomicCounters)
COPY_COMPUTE_PROPERTY(maxSharedMemorySize)
COPY_COMPUTE_PROPERTY(maxUniformComponents)
COPY_COMPUTE_PROPERTY(maxImageUniforms)
COPY_COMPUTE_PROPERTY(maxCombinedUniformComponents)
COPY_COMPUTE_PROPERTY(maxShaderStorageBlocks)
COPY_COMPUTE_PROPERTY(warpSize)
COPY_COMPUTE_PROPERTY(warpsPerStreamingMultiprocessor)
COPY_COMPUTE_PROPERTY(streamingMultiprocessorCount)
COPY_COMPUTE_PROPERTY(hasBallot)
COPY_COMPUTE_PROPERTY(hasShuffle)
DECLARE_AND_CALL_SERVICE_INIT(CountFill)
DECLARE_AND_CALL_SERVICE_INIT(Modes1D)
DECLARE_AND_CALL_SERVICE_INIT(Modes2D)
DECLARE_AND_CALL_SERVICE_INIT(Modes3D)
DECLARE_AND_CALL_SERVICE_INIT(Modes4D)
//
DECLARE_AND_CALL_SERVICE_INIT(_ID_Compact)
DECLARE_AND_CALL_SERVICE_INIT(_ID_MinMax)
DECLARE_AND_CALL_SERVICE_INIT(_ID_Histogram1D)
DECLARE_AND_CALL_SERVICE_INIT(_ID_Histogram2D)
DECLARE_AND_CALL_SERVICE_INIT(_ID_Histogram3D)
DECLARE_AND_CALL_SERVICE_INIT(_ID_Histogram4D)
DECLARE_AND_CALL_SERVICE_INIT(_ID_Filter1D)
DECLARE_AND_CALL_SERVICE_INIT(_ID_Filter2D)
DECLARE_AND_CALL_SERVICE_INIT(_ID_Filter3D)
DECLARE_AND_CALL_SERVICE_INIT(_ID_Filter4D)
DECLARE_AND_CALL_SERVICE_INIT(_ID_DtoaHistogram)
DECLARE_AND_CALL_SERVICE_INIT(_ID_MarkPulses)
DECLARE_AND_CALL_SERVICE_INIT(_ID_FilterUnmarked)
//
DECLARE_AND_CALL_SERVICE_INIT(_BM_MinMax)
DECLARE_AND_CALL_SERVICE_INIT(_BM_Histogram1D)
DECLARE_AND_CALL_SERVICE_INIT(_BM_Histogram2D)
DECLARE_AND_CALL_SERVICE_INIT(_BM_Histogram3D)
DECLARE_AND_CALL_SERVICE_INIT(_BM_Histogram4D)
DECLARE_AND_CALL_SERVICE_INIT(_BM_Filter1D)
DECLARE_AND_CALL_SERVICE_INIT(_BM_Filter2D)
DECLARE_AND_CALL_SERVICE_INIT(_BM_Filter3D)
DECLARE_AND_CALL_SERVICE_INIT(_BM_Filter4D)
DECLARE_AND_CALL_SERVICE_INIT(_BM_DtoaHistogram)
DECLARE_AND_CALL_SERVICE_INIT(_BM_MarkPulses)
DECLARE_AND_CALL_SERVICE_INIT(_BM_FilterUnmarked)
return true;
}

static
void
GpuElint_cleanup(GpuElint *ge)
{
DECLARE_AND_CALL_SERVICE_DESTROY(_BM_FilterUnmarked)
DECLARE_AND_CALL_SERVICE_DESTROY(_BM_MarkPulses)
DECLARE_AND_CALL_SERVICE_DESTROY(_BM_DtoaHistogram)
DECLARE_AND_CALL_SERVICE_DESTROY(_BM_Filter4D)
DECLARE_AND_CALL_SERVICE_DESTROY(_BM_Filter3D)
DECLARE_AND_CALL_SERVICE_DESTROY(_BM_Filter2D)
DECLARE_AND_CALL_SERVICE_DESTROY(_BM_Filter1D)
DECLARE_AND_CALL_SERVICE_DESTROY(_BM_Histogram4D)
DECLARE_AND_CALL_SERVICE_DESTROY(_BM_Histogram3D)
DECLARE_AND_CALL_SERVICE_DESTROY(_BM_Histogram2D)
DECLARE_AND_CALL_SERVICE_DESTROY(_BM_Histogram1D)
DECLARE_AND_CALL_SERVICE_DESTROY(_BM_MinMax)
//
DECLARE_AND_CALL_SERVICE_DESTROY(_ID_FilterUnmarked)
DECLARE_AND_CALL_SERVICE_DESTROY(_ID_MarkPulses)
DECLARE_AND_CALL_SERVICE_DESTROY(_ID_DtoaHistogram)
DECLARE_AND_CALL_SERVICE_DESTROY(_ID_Filter4D)
DECLARE_AND_CALL_SERVICE_DESTROY(_ID_Filter3D)
DECLARE_AND_CALL_SERVICE_DESTROY(_ID_Filter2D)
DECLARE_AND_CALL_SERVICE_DESTROY(_ID_Filter1D)
DECLARE_AND_CALL_SERVICE_DESTROY(_ID_Histogram4D)
DECLARE_AND_CALL_SERVICE_DESTROY(_ID_Histogram3D)
DECLARE_AND_CALL_SERVICE_DESTROY(_ID_Histogram2D)
DECLARE_AND_CALL_SERVICE_DESTROY(_ID_Histogram1D)
DECLARE_AND_CALL_SERVICE_DESTROY(_ID_MinMax)
DECLARE_AND_CALL_SERVICE_DESTROY(_ID_Compact)
//
DECLARE_AND_CALL_SERVICE_DESTROY(Modes4D)
DECLARE_AND_CALL_SERVICE_DESTROY(Modes3D)
DECLARE_AND_CALL_SERVICE_DESTROY(Modes2D)
DECLARE_AND_CALL_SERVICE_DESTROY(Modes1D)
DECLARE_AND_CALL_SERVICE_DESTROY(CountFill)
GpuElint_gpuBackend(ge,0);
GpuElintContext_destroy(&ge->context);
}

//----------------------------------------------------------------------------

int // identifier of the backend GPU-buffer (0 on failure)
GpuElint_gpuBackend(GpuElint *ge,
                    intptr_t capacity)
{
if(!capacity||(capacity>ge->gpuBackendCapacity))
  {
  if(ge->gpuBackend)
    {
    bool success=GpuElint_gpuFree(ge,ge->gpuBackend);
    ON_ERROR(!success,"GpuElint_gpuFree() failure",{});
    }
  ge->gpuBackendCapacity=capacity;
  ge->gpuBackend=0;
  if(capacity)
    {
    ge->gpuBackend=GpuElint_gpuAlloc(ge,capacity);
    ON_ERROR(!ge->gpuBackend,"GpuElint_gpuAlloc() failure",{});
    }
  }
return ge->gpuBackend;
}

int // maximal size supported by a 1D work-group
GpuElint_maxWorkGroupSize(const GpuElint *ge)
{
const int sz1=ge->properties.maxWorkGroupSize[0];
const int sz2=ge->properties.maxWorkGroupInvocations;
return sz1<sz2 ? sz1 : sz2;
}

int // maximal power-of-two size supported by a 1D work-group
GpuElint_maxPowerOfTwoWorkGroupSize(const GpuElint *ge)
{
const int maxWorkGroupSize=GpuElint_maxWorkGroupSize(ge);
int sz=ge->properties.warpSize;
while((sz<<1)<=maxWorkGroupSize) sz<<=1;
return sz;
}

int // a generaly suitable work-group size
GpuElint_chooseWorkGroupSize(const GpuElint *ge,
                             bool powerOfTwo)
{
(void)powerOfTwo; // avoid ``unused parameter'' warning
// A power-of-two work-group size is not mandatory but seems to be faster
// FIXME: this hardcoded setting gives good average performances for some
//        experiments on the actual device used during development.
//        The performances may vary for other devices and/or algorithms that
//        would require specific settings.
const int sz=GpuElint_maxPowerOfTwoWorkGroupSize(ge)/4;
const int warpSize=ge->properties.warpSize;
return sz<warpSize ? warpSize : sz;
}

int // a generaly suitable work-group count
GpuElint_chooseWorkGroupCount(const GpuElint *ge)
{
// FIXME: this hardcoded setting gives good average performances for some
//        experiments on the actual device used during development.
//        The performances may vary for other devices and/or algorithms that
//        would require specific settings.
return 8*ge->properties.streamingMultiprocessorCount;
}

void
GpuElint_appendText(GpuElint *ge,
                    char **inout_storage,
                    int *inout_size,
                    int *inout_capacity,
                    const char *format,
                    ...)
{
(void)ge; // avoid ``unused parameter'' warning
char *storage=*inout_storage;
int size=*inout_size;
int capacity=*inout_capacity;
for(;;)
  {
  const int remaining=capacity-size;
  va_list args;
  va_start(args,format);
  const int r=vsnprintf(storage+size,remaining,format,args);
  va_end(args);
  if((r>=0)&&(r<remaining))
    {
    size+=r;
    break;
    }
  else
    {
    capacity+=1024;
    storage=Memory_reAlloc(storage,capacity,char);
    }
  }
*inout_storage=storage;
*inout_size=size;
*inout_capacity=capacity;
}

void
GpuElint_appendCodePreamble(GpuElint *ge,
                            char **inout_storage,
                            int *inout_size,
                            int *inout_capacity,
                            int groupSize,
                            int groupCount)
{
if(ge->properties.hasBallot)
  {
  GpuElint_appendText(ge,inout_storage,inout_size,inout_capacity,
    " #extension GL_NV_shader_thread_group: enable \n");
  }
if(ge->properties.hasShuffle)
  {
  GpuElint_appendText(ge,inout_storage,inout_size,inout_capacity,
    " #extension GL_NV_shader_thread_shuffle: enable \n");
  }
const int warpSize=ge->properties.warpSize;
int warpShift=0;
while((1<<warpShift)<warpSize) ++warpShift;
GpuElint_appendText(ge,inout_storage,inout_size,inout_capacity,
  "                           \n"
  " // compile-time constants \n"
  " #define WARP_SIZE   %d    \n"
  " #define WARP_MASK   %d    \n"
  " #define WARP_SHIFT  %d    \n"
  " #define GROUP_SIZE  %d    \n",
  warpSize,
  warpSize-1,
  warpShift,
  groupSize);
if(!(groupSize&(groupSize-1))) // power-of-two
  {
  int groupShift=0;
  while((1<<groupShift)<groupSize) ++groupShift;
  GpuElint_appendText(ge,inout_storage,inout_size,inout_capacity,
    " #define GROUP_MASK  %d \n"
    " #define GROUP_SHIFT %d \n",
    groupSize-1,
    groupShift);
  }
if(groupCount)
  {
  GpuElint_appendText(ge,inout_storage,inout_size,inout_capacity,
    " #define GROUP_COUNT %d                       \n"
    " #define GRID_SIZE   (GROUP_COUNT*GROUP_SIZE) \n",
    groupCount);
  }
GpuElint_appendText(ge,inout_storage,inout_size,inout_capacity,
  "                                                  \n"
  " // run-time constants                            \n"
  " const int globalId=int(gl_GlobalInvocationID.x); \n"
  " const int groupId=int(gl_WorkGroupID.x);         \n"
  " const int localId=int(gl_LocalInvocationID.x);   \n"
  // gl_ThreadInWarpNV of GL_NV_shader_thread_group extension would be usable
  // but gl_WarpIDNV refers to the hardware warp (not necessarily less than
  // gl_WorkGroupSize.x/gl_WarpSizeNV)
  " const int warpId=localId>>WARP_SHIFT;            \n"
  " const int inWarpId=localId&WARP_MASK;            \n");
if(!groupCount)
  {
  GpuElint_appendText(ge,inout_storage,inout_size,inout_capacity,
  " const int groupCount=int(gl_NumWorkGroups.x); \n"
  " const int gridSize=groupCount*GROUP_SIZE;    \n");
  }
GpuElint_appendText(ge,inout_storage,inout_size,inout_capacity,
  "                                                       \n"
  " #define READ_ONLY_BUFFER(b,content)                 \\\n"
  "         layout(std430,binding=b) restrict readonly  \\\n"
  "         buffer SSBO_##b { content };                  \n"
  "                                                       \n"
  " #define WRITE_ONLY_BUFFER(b,content)                \\\n"
  "         layout(std430,binding=b) restrict writeonly \\\n"
  "         buffer SSBO_##b { content };                  \n"
  "                                                       \n"
  " #define READ_WRITE_BUFFER(b,content)                \\\n"
  "         layout(std430,binding=b) restrict           \\\n"
  "         buffer SSBO_##b { content };                  \n"
  "                                                       \n"
  " #define VOLATILE_BUFFER(b,content)                  \\\n"
  "         layout(std430,binding=b) volatile coherent  \\\n"
  "         buffer SSBO_##b { content };                  \n"
  "                                                       \n"
  " #define UNIFORM_PARAMETER(p,type)                   \\\n"
  "         layout(location=p) uniform type UNI_##p;      \n"
  "                                                       \n"
  " layout(local_size_x=GROUP_SIZE) in;                   \n"
  "                                                       \n");
}

GLuint // built GPU program (0 on failure)
GpuElint_buildProgram(GpuElint *ge,
                      const char *sourceCode)
{
(void)ge; // avoid ``unused parameter'' warning
#if 0
  printf("#########################################\n"
         "%s"
         "#########################################\n",
         sourceCode);
#endif
char *log=(char *)0;
GLuint program=CoreGl_buildComputeShader(sourceCode,true,&log);
if(log)
  {
  fprintf(stderr,"~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"
                 "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n%s"
                 "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"
                 "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n",
                 log);
  Memory_free(log);
  }
return program;
}

//----------------------------------------------------------------------------

GpuElint * // new GPU environment (0 on failure)
GpuElint_new(const GpuElint *shared)
{
GpuElint *ge=Memory_initAlloc(1,GpuElint);
if(!GpuElint_startup(ge,shared))
  {
  GpuElint_cleanup(ge);
  Memory_free(ge);
  ge=(GpuElint *)0;
  }
return ge;
}

void
GpuElint_delete(GpuElint *ge)
{
// FIXME: some execution environments destroy the original thread or
//        release the OpenGL context before destruction so we silently
//        ignore this problem but resources may leak...
if(ge)
  {
  if(!GpuElintContext_mismatch(&ge->context))
    {
    GpuElint_cleanup(ge);
    }
  Memory_free(ge);
  }
}

const GpuElintProperties * // GPU properties (0 on failure)
GpuElint_properties(const GpuElint *ge)
{
GPUELINT_API_INVOCATION_CHECK(ge,{ return (const GpuElintProperties *)0; });
return &ge->properties;
}

float // seconds elapsed since initialisation (-1.0f on failure)
GpuElint_elapsed(const GpuElint *ge)
{
GPUELINT_API_INVOCATION_CHECK(ge,{ return -1.0f; });
return (GpuElint_microseconds()-ge->initTime)*1e-6f;
}

bool // success
GpuElint_flush(GpuElint *ge,
               bool wait)
{
GPUELINT_API_INVOCATION_CHECK(ge,{ return false; });
if(wait)
  {
  glMemoryBarrier(GL_ALL_BARRIER_BITS);
  ON_GL_ERROR({ return false; });
  glFinish();
  ON_GL_ERROR({ return false; });
  }
else
  {
  glFlush();
  ON_GL_ERROR({ return false; });
  }
return true;
}

int // GPU-buffer identifier (0 on failure)
GpuElint_gpuAlloc(GpuElint *ge,
                  intptr_t capacity)
{
GPUELINT_API_INVOCATION_CHECK(ge,{ return 0; });
GLuint buffer=0;
glGenBuffers(1,&buffer);
ON_GL_ERROR({ goto undo; });
glBindBuffer(GL_SHADER_STORAGE_BUFFER,buffer);
ON_GL_ERROR({ goto undo; });
glBufferData(GL_SHADER_STORAGE_BUFFER,
             capacity,(const void *)0,
             GL_STATIC_COPY); // suitable for long term read/write on GPU
ON_GL_ERROR({ goto undo; });
return (int)buffer;
goto undo; // avoid ``unused label'' warning
undo:
if(buffer)
  {
  glDeleteBuffers(1,&buffer);
  ON_GL_ERROR({});
  }
return 0;
}

bool // success
GpuElint_gpuFree(GpuElint *ge,
                 int gpuBuffer)
{
GPUELINT_API_INVOCATION_CHECK(ge,{ return false; });
GLuint buffer=(GLuint)gpuBuffer;
glDeleteBuffers(1,&buffer);
ON_GL_ERROR({ return false; });
return true;
}

void * // mapped address (0 on failure)
GpuElint_beginMapping(GpuElint *ge,
                      int gpuBuffer,
                      intptr_t offset,
                      intptr_t size,
                      bool readAccess,
                      bool writeAccess)
{
void *addr=(void *)0;
GPUELINT_API_INVOCATION_CHECK(ge,{ return addr; });
GLuint buffer=(GLuint)gpuBuffer;
glBindBuffer(GL_SHADER_STORAGE_BUFFER,buffer);
ON_GL_ERROR({ return addr; });
GLbitfield access=0;
if(readAccess)  { access|=GL_MAP_READ_BIT;  }
if(writeAccess) { access|=GL_MAP_WRITE_BIT; }
addr=glMapBufferRange(GL_SHADER_STORAGE_BUFFER,offset,size,access);
ON_GL_ERROR({ return addr; });
return addr;
}

bool // success
GpuElint_endMapping(GpuElint *ge,
                    int gpuBuffer)
{
GPUELINT_API_INVOCATION_CHECK(ge,{ return false; });
GLuint buffer=(GLuint)gpuBuffer;
glBindBuffer(GL_SHADER_STORAGE_BUFFER,buffer);
ON_GL_ERROR({ return false; });
glUnmapBuffer(GL_SHADER_STORAGE_BUFFER);
ON_GL_ERROR({ return false; });
return true;
}

bool // success
GpuElint_hostToGpu(GpuElint *ge,
                   int gpuDst,
                   const void *hostSrc,
                   intptr_t dstOffset,
                   intptr_t srcOffset,
                   intptr_t size)
{
GPUELINT_API_INVOCATION_CHECK(ge,{ return false; });
GLuint dstBuffer=(GLuint)gpuDst;
glBindBuffer(GL_SHADER_STORAGE_BUFFER,dstBuffer);
ON_GL_ERROR({ return false; });
glBufferSubData(GL_SHADER_STORAGE_BUFFER,dstOffset,size,
                srcOffset+(uint8_t *)hostSrc);
ON_GL_ERROR({ return false; });
return true;
}

bool // success
GpuElint_gpuToHost(GpuElint *ge,
                   void *hostDst,
                   int gpuSrc,
                   intptr_t dstOffset,
                   intptr_t srcOffset,
                   intptr_t size)
{
GPUELINT_API_INVOCATION_CHECK(ge,{ return false; });
GLuint srcBuffer=(GLuint)gpuSrc;
glBindBuffer(GL_SHADER_STORAGE_BUFFER,srcBuffer);
ON_GL_ERROR({ return false; });
glGetBufferSubData(GL_SHADER_STORAGE_BUFFER,srcOffset,size,
                   dstOffset+(uint8_t *)hostDst);
ON_GL_ERROR({ return false; });
return true;
}

bool // success
GpuElint_gpuToGpu(GpuElint *ge,
                  int gpuDst,
                  int gpuSrc,
                  intptr_t dstOffset,
                  intptr_t srcOffset,
                  intptr_t size)
{
GPUELINT_API_INVOCATION_CHECK(ge,{ return false; });
GLuint dstBuffer=(GLuint)gpuDst;
glBindBuffer(GL_COPY_WRITE_BUFFER,dstBuffer);
GLuint srcBuffer=(GLuint)gpuSrc;
glBindBuffer(GL_COPY_READ_BUFFER,srcBuffer);
ON_GL_ERROR({ return false; });
glCopyBufferSubData(GL_COPY_READ_BUFFER,GL_COPY_WRITE_BUFFER,
                    srcOffset,dstOffset,size);
ON_GL_ERROR({ return false; });
return true;
}

int // GPU-time-marker identifier (0 on failure)
GpuElint_createMarker(GpuElint *ge)
{
GPUELINT_API_INVOCATION_CHECK(ge,{ return 0; });
GLuint query;
glGenQueries(1,&query);
ON_GL_ERROR({ return 0; });
return (int)query;
}

bool // success
GpuElint_destroyMarker(GpuElint *ge,
                       int marker)
{
GPUELINT_API_INVOCATION_CHECK(ge,{ return false; });
GLuint query=(GLuint)marker;
glDeleteQueries(1,&query);
ON_GL_ERROR({ return false; });
return true;
}

bool // success
GpuElint_setMarker(GpuElint *ge,
                   int marker)
{
GPUELINT_API_INVOCATION_CHECK(ge,{ return false; });
GLuint query=(GLuint)marker;
glQueryCounter(query,GL_TIMESTAMP);
ON_GL_ERROR({ return false; });
return true;
}

int // 0: marker is not reached  1: marker is reached  -1: failure
GpuElint_testMarker(GpuElint *ge,
                    int marker)
{
GPUELINT_API_INVOCATION_CHECK(ge,{ return -1; });
GLuint query=(GLuint)marker;
GLint available=0;
glGetQueryObjectiv(query,GL_QUERY_RESULT_AVAILABLE,&available);
ON_GL_ERROR({ return -1; });
return available==GL_TRUE ? 1 : 0;
}

float // seconds between first and second markers (-1.0f on failure)
GpuElint_markerInterval(GpuElint *ge,
                        int firstMarker,
                        int secondMarker)
{
GPUELINT_API_INVOCATION_CHECK(ge,{ return -1.0f; });
GLuint firstQuery=(GLuint)firstMarker;
GLuint secondQuery=(GLuint)secondMarker;
GLuint64 firstTime=0;
GLuint64 secondTime=0;
glGetQueryObjectui64v(firstQuery,GL_QUERY_RESULT,&firstTime);
ON_GL_ERROR({ return -1.0f; });
glGetQueryObjectui64v(secondQuery,GL_QUERY_RESULT,&secondTime);
ON_GL_ERROR({ return -1.0f; });
GLuint64 interval=secondTime>firstTime ? secondTime-firstTime
                                       : firstTime-secondTime;
return interval*1e-9f;
}

float // pulse-repetition-interval or 0.0f
GpuElint_estimateMajorPri(GpuElint *ge,
                          const int *hostSrc_counts,
                          const float *hostSrc_sums,
                          int binCount,
                          float minDtoa,
                          float dtoaRatio,
                          float exclusivePriLowerLimit,
                          float minDtoaSum)
{
(void)ge; // avoid ``unused parameter'' warning
const float log_minDtoa=logf(minDtoa);
const float div_log_dtoaRatio=1.0f/logf(dtoaRatio);
int bestCount=0;
float bestPri=0.0f;
for(int i=0;i<binCount;++i)
  {
  const float sum=hostSrc_sums[i];
  if(sum>=minDtoaSum)
    {
    int count=hostSrc_counts[i];
    float pri=sum/count;
    for(int skip=2;skip<=GPUELINT_DTOA_SKIP_COUNT;++skip)
      {
      const int i2=(int)floorf((logf(skip*pri)-log_minDtoa)*
                               div_log_dtoaRatio);
      if((i2<0)||(i2>=binCount)) { break; }
      const int count2=hostSrc_counts[i2];
      if(count2)
        {
        const float pri2=hostSrc_sums[i2]/skip/count2;
        count+=count2;
        pri=(pri2+pri*(skip-1))/skip;
        }
      }
    if((count>bestCount)&&(pri>exclusivePriLowerLimit))
      {
      bestCount=count;
      bestPri=pri;
      }
    }
  }
return bestPri;
}

//---------------------------------------------------------------------------
