//----------------------------------------------------------------------------
// Copyright (c) 2015-2016 Fabrice HARROUET (ENIB)
//
// This file is part of GpuElint.
//   http://www.enib.fr/~harrouet/gpuelint.html
//
// GpuElint is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GpuElint is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with GpuElint.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#ifndef GPUELINT_H
#define GPUELINT_H 1

#include "gpuElint_platform.h"

#define GPUELINT_DTOA_SKIP_COUNT 4 // adjust as needed

#define GPUELINT_MAX_MATCH_COUNT 8 // adjust as needed

/** @file
 * A GPU processing environment.
 * It provides means to create and transfer buffers between host and GPU
 * as well as predefined GPU processing algorithms.
 * Prior to any invocation, the current thread must hold an active OpenGL
 * context which must remain valid during the whole usage of this class.
 * If computation has to take place in another thread, this latter must
 * create its own processing environment which refers to the initial one;
 * their ressources will be shared into the initial OpenGL context.
 */
typedef struct _GpuElint GpuElint;

/**
 * GPU computing properties chosen or detected at initialisation.
 */
typedef struct _GpuElintProperties
{
int DTOA_SKIP_COUNT;
int MAX_MATCH_COUNT;
int glVersionMajor;
int glVersionMinor;
int glslVersion;
int maxWorkGroupCount[3];
int maxWorkGroupSize[3];
int maxWorkGroupInvocations;
int maxUniformBlocks;
int maxTextureImageUnits;
int maxAtomicCounterBuffers;
int maxAtomicCounters;
int maxSharedMemorySize;
int maxUniformComponents;
int maxImageUniforms;
int maxCombinedUniformComponents;
int maxShaderStorageBlocks;
int warpSize;
int warpsPerStreamingMultiprocessor;
int streamingMultiprocessorCount;
bool hasBallot;
bool hasShuffle;
} GpuElintProperties;

/**
 * Constructs a new GPU processing environment.
 * @param  shared
 *         A shared GPU processing environment used within another thread
 *         or 0.
 * @return A new GPU environment (0 on failure).
 */
GPUELINT_API
GpuElint * // new GPU environment (0 on failure)
GpuElint_new(const GpuElint *shared);

/**
 * Deletes a GPU processing environment that will not be used anymore.
 * @param  ge
 *         The concerned GPU processing environment.
 */
GPUELINT_API
void
GpuElint_delete(GpuElint *ge);

/**
 * Provides a copy of the GPU computing properties that were chosen or
 * detected at initialisation.
 * @param  ge
 *         The concerned GPU processing environment.
 * @return A set of properties (0 on failure).
 */
GPUELINT_API
const GpuElintProperties * // GPU properties (0 on failure)
GpuElint_properties(const GpuElint *ge);

/**
 * Gives the time elapsed since the initial GpuElint creation.
 * @param  ge
 *         The concerned GPU processing environment.
 * @return The time in seconds (-1.0f on failure).
 */
GPUELINT_API
float // seconds elapsed since initialisation (-1.0f on failure)
GpuElint_elapsed(const GpuElint *ge);

/**
 * Ensures all the preceding commands have been sent to the GPU.
 * @param  ge
 *         The concerned GPU processing environment.
 * @param  wait
 *         Tells whether to wait for these commands to complete or not.
 * @return Success.
 */
GPUELINT_API
bool // success
GpuElint_flush(GpuElint *ge,
               bool wait);

/**
 * Allocates a new GPU-buffer.
 * @param  ge
 *         The concerned GPU processing environment.
 * @param  capacity
 *         The capacity of the new GPU-buffer, in bytes.
 * @return The identifier of the new GPU-buffer (0 on failure).
 */
GPUELINT_API
int // GPU-buffer identifier (0 on failure)
GpuElint_gpuAlloc(GpuElint *ge,
                  intptr_t capacity);

/**
 * Frees a GPU-buffer that will not be used anymore.
 * @param  ge
 *         The concerned GPU processing environment.
 * @param  gpuBuffer
 *         The identifier of the GPU-buffer to be freed.
 * @return Success.
 */
GPUELINT_API
bool // success
GpuElint_gpuFree(GpuElint *ge,
                 int gpuBuffer);

/**
 * Maps a GPU-buffer to host-memory.
 * This GPU-buffer will be unusable by the GPU until it is unmapped.
 * @param  ge
 *         The concerned GPU processing environment.
 * @param  gpuBuffer
 *         The identifier of the GPU-buffer to be mapped.
 * @param  offset
 *         The offset inside the GPU-buffer, in bytes.
 * @param  size
 *         The amount of data to be mapped, in bytes.
 * @param  readAccess
 *         Tells if GPU data must be retreived.
 * @param  writeAccess
 *         Tells if host data must be sent to GPU at unmapping.
 * @return The host-memory address of the mapped GPU-buffer (0 on failure).
 */
GPUELINT_API
void * // mapped address (0 on failure)
GpuElint_beginMapping(GpuElint *ge,
                      int gpuBuffer,
                      intptr_t offset,
                      intptr_t size,
                      bool readAccess,
                      bool writeAccess);

/**
 * Unmaps a GPU-buffer from host-memory.
 * This GPU-buffer becomes usable again by the GPU, but the host-memory
 * address previously obtained for its mapping becomes unusable.
 * @param  ge
 *         The concerned GPU processing environment.
 * @param  gpuBuffer
 *         The identifier of the GPU-buffer to be unmapped.
 * @return Success.
 */
GPUELINT_API
bool // success
GpuElint_endMapping(GpuElint *ge,
                    int gpuBuffer);

/**
 * Copies some data from a host-buffer to a GPU-buffer.
 * @param  ge
 *         The concerned GPU processing environment.
 * @param  gpuDst
 *         The identifier of the destination GPU-buffer.
 * @param  hostSrc
 *         The source host-buffer.
 * @param  dstOffset
 *         The offset inside the destination GPU-buffer, in bytes.
 * @param  srcOffset
 *         The offset inside the source host-buffer, in bytes.
 * @param  size
 *         The amount of data to be copied, in bytes.
 * @return Success.
 */
GPUELINT_API
bool // success
GpuElint_hostToGpu(GpuElint *ge,
                   int gpuDst,
                   const void *hostSrc,
                   intptr_t dstOffset,
                   intptr_t srcOffset,
                   intptr_t size);

/**
 * Copies some data from a GPU-buffer to a host-buffer.
 * @param  ge
 *         The concerned GPU processing environment.
 * @param  hostDst
 *         The destination host-buffer.
 * @param  gpuSrc
 *         The identifier of the source GPU-buffer.
 * @param  dstOffset
 *         The offset inside the destination host-buffer, in bytes.
 * @param  srcOffset
 *         The offset inside the source GPU-buffer, in bytes.
 * @param  size
 *         The amount of data to be copied, in bytes.
 * @return Success.
 */
GPUELINT_API
bool // success
GpuElint_gpuToHost(GpuElint *ge,
                   void *hostDst,
                   int gpuSrc,
                   intptr_t dstOffset,
                   intptr_t srcOffset,
                   intptr_t size);

/**
 * Copies some data from a GPU-buffer to another.
 * @param  ge
 *         The concerned GPU processing environment.
 * @param  gpuDst
 *         The identifier of the destination GPU-buffer.
 * @param  gpuSrc
 *         The identifier of the source GPU-buffer.
 * @param  dstOffset
 *         The offset inside the destination GPU-buffer, in bytes.
 * @param  srcOffset
 *         The offset inside the source GPU-buffer, in bytes.
 * @param  size
 *         The amount of data to be copied, in bytes.
 * @return Success.
 */
GPUELINT_API
bool // success
GpuElint_gpuToGpu(GpuElint *ge,
                  int gpuDst,
                  int gpuSrc,
                  intptr_t dstOffset,
                  intptr_t srcOffset,
                  intptr_t size);

/**
 * Creates a new GPU-time-marker to measure GPU processing durations.
 * @param  ge
 *         The concerned GPU processing environment.
 * @return The identifier of the new GPU-time-marker (0 on failure).
 */
GPUELINT_API
int // GPU-time-marker identifier (0 on failure)
GpuElint_createMarker(GpuElint *ge);

/**
 * Destroys a GPU-time-marker that will not be used anymore.
 * @param  ge
 *         The concerned GPU processing environment.
 * @param  marker
 *         The identifier of the GPU-time-marker to be destroyed.
 * @return Success.
 */
GPUELINT_API
bool // success
GpuElint_destroyMarker(GpuElint *ge,
                       int marker);

/**
 * Inserts a GPU-time-marker in the processing stream of the GPU.
 * @param  ge
 *         The concerned GPU processing environment.
 * @param  marker
 *         The identifier of the GPU-time-marker to be inserted.
 * @return Success.
 */
GPUELINT_API
bool // success
GpuElint_setMarker(GpuElint *ge,
                   int marker);

/**
 * Tests whether a GPU-time-marker has been reached by the GPU processing
 * stream or not.
 * @param  ge
 *         The concerned GPU processing environment.
 * @param  marker
 *         The identifier of the GPU-time-marker to be tested.
 * @return 0 if the marker is not reached, 1 if it is (-1 on failure).
 */
GPUELINT_API
int // 0: marker is not reached  1: marker is reached  -1: failure
GpuElint_testMarker(GpuElint *ge,
                    int marker);

/**
 * Gives the time elapsed between two GPU-time-markers.
 * @param  ge
 *         The concerned GPU processing environment.
 * @param  firstMarker
 *         The identifier of the first GPU-time-marker to reach.
 * @param  secondMarker
 *         The identifier of the second GPU-time-marker to reach.
 * @return The duration in seconds (-1.0f on failure).
 */
GPUELINT_API
float // seconds between first and second markers (-1.0f on failure)
GpuElint_markerInterval(GpuElint *ge,
                        int firstMarker,
                        int secondMarker);

/**
 * Sets to zero the float or integer value of each element.
 * @param  ge
 *         The concerned GPU processing environment.
 * @param  gpuDst
 *         The identifier of the destination GPU-buffer.
 * @param  begin
 *         The index of the first element to consider.
 * @param  end
 *         The index after the last element to consider.
 * @return Success.
 */
GPUELINT_API
bool // success
GpuElint_zeroFill(GpuElint *ge,
                  int gpuDst,
                  int begin,
                  int end);

/**
 * Initialises the integer value of each element by counting them
 * as follows:
 *   value of element at (begin+i) = countStart + i * countIncrement
 * @param  ge
 *         The concerned GPU processing environment.
 * @param  gpuDst
 *         The identifier of the destination GPU-buffer.
 * @param  begin
 *         The index of the first element to consider.
 * @param  end
 *         The index after the last element to consider.
 * @param  countStart
 *         The starting value to count from.
 * @param  countIncrement
 *         The increment between each consecutive value.
 * @return Success.
 */
GPUELINT_API
bool // success
GpuElint_countFill(GpuElint *ge,
                   int gpuDst,
                   int begin,
                   int end,
                   int countStart,
                   int countIncrement);

/**
 * Description of a mode detected in a 1D histogram.
 */
typedef struct _HistoMode1D
{
int maxCount;
int weight;
int firstBin;
int lastBin;
} HistoMode1D;

/**
 * Detects the modes of a 1D histogram.
 * @param  ge
 *         The concerned GPU processing environment.
 * @param  hostDst_modes
 *         The destination host-buffer to store the detected modes
 *         (must be preallocated and large enough for modeCapacity).
 * @param  modeCapacity
 *         The capacity of hostDst_modes, in modes.
 * @param  gpuSrc_counts
 *         The identifier of the source GPU-buffer containing the
 *         counts of each bin.
 * @param  cutoffRatio
 *         The cutoff level of a mode relative its maximum.
 * @param  binCount
 *         The number of bins in the histogram.
 * @param  cyclic
 *         Tells if detection should cycle through histogram boundaries.
 * @return The number of detected modes (-1 on failure).
 */
GPUELINT_API
int // number of modes (-1 on failure)
GpuElint_modes1D(GpuElint *ge,
                 HistoMode1D *hostDst_modes,
                 int modeCapacity,
                 int gpuSrc_counts,
                 float cutoffRatio,
                 int binCount,
                 bool cyclic);

/**
 * Description of a mode detected in a 2D histogram.
 */
typedef struct _HistoMode2D
{
int maxCount;
int weight;
int firstBinA;
int lastBinA;
int firstBinB;
int lastBinB;
} HistoMode2D;

/**
 * Detects the modes of a 2D histogram.
 * The histogram is firstly ordered along A values then B.
 * @param  ge
 *         The concerned GPU processing environment.
 * @param  hostDst_modes
 *         The destination host-buffer to store the detected modes
 *         (must be preallocated and large enough for modeCapacity).
 * @param  modeCapacity
 *         The capacity of hostDst_modes, in modes.
 * @param  gpuSrc_counts
 *         The identifier of the source GPU-buffer containing the
 *         counts of each bin.
 * @param  cutoffRatio
 *         The cutoff level of a mode relative its maximum.
 * @param  binCountA
 *         The number of bins for A values in the histogram.
 * @param  cyclicA
 *         Tells if detection should cycle through histogram boundaries
 *         for A values.
 * @param  binCountB
 *         The number of bins for B values in the histogram.
 * @param  cyclicB
 *         Tells if detection should cycle through histogram boundaries
 *         for B values.
 * @return The number of detected modes (-1 on failure).
 */
GPUELINT_API
int // number of modes (-1 on failure)
GpuElint_modes2D(GpuElint *ge,
                 HistoMode2D *hostDst_modes,
                 int modeCapacity,
                 int gpuSrc_counts,
                 float cutoffRatio,
                 int binCountA,
                 bool cyclicA,
                 int binCountB,
                 bool cyclicB);

/**
 * Description of a mode detected in a 3D histogram.
 */
typedef struct _HistoMode3D
{
int maxCount;
int weight;
int firstBinA;
int lastBinA;
int firstBinB;
int lastBinB;
int firstBinC;
int lastBinC;
} HistoMode3D;

/**
 * Detects the modes of a 3D histogram.
 * The histogram is firstly ordered along A values then B and C.
 * @param  ge
 *         The concerned GPU processing environment.
 * @param  hostDst_modes
 *         The destination host-buffer to store the detected modes
 *         (must be preallocated and large enough for modeCapacity).
 * @param  modeCapacity
 *         The capacity of hostDst_modes, in modes.
 * @param  gpuSrc_counts
 *         The identifier of the source GPU-buffer containing the
 *         counts of each bin.
 * @param  cutoffRatio
 *         The cutoff level of a mode relative its maximum.
 * @param  binCountA
 *         The number of bins for A values in the histogram.
 * @param  cyclicA
 *         Tells if detection should cycle through histogram boundaries
 *         for A values.
 * @param  binCountB
 *         The number of bins for B values in the histogram.
 * @param  cyclicB
 *         Tells if detection should cycle through histogram boundaries
 *         for B values.
 * @param  binCountC
 *         The number of bins for C values in the histogram.
 * @param  cyclicC
 *         Tells if detection should cycle through histogram boundaries
 *         for C values.
 * @return The number of detected modes (-1 on failure).
 */
GPUELINT_API
int // number of modes (-1 on failure)
GpuElint_modes3D(GpuElint *ge,
                 HistoMode3D *hostDst_modes,
                 int modeCapacity,
                 int gpuSrc_counts,
                 float cutoffRatio,
                 int binCountA,
                 bool cyclicA,
                 int binCountB,
                 bool cyclicB,
                 int binCountC,
                 bool cyclicC);

/**
 * Description of a mode detected in a 4D histogram.
 */
typedef struct _HistoMode4D
{
int maxCount;
int weight;
int firstBinA;
int lastBinA;
int firstBinB;
int lastBinB;
int firstBinC;
int lastBinC;
int firstBinD;
int lastBinD;
} HistoMode4D;

/**
 * Detects the modes of a 4D histogram.
 * The histogram is firstly ordered along A values then B, C and D.
 * @param  ge
 *         The concerned GPU processing environment.
 * @param  hostDst_modes
 *         The destination host-buffer to store the detected modes
 *         (must be preallocated and large enough for modeCapacity).
 * @param  modeCapacity
 *         The capacity of hostDst_modes, in modes.
 * @param  gpuSrc_counts
 *         The identifier of the source GPU-buffer containing the
 *         counts of each bin.
 * @param  cutoffRatio
 *         The cutoff level of a mode relative its maximum.
 * @param  binCountA
 *         The number of bins for A values in the histogram.
 * @param  cyclicA
 *         Tells if detection should cycle through histogram boundaries
 *         for A values.
 * @param  binCountB
 *         The number of bins for B values in the histogram.
 * @param  cyclicB
 *         Tells if detection should cycle through histogram boundaries
 *         for B values.
 * @param  binCountC
 *         The number of bins for C values in the histogram.
 * @param  cyclicC
 *         Tells if detection should cycle through histogram boundaries
 *         for C values.
 * @param  binCountD
 *         The number of bins for D values in the histogram.
 * @param  cyclicD
 *         Tells if detection should cycle through histogram boundaries
 *         for D values.
 * @return The number of detected modes (-1 on failure).
 */
GPUELINT_API
int // number of modes (-1 on failure)
GpuElint_modes4D(GpuElint *ge,
                 HistoMode4D *hostDst_modes,
                 int modeCapacity,
                 int gpuSrc_counts,
                 float cutoffRatio,
                 int binCountA,
                 bool cyclicA,
                 int binCountB,
                 bool cyclicB,
                 int binCountC,
                 bool cyclicC,
                 int binCountD,
                 bool cyclicD);

/**
 * Attempts to estimate the pulse-repetition-interval (PRI) which is the
 * most significant in a DTOA histogram.
 * @param  ge
 *         The concerned GPU processing environment.
 * @param  hostSrc_counts
 *         The source host-buffer containing the DTOA counts of each bin.
 * @param  hostSrc_sums
 *         The source host-buffer containing the DTOA sums of each bin.
 * @param  binCount
 *         The number of bins of the histogram.
 * @param  minDtoa
 *         The lower bound (strictly greater than 0.0f) for a DTOA to be
 *         inserted in the histogram.
 * @param  dtoaRatio
 *         The multiplicative factor (strictly greater than 1.0f) which
 *         determines the logarithmic width of each bin.
 * @param  exclusivePriLowerLimit
 *         The strict lower limit (possibly 0.0f) above which a PRI will be
 *         considered in the estimation.
 * @param  minDtoaSum
 *         The minimal DTOA sum for a bin to be considered.
 * @return The most significant pulse-repetition-interval or 0.0f
 */
GPUELINT_API
float // pulse-repetition-interval or 0.0f
GpuElint_estimateMajorPri(GpuElint *ge,
                          const int *hostSrc_counts,
                          const float *hostSrc_sums,
                          int binCount,
                          float minDtoa,
                          float dtoaRatio,
                          float exclusivePriLowerLimit,
                          float minDtoaSum);

//----------------------------------------------------------------------------

/**
 * Determines the min and max of the float value of each pulse,
 * as well as their respective indices.
 * @param  ge
 *         The concerned GPU processing environment.
 * @param  gpuSrc_indices
 *         The identifier of the source GPU-buffer containing the indices
 *         of the pulses to analyse.
 * @param  pulseCount
 *         The number of pulses to consider.
 * @param  gpuSrc_values
 *         The identifier of the source GPU-buffer containing the values
 *         to analyse.
 * @param  out_minValue
 *         The output storage for the minimal value.
 * @param  out_minId
 *         The output storage for the index of the minimal value.
 * @param  out_maxValue
 *         The output storage for the maximal value found.
 * @param  out_maxId
 *         The output storage for the index of the maximal value.
 * @return Success.
 */
GPUELINT_API
bool // success
GpuElint_ID_minMax(GpuElint *ge,
                   int gpuSrc_indices,
                   int pulseCount,
                   int gpuSrc_values,
                   float *out_minValue,
                   int *out_minId,
                   float *out_maxValue,
                   int *out_maxId);

/**
 * Builds up an histogram of the float value of each pulse.
 * The destination GPU-buffer is initially reset to zero.
 * @param  ge
 *         The concerned GPU processing environment.
 * @param  gpuSrc_indices
 *         The identifier of the source GPU-buffer containing the indices
 *         of the pulses to analyse.
 * @param  pulseCount
 *         The number of pulses to consider.
 * @param  gpuDst_counts
 *         The identifier of the destination GPU-buffer to store the
 *         counts of each bin.
 * @param  gpuSrc_values
 *         The identifier of the source GPU-buffer containing the values
 *         to analyse.
 * @param  minValue
 *         The lower bound of the range used for counting values.
 * @param  maxValue
 *         The upper bound of the range used for counting values.
 * @param  binCount
 *         The number of bins of the range used for counting values.
 * @return Success.
 */
GPUELINT_API
bool // success
GpuElint_ID_histogram1D(GpuElint *ge,
                        int gpuSrc_indices,
                        int pulseCount,
                        int gpuDst_counts,
                        int gpuSrc_values,
                        float minValue,
                        float maxValue,
                        int binCount);

/**
 * Builds up a 2D histogram of the float (A,B) values of each pulse.
 * The destination GPU-buffer is initially reset to zero and is firstly
 * ordered along A then B.
 * @param  ge
 *         The concerned GPU processing environment.
 * @param  gpuSrc_indices
 *         The identifier of the source GPU-buffer containing the indices
 *         of the pulses to analyse.
 * @param  pulseCount
 *         The number of pulses to consider.
 * @param  gpuDst_counts
 *         The identifier of the destination GPU-buffer to store the
 *         counts of each bin.
 * @param  gpuSrc_valuesA
 *         The identifier of the source GPU-buffer containing the A values
 *         to analyse.
 * @param  minValueA
 *         The lower bound of the range used for counting A values.
 * @param  maxValueA
 *         The upper bound of the range used for counting A values.
 * @param  binCountA
 *         The number of bins of the range used for counting A values.
 * @param  gpuSrc_valuesB
 *         The identifier of the source GPU-buffer containing the B values
 *         to analyse.
 * @param  minValueB
 *         The lower bound of the range used for counting B values.
 * @param  maxValueB
 *         The upper bound of the range used for counting B values.
 * @param  binCountB
 *         The number of bins of the range used for counting B values.
 * @return Success.
 */
GPUELINT_API
bool // success
GpuElint_ID_histogram2D(GpuElint *ge,
                        int gpuSrc_indices,
                        int pulseCount,
                        int gpuDst_counts,
                        int gpuSrc_valuesA,
                        float minValueA,
                        float maxValueA,
                        int binCountA,
                        int gpuSrc_valuesB,
                        float minValueB,
                        float maxValueB,
                        int binCountB);

/**
 * Builds up a 3D histogram of the float (A,B,C) values of each pulse.
 * The destination GPU-buffer is initially reset to zero and is firstly
 * ordered along A then B, and C.
 * @param  ge
 *         The concerned GPU processing environment.
 * @param  gpuSrc_indices
 *         The identifier of the source GPU-buffer containing the indices
 *         of the pulses to analyse.
 * @param  pulseCount
 *         The number of pulses to consider.
 * @param  gpuDst_counts
 *         The identifier of the destination GPU-buffer to store the
 *         counts of each bin.
 * @param  gpuSrc_valuesA
 *         The identifier of the source GPU-buffer containing the A values
 *         to analyse.
 * @param  minValueA
 *         The lower bound of the range used for counting A values.
 * @param  maxValueA
 *         The upper bound of the range used for counting A values.
 * @param  binCountA
 *         The number of bins of the range used for counting A values.
 * @param  gpuSrc_valuesB
 *         The identifier of the source GPU-buffer containing the B values
 *         to analyse.
 * @param  minValueB
 *         The lower bound of the range used for counting B values.
 * @param  maxValueB
 *         The upper bound of the range used for counting B values.
 * @param  binCountB
 *         The number of bins of the range used for counting B values.
 * @param  gpuSrc_valuesC
 *         The identifier of the source GPU-buffer containing the C values
 *         to analyse.
 * @param  minValueC
 *         The lower bound of the range used for counting C values.
 * @param  maxValueC
 *         The upper bound of the range used for counting C values.
 * @param  binCountC
 *         The number of bins of the range used for counting C values.
 * @return Success.
 */
GPUELINT_API
bool // success
GpuElint_ID_histogram3D(GpuElint *ge,
                        int gpusrc_indices,
                        int pulseCount,
                        int gpuDst_counts,
                        int gpuSrc_valuesA,
                        float minValueA,
                        float maxValueA,
                        int binCountA,
                        int gpuSrc_valuesB,
                        float minValueB,
                        float maxValueB,
                        int binCountB,
                        int gpuSrc_valuesC,
                        float minValueC,
                        float maxValueC,
                        int binCountC);

/**
 * Builds up a 4D histogram of the float (A,B,C,D) values of each pulse.
 * The destination GPU-buffer is initially reset to zero and is firstly
 * ordered along A then B, C and D.
 * @param  ge
 *         The concerned GPU processing environment.
 * @param  gpuSrc_indices
 *         The identifier of the source GPU-buffer containing the indices
 *         of the pulses to analyse.
 * @param  pulseCount
 *         The number of pulses to consider.
 * @param  gpuDst_counts
 *         The identifier of the destination GPU-buffer to store the
 *         counts of each bin.
 * @param  gpuSrc_valuesA
 *         The identifier of the source GPU-buffer containing the A values
 *         to analyse.
 * @param  minValueA
 *         The lower bound of the range used for counting A values.
 * @param  maxValueA
 *         The upper bound of the range used for counting A values.
 * @param  binCountA
 *         The number of bins of the range used for counting A values.
 * @param  gpuSrc_valuesB
 *         The identifier of the source GPU-buffer containing the B values
 *         to analyse.
 * @param  minValueB
 *         The lower bound of the range used for counting B values.
 * @param  maxValueB
 *         The upper bound of the range used for counting B values.
 * @param  binCountB
 *         The number of bins of the range used for counting B values.
 * @param  gpuSrc_valuesC
 *         The identifier of the source GPU-buffer containing the C values
 *         to analyse.
 * @param  minValueC
 *         The lower bound of the range used for counting C values.
 * @param  maxValueC
 *         The upper bound of the range used for counting C values.
 * @param  binCountC
 *         The number of bins of the range used for counting C values.
 * @param  gpuSrc_valuesD
 *         The identifier of the source GPU-buffer containing the D values
 *         to analyse.
 * @param  minValueD
 *         The lower bound of the range used for counting D values.
 * @param  maxValueD
 *         The upper bound of the range used for counting D values.
 * @param  binCountD
 *         The number of bins of the range used for counting D values.
 * @return Success.
 */
GPUELINT_API
bool // success
GpuElint_ID_histogram4D(GpuElint *ge,
                        int gpuSrc_indices,
                        int pulseCount,
                        int gpuDst_counts,
                        int gpuSrc_valuesA,
                        float minValueA,
                        float maxValueA,
                        int binCountA,
                        int gpuSrc_valuesB,
                        float minValueB,
                        float maxValueB,
                        int binCountB,
                        int gpuSrc_valuesC,
                        float minValueC,
                        float maxValueC,
                        int binCountC,
                        int gpuSrc_valuesD,
                        float minValueD,
                        float maxValueD,
                        int binCountD);

/**
 * Keep only the pulses which float value is in the specified interval.
 * @param  ge
 *         The concerned GPU processing environment.
 * @param  gpuSrc_indices
 *         The identifier of the source GPU-buffer containing the indices
 *         of the pulses to analyse.
 * @param  pulseCount
 *         The number of pulses to consider.
 * @param  gpuDst_indices
 *         The identifier of the destination GPU-buffer containing the
 *         indices of the remaining pulses.
 * @param  gpuSrc_values
 *         The identifier of the source GPU-buffer containing the values
 *         to analyse.
 * @param  minValue
 *         The minimal value of the interval.
 * @param  minValue
 *         The maximal value of the interval.
 * @return The number of remaining pulses (-1 on failure).
 */
GPUELINT_API
int // number of remaining pulses (-1 on failure)
GpuElint_ID_filter1D(GpuElint *ge,
                     int gpuSrc_indices,
                     int pulseCount,
                     int gpuDst_indices,
                     int gpuSrc_values,
                     float minValue,
                     float maxValue);

/**
 * Keep only the pulses which float (A,B) values are in the specified
 * interval.
 * @param  ge
 *         The concerned GPU processing environment.
 * @param  gpuSrc_indices
 *         The identifier of the source GPU-buffer containing the indices
 *         of the pulses to analyse.
 * @param  pulseCount
 *         The number of pulses to consider.
 * @param  gpuDst_indices
 *         The identifier of the destination GPU-buffer containing the
 *         indices of the remaining pulses.
 * @param  gpuSrc_valuesA
 *         The identifier of the source GPU-buffer containing the A values
 *         to analyse.
 * @param  minValueA
 *         The minimal value of the A interval.
 * @param  minValueA
 *         The maximal value of the A interval.
 * @param  gpuSrc_valuesB
 *         The identifier of the source GPU-buffer containing the B values
 *         to analyse.
 * @param  minValueB
 *         The minimal value of the B interval.
 * @param  minValueB
 *         The maximal value of the B interval.
 * @return The number of remaining pulses (-1 on failure).
 */
GPUELINT_API
int // number of remaining pulses (-1 on failure)
GpuElint_ID_filter2D(GpuElint *ge,
                     int gpuSrc_indices,
                     int pulseCount,
                     int gpuDst_indices,
                     int gpuSrc_valuesA,
                     float minValueA,
                     float maxValueA,
                     int gpuSrc_valuesB,
                     float minValueB,
                     float maxValueB);

/**
 * Keep only the pulses which float (A,B,C) values are in the specified
 * interval.
 * @param  ge
 *         The concerned GPU processing environment.
 * @param  gpuSrc_indices
 *         The identifier of the source GPU-buffer containing the indices
 *         of the pulses to analyse.
 * @param  pulseCount
 *         The number of pulses to consider.
 * @param  gpuDst_indices
 *         The identifier of the destination GPU-buffer containing the
 *         indices of the remaining pulses.
 * @param  gpuSrc_valuesA
 *         The identifier of the source GPU-buffer containing the A values
 *         to analyse.
 * @param  minValueA
 *         The minimal value of the A interval.
 * @param  minValueA
 *         The maximal value of the A interval.
 * @param  gpuSrc_valuesB
 *         The identifier of the source GPU-buffer containing the B values
 *         to analyse.
 * @param  minValueB
 *         The minimal value of the B interval.
 * @param  minValueB
 *         The maximal value of the B interval.
 * @param  gpuSrc_valuesC
 *         The identifier of the source GPU-buffer containing the C values
 *         to analyse.
 * @param  minValueC
 *         The minimal value of the C interval.
 * @param  minValueC
 *         The maximal value of the C interval.
 * @return The number of remaining pulses (-1 on failure).
 */
GPUELINT_API
int // number of remaining pulses (-1 on failure)
GpuElint_ID_filter3D(GpuElint *ge,
                     int gpuSrc_indices,
                     int pulseCount,
                     int gpuDst_indices,
                     int gpuSrc_valuesA,
                     float minValueA,
                     float maxValueA,
                     int gpuSrc_valuesB,
                     float minValueB,
                     float maxValueB,
                     int gpuSrc_valuesC,
                     float minValueC,
                     float maxValueC);

/**
 * Keep only the pulses which float (A,B,C,D) values are in the specified
 * interval.
 * @param  ge
 *         The concerned GPU processing environment.
 * @param  gpuSrc_indices
 *         The identifier of the source GPU-buffer containing the indices
 *         of the pulses to analyse.
 * @param  pulseCount
 *         The number of pulses to consider.
 * @param  gpuDst_indices
 *         The identifier of the destination GPU-buffer containing the
 *         indices of the remaining pulses.
 * @param  gpuSrc_valuesA
 *         The identifier of the source GPU-buffer containing the A values
 *         to analyse.
 * @param  minValueA
 *         The minimal value of the A interval.
 * @param  minValueA
 *         The maximal value of the A interval.
 * @param  gpuSrc_valuesB
 *         The identifier of the source GPU-buffer containing the B values
 *         to analyse.
 * @param  minValueB
 *         The minimal value of the B interval.
 * @param  minValueB
 *         The maximal value of the B interval.
 * @param  gpuSrc_valuesC
 *         The identifier of the source GPU-buffer containing the C values
 *         to analyse.
 * @param  minValueC
 *         The minimal value of the C interval.
 * @param  minValueC
 *         The maximal value of the C interval.
 * @param  gpuSrc_valuesD
 *         The identifier of the source GPU-buffer containing the D values
 *         to analyse.
 * @param  minValueD
 *         The minimal value of the D interval.
 * @param  minValueD
 *         The maximal value of the D interval.
 * @return The number of remaining pulses (-1 on failure).
 */
GPUELINT_API
int // number of remaining pulses (-1 on failure)
GpuElint_ID_filter4D(GpuElint *ge,
                     int gpuSrc_indices,
                     int pulseCount,
                     int gpuDst_indices,
                     int gpuSrc_valuesA,
                     float minValueA,
                     float maxValueA,
                     int gpuSrc_valuesB,
                     float minValueB,
                     float maxValueB,
                     int gpuSrc_valuesC,
                     float minValueC,
                     float maxValueC,
                     int gpuSrc_valuesD,
                     float minValueD,
                     float maxValueD);

/**
 * Builds up a histogram of the differences (DTOA) between successives
 * time-of-arrival (TOA) values.
 * These TOA differences are computed up to the DTOA_SKIP_COUNT
 * built-in constant.
 * The destination host-buffers are initialy reset to zero.
 * @param  ge
 *         The concerned GPU processing environment.
 * @param  gpuSrc_indices
 *         The identifier of the source GPU-buffer containing the indices
 *         of the pulses to analyse.
 * @param  pulseCount
 *         The number of pulses to consider.
 * @param  gpuSrc_toa64
 *         The identifier of the source GPU-buffer containing the TOA values
 *         (64-bit real) to analyse.
 * @param  hostDst_counts
 *         The destination host-buffer to store the DTOA counts of each bin.
 * @param  hostDst_sums
 *         The destination host-buffer to store the DTOA sums of each bin.
 * @param  binCount
 *         The number of bins of the histogram.
 * @param  minDtoa
 *         The lower bound (strictly greater than 0.0f) for a DTOA to be
 *         inserted in the histogram.
 * @param  dtoaRatio
 *         The multiplicative factor (strictly greater than 1.0f) which
 *         determines the logarithmic width of each bin.
 * @return Success.
 */
GPUELINT_API
bool // success
GpuElint_ID_dtoaHistogram(GpuElint *ge,
                          int gpuSrc_indices,
                          int pulseCount,
                          int gpuSrc_toa64,
                          int *hostDst_counts,
                          float *hostDst_sums,
                          int binCount,
                          float minDtoa,
                          float dtoaRatio);

/**
 * Find sequences of pulses which DTOA match the given PRI and mark them.
 * @param  ge
 *         The concerned GPU processing environment.
 * @param  gpuSrc_indices
 *         The identifier of the source GPU-buffer containing the indices
 *         of the pulses to analyse.
 * @param  pulseCount
 *         The number of pulses to consider.
 * @param  gpuSrc_toa64
 *         The identifier of the source GPU-buffer containing the TOA values
 *         (64-bit real) to analyse.
 * @param  gpu_marks
 *         The identifier of the GPU-buffer to store the mark of each pulse.
 * @param  mark
 *         The mark to assign to each matching pulse.
 * @param  estimatedPri
 *         The previously estimated PRI to be looked for.
 * @param  priRatioThreshold
 *         The precision factor (greater than 1.0f) used to compare the DTOA
 *         in a sequence with the expected PRI.
 * @param  minMatchCount
 *         The minimal number of matching DTOA to detect a sequence.
 * @return The actual PRI of the marked pulses or 0.0f (-1.0f on failure).
 */
GPUELINT_API
float // actual pulse-repetition-interval or 0.0f (-1.0f on failure)
GpuElint_ID_markPulses(GpuElint *ge,
                       int gpuSrc_indices,
                       int pulseCount,
                       int gpuSrc_toa64,
                       int gpu_marks,
                       int mark,
                       float estimatedPri,
                       float priRatioThreshold,
                       int minMatchCount);

/**
 * Keep only the pulses with no mark (<0).
 * @param  ge
 *         The concerned GPU processing environment.
 * @param  gpuSrc_indices
 *         The identifier of the source GPU-buffer containing the indices
 *         of the pulses to analyse.
 * @param  pulseCount
 *         The number of pulses to consider.
 * @param  gpuDst_indices
 *         The identifier of the destination GPU-buffer containing the
 *         indices of the remaining pulses.
 * @param  gpuSrc_marks
 *         The identifier of the source GPU-buffer containing the marks
 *         to analyse.
 * @return The number of remaining pulses (-1 on failure).
 */
GPUELINT_API
int // number of remaining pulses (-1 on failure)
GpuElint_ID_filterUnmarked(GpuElint *ge,
                           int gpuSrc_indices,
                           int pulseCount,
                           int gpuDst_indices,
                           int gpuSrc_marks);

//----------------------------------------------------------------------------

/**
 * Bitmask-based version of GpuElint_ID_minMax() (testing purpose only).
 */
GPUELINT_API
bool // success
GpuElint_BM_minMax(GpuElint *ge,
                   int gpuSrc_bitmask,
                   int pulseCount,
                   int gpuSrc_values,
                   float *out_minValue,
                   int *out_minId,
                   float *out_maxValue,
                   int *out_maxId);

/**
 * Bitmask-based version of GpuElint_ID_histogram1D() (testing purpose only).
 */
GPUELINT_API
bool // success
GpuElint_BM_histogram1D(GpuElint *ge,
                        int gpuSrc_bitmask,
                        int pulseCount,
                        int gpuDst_counts,
                        int gpuSrc_values,
                        float minValue,
                        float maxValue,
                        int binCount);

/**
 * Bitmask-based version of GpuElint_ID_histogram2D() (testing purpose only).
 */
GPUELINT_API
bool // success
GpuElint_BM_histogram2D(GpuElint *ge,
                        int gpuSrc_bitmask,
                        int pulseCount,
                        int gpuDst_counts,
                        int gpuSrc_valuesA,
                        float minValueA,
                        float maxValueA,
                        int binCountA,
                        int gpuSrc_valuesB,
                        float minValueB,
                        float maxValueB,
                        int binCountB);

/**
 * Bitmask-based version of GpuElint_ID_histogram3D() (testing purpose only).
 */
GPUELINT_API
bool // success
GpuElint_BM_histogram3D(GpuElint *ge,
                        int gpuSrc_bitmask,
                        int pulseCount,
                        int gpuDst_counts,
                        int gpuSrc_valuesA,
                        float minValueA,
                        float maxValueA,
                        int binCountA,
                        int gpuSrc_valuesB,
                        float minValueB,
                        float maxValueB,
                        int binCountB,
                        int gpuSrc_valuesC,
                        float minValueC,
                        float maxValueC,
                        int binCountC);

/**
 * Bitmask-based version of GpuElint_ID_histogram4D() (testing purpose only).
 */
GPUELINT_API
bool // success
GpuElint_BM_histogram4D(GpuElint *ge,
                        int gpuSrc_bitmask,
                        int pulseCount,
                        int gpuDst_counts,
                        int gpuSrc_valuesA,
                        float minValueA,
                        float maxValueA,
                        int binCountA,
                        int gpuSrc_valuesB,
                        float minValueB,
                        float maxValueB,
                        int binCountB,
                        int gpuSrc_valuesC,
                        float minValueC,
                        float maxValueC,
                        int binCountC,
                        int gpuSrc_valuesD,
                        float minValueD,
                        float maxValueD,
                        int binCountD);

/**
 * Bitmask-based version of GpuElint_ID_filter1D() (testing purpose only).
 */
GPUELINT_API
bool // success
GpuElint_BM_filter1D(GpuElint *ge,
                     int gpuSrc_bitmask,
                     int pulseCount,
                     int gpuDst_bitmask,
                     int gpuSrc_values,
                     float minValue,
                     float maxValue);

/**
 * Bitmask-based version of GpuElint_ID_filter2D() (testing purpose only).
 */
GPUELINT_API
bool // success
GpuElint_BM_filter2D(GpuElint *ge,
                     int gpuSrc_bitmask,
                     int pulseCount,
                     int gpuDst_bitmask,
                     int gpuSrc_valuesA,
                     float minValueA,
                     float maxValueA,
                     int gpuSrc_valuesB,
                     float minValueB,
                     float maxValueB);

/**
 * Bitmask-based version of GpuElint_ID_filter3D() (testing purpose only).
 */
GPUELINT_API
bool // success
GpuElint_BM_filter3D(GpuElint *ge,
                     int gpuSrc_bitmask,
                     int pulseCount,
                     int gpuDst_bitmask,
                     int gpuSrc_valuesA,
                     float minValueA,
                     float maxValueA,
                     int gpuSrc_valuesB,
                     float minValueB,
                     float maxValueB,
                     int gpuSrc_valuesC,
                     float minValueC,
                     float maxValueC);

/**
 * Bitmask-based version of GpuElint_ID_filter4D() (testing purpose only).
 */
GPUELINT_API
bool // success
GpuElint_BM_filter4D(GpuElint *ge,
                     int gpuSrc_bitmask,
                     int pulseCount,
                     int gpuDst_bitmask,
                     int gpuSrc_valuesA,
                     float minValueA,
                     float maxValueA,
                     int gpuSrc_valuesB,
                     float minValueB,
                     float maxValueB,
                     int gpuSrc_valuesC,
                     float minValueC,
                     float maxValueC,
                     int gpuSrc_valuesD,
                     float minValueD,
                     float maxValueD);

/**
 * Bitmask-based version of GpuElint_ID_dtoahistogram() (testing purpose only).
 */
GPUELINT_API
bool // success
GpuElint_BM_dtoaHistogram(GpuElint *ge,
                          int gpuSrc_bitmask,
                          int pulseCount,
                          int gpuSrc_toa64,
                          int *hostDst_counts,
                          float *hostDst_sums,
                          int binCount,
                          float minDtoa,
                          float dtoaRatio);

/**
 * Bitmask-based version of GpuElint_ID_markPulses() (testing purpose only).
 */
GPUELINT_API
float // actual pulse-repetition-interval or 0.0f (-1.0f on failure)
GpuElint_BM_markPulses(GpuElint *ge,
                       int gpuSrc_bitmask,
                       int pulseCount,
                       int gpuSrc_toa64,
                       int gpu_marks,
                       int mark,
                       float estimatedPri,
                       float priRatioThreshold,
                       int minMatchCount);

/**
 * Bitmask-based version of GpuElint_ID_filterUnmarked() (testing purpose only).
 */
GPUELINT_API
bool // success
GpuElint_BM_filterUnmarked(GpuElint *ge,
                           int gpuSrc_bitmask,
                           int pulseCount,
                           int gpuDst_bitmask,
                           int gpuSrc_marks);

#endif // GPUELINT_H

//----------------------------------------------------------------------------
