//----------------------------------------------------------------------------
// Copyright (c) 2015-2016 Fabrice HARROUET (ENIB)
//
// This file is part of GpuElint.
//   http://www.enib.fr/~harrouet/gpuelint.html
//
// GpuElint is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GpuElint is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with GpuElint.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "gpuElint_internals.h"

typedef struct _GpuService_ID_Compact
{
GLuint program[2];
int groupSize;
int groupCount;
int gpu_remaining;
int backend;
intptr_t backendCapacity;
} GpuService_ID_Compact;

INLINE
int
computeAccumOffset(const GpuElint *ge,
                   int pulseCount)
{
const int warpSize=ge->properties.warpSize;
pulseCount+=warpSize-1;
return pulseCount-(pulseCount%warpSize);
}

int // identifier of the backend GPU-buffer for ID_compact or 0
GpuElint_ID_compactionBackend(GpuElint *ge,
                           int pulseCount)
{
GpuService_ID_Compact *s=ge->ID_compact_service;
const intptr_t capacity=(computeAccumOffset(ge,pulseCount)+pulseCount)*
                        sizeof(int32_t);
if(!capacity||(capacity>s->backendCapacity))
  {
  if(s->backend)
    {
    bool success=GpuElint_gpuFree(ge,s->backend);
    ON_ERROR(!success,"GpuElint_gpuFree() failure",{});
    }
  s->backendCapacity=capacity;
  s->backend=0;
  if(capacity)
    {
    s->backend=GpuElint_gpuAlloc(ge,capacity);
    ON_ERROR(!s->backend,"GpuElint_gpuAlloc() failure",{});
    }
  }
return s->backend;
}

static
bool // success
buildProgram0(GpuElint *ge,
              GpuService_ID_Compact *s)
{
char *src=(char *)0;
int srcSize=0;
int srcCapacity=0;
// a power-of-two group size is MANDATORY for prefix-sum
s->groupSize=GpuElint_chooseWorkGroupSize(ge,true);
s->groupCount=GpuElint_chooseWorkGroupCount(ge);
GpuElint_appendCodePreamble(ge,&src,&srcSize,&srcCapacity,
                            s->groupSize,s->groupCount);
GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
  " shared int sh_accum[GROUP_SIZE];            \n"
  "                                             \n"
  " READ_WRITE_BUFFER(0, int backend[]; )       \n"
  "                                             \n"
  " UNIFORM_PARAMETER(0, ivec2 )                \n"
  " const int pulseCount =UNI_0.x,              \n"
  "           accumOffset=UNI_0.y;              \n"
  "                                             \n"
  " #define ACCUM(i) (backend[accumOffset+(i)]) \n"
  "                                             \n");
GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
  " void                                                   \n"
  " main(void) // ID_compact() bottom-up pass              \n"
  " {                                                      \n"
  " const int highMask=1<<31;                              \n"
  " for(int id=globalId;id<pulseCount;id+=GRID_SIZE)       \n"
  "   {                                                    \n"
  "   barrier(); // no early overwrite of sh_accum[]       \n"
  "   sh_accum[localId]=(backend[id]&highMask)==0 ? 1 : 0; \n"
  "   for(int n=1;n<=GROUP_SHIFT;++n)                      \n"
  "     {                                                  \n"
  "     memoryBarrierShared();                             \n"
  "     barrier(); // wait for sh_accum[]                  \n"
  "     const int m=(1<<n)-1,s=1<<(n-1);                   \n"
  "     if((localId&m)==m)                                 \n"
  "       { sh_accum[localId]+=sh_accum[localId-s]; }      \n"
  "     }                                                  \n"
  "   ACCUM(id)=sh_accum[localId];                         \n"
  "   }                                                    \n"
  " }                                                      \n");
s->program[0]=GpuElint_buildProgram(ge,src);
Memory_free(src);
return s->program[0]!=0;
}

static
bool // success
buildProgram1(GpuElint *ge,
              GpuService_ID_Compact *s)
{
char *src=(char *)0;
int srcSize=0;
int srcCapacity=0;
GpuElint_appendCodePreamble(ge,&src,&srcSize,&srcCapacity,
                            s->groupSize,s->groupCount);
GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
  " shared int sh_accum[GROUP_SIZE];            \n"
  " shared int sh_prevAccum;                    \n"
  "                                             \n"
  " WRITE_ONLY_BUFFER(0, int remaining; )       \n"
  "  READ_ONLY_BUFFER(1, int backend[]; )       \n"
  " WRITE_ONLY_BUFFER(2, int indices[]; )       \n"
  "                                             \n"
  " UNIFORM_PARAMETER(0, ivec2 )                \n"
  " const int pulseCount =UNI_0.x,              \n"
  "           accumOffset=UNI_0.y;              \n"
  "                                             \n"
  " #define ACCUM(i) (backend[accumOffset+(i)]) \n"
  "                                             \n");
GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
  " void                                                 \n"
  " main(void) // ID_compact() top-down pass             \n"
  " {                                                    \n"
  " const int highMask=1<<31;                            \n"
  " int lastPrevAccum=0;                                 \n"
  " int lastPrevId=GROUP_MASK;                           \n"
  " for(int id=globalId;id<pulseCount;id+=GRID_SIZE)     \n"
  "   {                                                  \n"
  "   barrier(); // no early overwrite of sh_accum[]     \n"
  "   if(localId==0)                                     \n"
  "     {                                                \n"
  "     for(;lastPrevId<id;lastPrevId+=GROUP_SIZE)       \n"
  "       { lastPrevAccum+=ACCUM(lastPrevId); }          \n"
  "     sh_prevAccum=lastPrevAccum;                      \n"
  "     }                                                \n"
  "   sh_accum[localId]=ACCUM(id);                       \n"
  "   for(int n=GROUP_SHIFT-1;n>=1;--n)                  \n"
  "     {                                                \n"
  "     memoryBarrierShared();                           \n"
  "     barrier(); // wait for sh_accum[]                \n"
  "     const int m=(1<<n)-1,s=1<<(n-1);                 \n"
  "     if((localId>=s)&&((localId&m)==(m>>1)))          \n"
  "       { sh_accum[localId]+=sh_accum[localId-s]; }    \n"
  "     }                                                \n"
  "   memoryBarrierShared();                             \n"
  "   barrier(); // wait for sh_accum[localId-1]         \n"
  "   int prevAccum=sh_prevAccum;                        \n"
  "   int accum=prevAccum+sh_accum[localId];             \n"
  "   if(localId!=0) { prevAccum+=sh_accum[localId-1]; } \n"
  "   if(accum!=prevAccum) // store remaining pulses     \n"
  "     { indices[accum-1]=backend[id]; }                \n"
  "   if(id==pulseCount-1)                               \n"
  "     { remaining=accum; }                             \n"
  "   }                                                  \n"
  " }                                                    \n");
s->program[1]=GpuElint_buildProgram(ge,src);
Memory_free(src);
return s->program[1]!=0;
}

bool // success
GpuService_ID_Compact_init(GpuElint *ge)
{
GpuService_ID_Compact *s=Memory_alloc(1,GpuService_ID_Compact);
if(!buildProgram0(ge,s))
  {
  Memory_free(s);
  return false;
  }
if(!buildProgram1(ge,s))
  {
  glDeleteProgram(s->program[0]);
  Memory_free(s);
  return false;
  }
s->gpu_remaining=GpuElint_gpuAlloc(ge,sizeof(int32_t));
ON_ERROR(!s->gpu_remaining,"GpuElint_gpuAlloc() failure",
         {
         glDeleteProgram(s->program[1]);
         glDeleteProgram(s->program[0]);
         Memory_free(s);
         return false;
         });
s->backend=0;
s->backendCapacity=0;
ge->ID_compact_service=s;
return true;
}

void
GpuService_ID_Compact_destroy(GpuElint *ge)
{
GpuService_ID_Compact *s=ge->ID_compact_service;
if(!s) return;
GpuElint_ID_compactionBackend(ge,0);
bool success=GpuElint_gpuFree(ge,s->gpu_remaining);
ON_ERROR(!success,"GpuElint_gpuFree() failure",{});
glDeleteProgram(s->program[1]);
ON_GL_ERROR({});
glDeleteProgram(s->program[0]);
ON_GL_ERROR({});
Memory_free(s);
}

int // number of remaining pulses or -1
GpuElint_ID_compact(GpuElint *ge,
                    int gpu_backend,
                    int pulseCount,
                    int gpuDst_indices)
{
GPUELINT_API_INVOCATION_CHECK(ge,{ return -1; });
const GpuService_ID_Compact *s=ge->ID_compact_service;
const int accumOffset=computeAccumOffset(ge,pulseCount);
glUseProgram(s->program[0]); // bottom-up pass
ON_GL_ERROR({ return -1; });
glBindBufferBase(GL_SHADER_STORAGE_BUFFER,0,gpu_backend);
ON_GL_ERROR({ return -1; });
glUniform2i(0,pulseCount,accumOffset);
ON_GL_ERROR({ return -1; });
glDispatchCompute(s->groupCount,1,1);
ON_GL_ERROR({ return -1; });
glMemoryBarrier(GL_SHADER_STORAGE_BARRIER_BIT);
ON_GL_ERROR({ return -1; });
glUseProgram(s->program[1]); // top-down pass
ON_GL_ERROR({ return -1; });
glBindBufferBase(GL_SHADER_STORAGE_BUFFER,0,s->gpu_remaining);
ON_GL_ERROR({ return -1; });
glBindBufferBase(GL_SHADER_STORAGE_BUFFER,1,gpu_backend);
ON_GL_ERROR({ return -1; });
glBindBufferBase(GL_SHADER_STORAGE_BUFFER,2,gpuDst_indices);
ON_GL_ERROR({ return -1; });
glUniform2i(0,pulseCount,accumOffset);
ON_GL_ERROR({ return -1; });
glDispatchCompute(s->groupCount,1,1);
ON_GL_ERROR({ return -1; });
glMemoryBarrier(GL_SHADER_STORAGE_BARRIER_BIT);
ON_GL_ERROR({ return -1; });
int32_t remaining;
bool success=GpuElint_gpuToHost(ge,&remaining,s->gpu_remaining,
                                0,0,sizeof(int32_t));
ON_ERROR(!success,"GpuElint_gpuToHost() failure",{ return -1; });
return remaining;
}

//----------------------------------------------------------------------------
