//----------------------------------------------------------------------------
// Copyright (c) 2015-2016 Fabrice HARROUET (ENIB)
//
// This file is part of GpuElint.
//   http://www.enib.fr/~harrouet/gpuelint.html
//
// GpuElint is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GpuElint is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with GpuElint.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "gpuElint_internals.h"

typedef struct _GpuService_ID_Histogram4D
{
GLuint program;
int groupSize;
int groupCount;
} GpuService_ID_Histogram4D;

static
bool // success
buildProgram(GpuElint *ge,
             GpuService_ID_Histogram4D *s)
{
char *src=(char *)0;
int srcSize=0;
int srcCapacity=0;
s->groupSize=GpuElint_chooseWorkGroupSize(ge,false);
s->groupCount=GpuElint_chooseWorkGroupCount(ge);
GpuElint_appendCodePreamble(ge,&src,&srcSize,&srcCapacity,
                           s->groupSize,s->groupCount);
GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
  "   VOLATILE_BUFFER(0, int   counts[];  ) \n"
  "  READ_ONLY_BUFFER(1, int   indices[]; ) \n"
  "  READ_ONLY_BUFFER(2, float valuesA[]; ) \n"
  "  READ_ONLY_BUFFER(3, float valuesB[]; ) \n"
  "  READ_ONLY_BUFFER(4, float valuesC[]; ) \n"
  "  READ_ONLY_BUFFER(5, float valuesD[]; ) \n"
  "                                         \n"
  " UNIFORM_PARAMETER(0, int   )            \n"
  " const int pulseCount=UNI_0.x;           \n"
  "                                         \n"
  " UNIFORM_PARAMETER(1, ivec4 )            \n"
  " const ivec4 binCount=UNI_1.xyzw;        \n"
  "                                         \n"
  " UNIFORM_PARAMETER(2, vec4  )            \n"
  " const vec4 factor=UNI_2.xyzw;           \n"
  "                                         \n"
  " UNIFORM_PARAMETER(3, vec4  )            \n"
  " const vec4 increment=UNI_3.xyzw;        \n"
  "                                         \n");
GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
  " void                                                           \n"
  " main(void) // ID_histogram4D()                                 \n"
  " {                                                              \n"
  " const ivec4 firstBin=ivec4(0);                                 \n"
  " const ivec4 lastBin=binCount-ivec4(1);                         \n"
  " for(int id=globalId;id<pulseCount;id+=GRID_SIZE)               \n"
  "   {                                                            \n"
  "   const int index=indices[id];                                 \n"
  "   const vec4 value=vec4(valuesA[index],                        \n"
  "                         valuesB[index],                        \n"
  "                         valuesC[index],                        \n"
  "                         valuesD[index]);                       \n"
  "   const vec4 pos=fma(value,factor,increment);                  \n"
  "   const ivec4 bin=clamp(ivec4(floor(pos)),firstBin,lastBin);   \n"
  "   const int binId=((bin.w*binCount.z+bin.z)*binCount.y+bin.y)* \n"
  "                   binCount.x+bin.x;                            \n"
  "   atomicAdd(counts[binId],1);                                  \n"
  "   }                                                            \n"
  " }                                                              \n");
s->program=GpuElint_buildProgram(ge,src);
Memory_free(src);
return s->program!=0;
}

bool // success
GpuService_ID_Histogram4D_init(GpuElint *ge)
{
GpuService_ID_Histogram4D *s=Memory_alloc(1,GpuService_ID_Histogram4D);
if(!buildProgram(ge,s))
  {
  Memory_free(s);
  return false;
  }
ge->ID_histogram4D_service=s;
return true;
}

void
GpuService_ID_Histogram4D_destroy(GpuElint *ge)
{
GpuService_ID_Histogram4D *s=ge->ID_histogram4D_service;
if(!s) return;
glDeleteProgram(s->program);
ON_GL_ERROR({});
Memory_free(s);
}

bool // success
GpuElint_ID_histogram4D(GpuElint *ge,
                        int gpuSrc_indices,
                        int pulseCount,
                        int gpuDst_counts,
                        int gpuSrc_valuesA,
                        float minValueA,
                        float maxValueA,
                        int binCountA,
                        int gpuSrc_valuesB,
                        float minValueB,
                        float maxValueB,
                        int binCountB,
                        int gpuSrc_valuesC,
                        float minValueC,
                        float maxValueC,
                        int binCountC,
                        int gpuSrc_valuesD,
                        float minValueD,
                        float maxValueD,
                        int binCountD)
{
GPUELINT_API_INVOCATION_CHECK(ge,{ return false; });
const GpuService_ID_Histogram4D *s=ge->ID_histogram4D_service;
// binCount*(v-minValue)/(maxValue-minValue)
const float factorA=binCountA/(maxValueA-minValueA);
const float factorB=binCountB/(maxValueB-minValueB);
const float factorC=binCountC/(maxValueC-minValueC);
const float factorD=binCountC/(maxValueD-minValueD);
const float incrementA=-minValueA*factorA;
const float incrementB=-minValueB*factorB;
const float incrementC=-minValueC*factorC;
const float incrementD=-minValueD*factorD;
bool success=GpuElint_zeroFill(ge,gpuDst_counts,
                               0,binCountA*binCountB*binCountC*binCountD);
ON_ERROR(!success,"GpuElint_zeroFill() failure",{ return false; });
glUseProgram(s->program);
ON_GL_ERROR({ return false; });
glBindBufferBase(GL_SHADER_STORAGE_BUFFER,0,gpuDst_counts);
ON_GL_ERROR({ return false; });
glBindBufferBase(GL_SHADER_STORAGE_BUFFER,1,gpuSrc_indices);
ON_GL_ERROR({ return false; });
glBindBufferBase(GL_SHADER_STORAGE_BUFFER,2,gpuSrc_valuesA);
ON_GL_ERROR({ return false; });
glBindBufferBase(GL_SHADER_STORAGE_BUFFER,3,gpuSrc_valuesB);
ON_GL_ERROR({ return false; });
glBindBufferBase(GL_SHADER_STORAGE_BUFFER,4,gpuSrc_valuesC);
ON_GL_ERROR({ return false; });
glBindBufferBase(GL_SHADER_STORAGE_BUFFER,5,gpuSrc_valuesD);
ON_GL_ERROR({ return false; });
glUniform1i(0,pulseCount);
ON_GL_ERROR({ return false; });
glUniform4i(1,binCountA,binCountB,binCountC,binCountD);
ON_GL_ERROR({ return false; });
glUniform4f(2,factorA,factorB,factorC,factorD);
ON_GL_ERROR({ return false; });
glUniform4f(3,incrementA,incrementB,incrementC,incrementD);
ON_GL_ERROR({ return false; });
glDispatchCompute(s->groupCount,1,1);
ON_GL_ERROR({ return false; });
glMemoryBarrier(GL_SHADER_STORAGE_BARRIER_BIT);
ON_GL_ERROR({ return false; });
return true;
}

//----------------------------------------------------------------------------
