//----------------------------------------------------------------------------
// Copyright (c) 2015-2016 Fabrice HARROUET (ENIB)
//
// This file is part of GpuElint.
//   http://www.enib.fr/~harrouet/gpuelint.html
//
// GpuElint is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GpuElint is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with GpuElint.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "gpuElint_internals.h"

typedef struct _GpuServiceModes1D
{
GLuint program;
int groupSize;
int groupCount;
int gpu_resultCount;
} GpuServiceModes1D;

static
bool // success
buildProgram(GpuElint *ge,
             GpuServiceModes1D *s)
{
char *src=(char *)0;
int srcSize=0;
int srcCapacity=0;
s->groupSize=GpuElint_chooseWorkGroupSize(ge,false);
s->groupCount=GpuElint_chooseWorkGroupCount(ge);
GpuElint_appendCodePreamble(ge,&src,&srcSize,&srcCapacity,
                            s->groupSize,s->groupCount);
GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
  " struct HistoMode1D                            \n"
  " {                                             \n"
  " int maxCount,weight;                          \n"
  " int firstBin,lastBin;                         \n"
  " };                                            \n"
  "                                               \n"
  "   VOLATILE_BUFFER(0, int resultCount;       ) \n"
  " WRITE_ONLY_BUFFER(1, HistoMode1D results[]; ) \n"
  "  READ_ONLY_BUFFER(2, int counts[];          ) \n"
  "                                               \n"
  " UNIFORM_PARAMETER(0, float )                  \n"
  " const float cutoffRatio=UNI_0.x;              \n"
  "                                               \n"
  " UNIFORM_PARAMETER(1, ivec3 )                  \n"
  " const int modeCapacity=UNI_1.x,               \n"
  "           binCount    =UNI_1.y;               \n"
  " const bool cyclic     =UNI_1.z!=0;            \n"
  "                                               \n");
GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
  " void                                                   \n"
  " main(void) // modes1D()                                \n"
  " {                                                      \n"
  " for(int bin=globalId;bin<binCount;bin+=GRID_SIZE)      \n"
  "   {                                                    \n"
  "   int count=counts[bin];                               \n"
  "   if(count==0) { continue; }                           \n"
  "   HistoMode1D m={ count, count, bin, bin };            \n"
  "   const int cutoff=int(floor(cutoffRatio*m.maxCount)); \n"
  "   bool giveUp=false;                                   \n"
  "   int stop=cyclic ? bin : binCount-1;                  \n"
  "   int prevCount=m.maxCount;                            \n"
  "   bool lowEnough=false;                                \n"
  "   for(int i=bin;(i=(i+binCount-1)%%binCount)!=stop;)   \n"
  "     {                                                  \n"
  "     count=counts[i];                                   \n"
  "     if(count>m.maxCount) { giveUp=true; break; }       \n"
  "     if(count<=cutoff) { lowEnough=true; }              \n"
  "     if((count==0)||((count>=prevCount)&&lowEnough))    \n"
  "       { break; }                                       \n"
  "     m.weight+=count;                                   \n"
  "     m.firstBin=i;                                      \n"
  "     prevCount=count;                                   \n"
  "     }                                                  \n"
  "   if(giveUp) { continue; }                             \n"
  "   stop=cyclic ? bin : 0;                               \n"
  "   prevCount=m.maxCount;                                \n"
  "   lowEnough=false;                                     \n"
  "   for(int i=bin;(i=(i+1)%%binCount)!=stop;)            \n"
  "     {                                                  \n"
  "     count=counts[i];                                   \n"
  "     if(count>m.maxCount) { giveUp=true; break; }       \n"
  "     if(count<=cutoff) { lowEnough=true; }              \n"
  "     if((count==0)||((count>=prevCount)&&lowEnough))    \n"
  "       { break; }                                       \n"
  "     m.weight+=count;                                   \n"
  "     m.lastBin=i;                                       \n"
  "     prevCount=count;                                   \n"
  "     }                                                  \n"
  "   if(giveUp) { continue; }                             \n"
  "   const int pos=atomicAdd(resultCount,1);              \n"
  "   if(pos<modeCapacity) { results[pos]=m; }             \n"
  "   }                                                    \n"
  " }                                                      \n");
s->program=GpuElint_buildProgram(ge,src);
Memory_free(src);
return s->program!=0;
}

bool // success
GpuServiceModes1D_init(GpuElint *ge)
{
GpuServiceModes1D *s=Memory_alloc(1,GpuServiceModes1D);
if(!buildProgram(ge,s))
  {
  Memory_free(s);
  return false;
  }
s->gpu_resultCount=GpuElint_gpuAlloc(ge,sizeof(int32_t));
ON_ERROR(!s->gpu_resultCount,"GpuElint_gpuAlloc() failure",
         {
         glDeleteProgram(s->program);
         Memory_free(s);
         return false;
         });
ge->modes1D_service=s;
return true;
}

void
GpuServiceModes1D_destroy(GpuElint *ge)
{
GpuServiceModes1D *s=ge->modes1D_service;
if(!s) return;
bool success=GpuElint_gpuFree(ge,s->gpu_resultCount);
ON_ERROR(!success,"GpuElint_gpuFree() failure",{});
glDeleteProgram(s->program);
ON_GL_ERROR({});
Memory_free(s);
}

int // number of modes (-1 on failure)
GpuElint_modes1D(GpuElint *ge,
                 HistoMode1D *hostDst_modes,
                 int modeCapacity,
                 int gpuSrc_counts,
                 float cutoffRatio,
                 int binCount,
                 bool cyclic)
{
GPUELINT_API_INVOCATION_CHECK(ge,{ return -1; });
const GpuServiceModes1D *s=ge->modes1D_service;
int gpu_results=GpuElint_gpuBackend(ge,modeCapacity*sizeof(HistoMode1D));
ON_ERROR(!gpu_results,"GpuElint_gpuBackend() failure",{ return -1; });
int32_t modeCount=0;
bool success=GpuElint_hostToGpu(ge,s->gpu_resultCount,&modeCount,
                                0,0,sizeof(int32_t));
ON_ERROR(!success,"GpuElint_hostToGpu() failure",{ return -1; });
glUseProgram(s->program);
ON_GL_ERROR({ return -1; });
glBindBufferBase(GL_SHADER_STORAGE_BUFFER,0,s->gpu_resultCount);
ON_GL_ERROR({ return -1; });
glBindBufferBase(GL_SHADER_STORAGE_BUFFER,1,gpu_results);
ON_GL_ERROR({ return -1; });
glBindBufferBase(GL_SHADER_STORAGE_BUFFER,2,gpuSrc_counts);
ON_GL_ERROR({ return -1; });
glUniform1f(0,cutoffRatio);
ON_GL_ERROR({ return -1; });
glUniform3i(1,modeCapacity,binCount,cyclic);
ON_GL_ERROR({ return -1; });
glDispatchCompute(s->groupCount,1,1);
ON_GL_ERROR({ return -1; });
glMemoryBarrier(GL_SHADER_STORAGE_BARRIER_BIT);
ON_GL_ERROR({ return -1; });
success=GpuElint_gpuToHost(ge,&modeCount,s->gpu_resultCount,
                           0,0,sizeof(int32_t));
ON_ERROR(!success,"GpuElint_gpuToHost() failure",{ return -1; });
if(modeCount)
  {
  const int copyCount=modeCount>modeCapacity ? modeCapacity : modeCount;
  success=GpuElint_gpuToHost(ge,hostDst_modes,gpu_results,
                             0,0,copyCount*sizeof(HistoMode1D));
  ON_ERROR(!success,"GpuElint_gpuToHost() failure",{ return -1; });
  // FIXME: if the maximal count appears many times in a mode, this mode
  //        will then be reported many times.
  //        Would it be worth filtering to remove duplicates?
  }
return modeCount;
}

//----------------------------------------------------------------------------
