//----------------------------------------------------------------------------
// Copyright (c) 2015-2016 Fabrice HARROUET (ENIB)
//
// This file is part of GpuElint.
//   http://www.enib.fr/~harrouet/gpuelint.html
//
// GpuElint is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GpuElint is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with GpuElint.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

using System;
using System.Threading;
using System.IO;
using System.Drawing;
using OpenTK;
using OpenTK.Graphics;
using OpenTK.Graphics.OpenGL;
using OpenTK.Input;

//----------------------------------------------------------------------------
// Application-specific code
//----------------------------------------------------------------------------

internal class Application
{
const int HISTO_A=0;
const int HISTO_B=3;

static string[] propertyName={ "DI", "FN", "NIV", "AZIMUT" };
static int[] propertyBinCount={ 128,  32,   32,    128      };
static string NAME_A=propertyName[HISTO_A];
static string NAME_B=propertyName[HISTO_B];
static int BIN_COUNT_A=propertyBinCount[HISTO_A];
static int BIN_COUNT_B=propertyBinCount[HISTO_B];

const int MAX_HISTO_MODE_COUNT=128;
const float HISTO_CUTOFF_RATIO=0.1f;
const float HIGHEST_PRF=100000.0f;
const float LOWEST_PRF=100.0f;
const float DTOA_RATIO=1.025f;
const float DTOA_DETECTION_THRESHOLD=0.01f;
const int MIN_MATCH_COUNT=4;

//-- properties set by execution environment --
internal GameWindow game=null;
internal int width=0,height=0;
internal string fileName=null;

//-- simple pass-through vertex-shader source code --
string vertexSrc=
" #version 430                                                            \n"+
" in vec2 v_Vertex;                                                       \n"+
" in int v_Mark;                                                          \n"+
" flat out int f_Mark;                                                    \n"+
" uniform vec2 u_Delta;                                                   \n"+
" void                                                                    \n"+
" main(void)                                                              \n"+
" {                                                                       \n"+
" f_Mark=v_Mark;                                                          \n"+
" gl_Position=vec4(v_Vertex+u_Delta,0.0,1.0);                             \n"+
" }                                                                       \n";

//-- simple unicolour fragment-shader source code --
string fragmentSrc=
" #version 430                                                            \n"+
" layout(location=0) out vec4 defaultFragColour;                          \n"+
" #define gl_FragColor defaultFragColour                                  \n"+
" flat in int f_Mark;                                                     \n"+
" void                                                                    \n"+
" main(void)                                                              \n"+
" {                                                                       \n"+
" gl_FragColor=f_Mark<0 ? vec4(0.2,0.8,0.2,1.0)                           \n"+
"                       : vec4(0.8,0.2,0.2,1.0);                          \n"+
" }                                                                       \n";

//-- shader-program and parameters --
int program;
int v_Vertex;
int v_Mark;
int u_Delta;

//-- informations required to draw a simple shape --
int vertexCount;
int vertex_vbo;
int mark_vbo;
int vao;
volatile int stepCount;
volatile int redrawCount;

//-- GPU execution environment --
GpuElint ge;

internal
bool // success
init()
  {
  try
    {
    ge=new GpuElint(null,true);
    }
  catch(Exception e)
    {
    Console.Write("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n");
    Console.Write("~~~~ GpuElint initialisation failure ~~~~\n");
    Console.Write("{0}\n",e);
    Console.Write("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n");
    return false;
    }
  GpuElint.Properties properties=ge.properties();
  Console.Write("DTOA_SKIP_COUNT={0}\n",
                properties.DTOA_SKIP_COUNT);
  Console.Write("MAX_MATCH_COUNT={0}\n",
                properties.MAX_MATCH_COUNT);
  Console.Write("glVersionMajor={0}\n",
                properties.glVersionMajor);
  Console.Write("glVersionMinor={0}\n",
                properties.glVersionMinor);
  Console.Write("glslVersion={0}\n",
                properties.glslVersion);
  Console.Write("maxWorkGroupCountX={0}\n",
                properties.maxWorkGroupCountX);
  Console.Write("maxWorkGroupCountY={0}\n",
                properties.maxWorkGroupCountY);
  Console.Write("maxWorkGroupCountZ={0}\n",
                properties.maxWorkGroupCountZ);
  Console.Write("maxWorkGroupSizeX={0}\n",
                properties.maxWorkGroupSizeX);
  Console.Write("maxWorkGroupSizeY={0}\n",
                properties.maxWorkGroupSizeY);
  Console.Write("maxWorkGroupSizeZ={0}\n",
                properties.maxWorkGroupSizeZ);
  Console.Write("maxWorkGroupInvocations={0}\n",
                properties.maxWorkGroupInvocations);
  Console.Write("maxUniformBlocks={0}\n",
                properties.maxUniformBlocks);
  Console.Write("maxTextureImageUnits={0}\n",
                properties.maxTextureImageUnits);
  Console.Write("maxAtomicCounterBuffers={0}\n",
                properties.maxAtomicCounterBuffers);
  Console.Write("maxAtomicCounters={0}\n",
                properties.maxAtomicCounters);
  Console.Write("maxSharedMemorySize={0}\n",
                properties.maxSharedMemorySize);
  Console.Write("maxUniformComponents={0}\n",
                properties.maxUniformComponents);
  Console.Write("maxImageUniforms={0}\n",
                properties.maxImageUniforms);
  Console.Write("maxCombinedUniformComponents={0}\n",
                properties.maxCombinedUniformComponents);
  Console.Write("maxShaderStorageBlocks={0}\n",
                properties.maxShaderStorageBlocks);
  Console.Write("warpSize={0}\n",
                properties.warpSize);
  Console.Write("warpsPerStreamingMultiprocessor={0}\n",
                properties.warpsPerStreamingMultiprocessor);
  Console.Write("streamingMultiprocessorCount={0}\n",
                properties.streamingMultiprocessorCount);
  Console.Write("hasBallot={0}\n",
                properties.hasBallot);
  Console.Write("hasShuffle={0}\n",
                properties.hasShuffle);

  //-- detect platform-specific features --
  Console.Write("OpenGL version: {0}\n",
                GL.GetString(StringName.Version));
  Console.Write("          GLSL: {0}\n",
                GL.GetString(StringName.ShadingLanguageVersion));
  Console.Write("        vendor: {0}\n",
                GL.GetString(StringName.Vendor));
  Console.Write("      renderer: {0}\n",
                GL.GetString(StringName.Renderer));

  //-- load data from file --
  BinaryReader f=new BinaryReader(File.Open(fileName,FileMode.Open));
  int pulseCount=f.ReadInt32();
  f.BaseStream.Position=sizeof(int);
  double[] host_toa64=new double[pulseCount];
  for(int i=0;i<pulseCount;++i)
    { host_toa64[i]=f.ReadDouble(); }
  f.BaseStream.Position=sizeof(int)+
                        pulseCount*(sizeof(double)+HISTO_A*sizeof(float));
  float[] host_valuesA=new float[pulseCount];
  for(int i=0;i<pulseCount;++i)
    { host_valuesA[i]=f.ReadSingle(); }
  f.BaseStream.Position=sizeof(int)+
                        pulseCount*(sizeof(double)+HISTO_B*sizeof(float));
  float[] host_valuesB=new float[pulseCount];
  for(int i=0;i<pulseCount;++i)
    { host_valuesB[i]=f.ReadSingle(); }
  f.Close();

  //-- change initial value for some OpenGL settings --
  GL.ClearColor(0.3f,0.3f,0.3f,1.0f);

  //-- create shader-program and access its parameters --
  int vs=GL.CreateShader(ShaderType.VertexShader);
  GL.ShaderSource(vs,vertexSrc);
  GL.CompileShader(vs);
  int fs=GL.CreateShader(ShaderType.FragmentShader);
  GL.ShaderSource(fs,fragmentSrc);
  GL.CompileShader(fs);
  program=GL.CreateProgram();
  GL.AttachShader(program,vs);
  GL.AttachShader(program,fs);
  GL.LinkProgram(program);
  GL.DeleteShader(vs);
  GL.DeleteShader(fs);
  v_Vertex=GL.GetAttribLocation(program,"v_Vertex");
  v_Mark=GL.GetAttribLocation(program,"v_Mark");
  u_Delta=GL.GetUniformLocation(program,"u_Delta");

  //-- compute vertices --
  vertexCount=pulseCount;
  float[] tmpV=new float[2*vertexCount];
  int side=(int)Math.Sqrt(vertexCount+1.0);
  double factor=1.8/side;
  double increment=-0.9;
  for(int i=0;i<vertexCount;++i)
    {
    tmpV[2*i+0]=(float)(factor*(i%side)+increment);
    tmpV[2*i+1]=(float)(factor*(i/side)+increment);
    }

  //-- create and fill buffers with vertices --
  GL.GenBuffers(1,out vertex_vbo);
  GL.BindBuffer(BufferTarget.ArrayBuffer,vertex_vbo);
  GL.BufferData(BufferTarget.ArrayBuffer,
                (IntPtr)(tmpV.Length*sizeof(float)),tmpV,
                BufferUsageHint.StaticDraw);
  mark_vbo=ge.gpuAlloc((IntPtr)(vertexCount*sizeof(int)));
  ge.countFill(mark_vbo,0,vertexCount,-1,0);

  //-- create and set vertex-array up --
  GL.GenVertexArrays(1,out vao);
  GL.BindVertexArray(vao);
  //-- associate vertex-array attribute and buffer --
  GL.EnableVertexAttribArray(v_Vertex);
  GL.BindBuffer(BufferTarget.ArrayBuffer,vertex_vbo);
  GL.VertexAttribPointer(v_Vertex,2,VertexAttribPointerType.Float,false,0,0);
  GL.EnableVertexAttribArray(v_Mark);
  GL.BindBuffer(BufferTarget.ArrayBuffer,mark_vbo);
  GL.VertexAttribPointer(v_Mark,1,VertexAttribPointerType.Int,false,0,0);
  GL.BindVertexArray(0); // avoid accidental change

  //-- start an asynchronous computation thread if requested --
  GL.Finish(); // ensure graphics initialisation is done
  stepCount=0;
  redrawCount=0;
  Console.Write("----> launching an asynchronous computation thread\n");
  Thread th=new Thread(delegate()
    { computeThread(pulseCount,host_toa64,host_valuesA,host_valuesB); });
  th.IsBackground=true;
  th.Start();
  return true;
  }

void
computeThread(int pulseCount,
              double[] host_toa64,
              float[] host_valuesA,
              float[] host_valuesB)
  {
  GpuElint ge=null;
  try
    {
    ge=new GpuElint(this.ge,true);
    }
  catch(Exception e)
    {
    Console.Write("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n");
    Console.Write("~~~~ shared GpuElint initialisation failure ~~~~\n");
    Console.Write("{0}\n",e);
    Console.Write("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n");
    }
  for(;;)
    {
    for(int i=0;i<100;++i)
      { Thread.Sleep(10); step(); }
    if(ge!=null)
      { compute(ge,pulseCount,host_toa64,host_valuesA,host_valuesB); }
    }
  }

void
step()
  {
  lock(this) { ++stepCount; Monitor.PulseAll(this); }
  }

int
toaAnalysis(GpuElint ge,
            int indices,
            int pulseCount,
            int toa64,
            int marks,
            ref int seqMark,
            float minDtoa,
            int dtoaBinCount,
            int[] dtoaCounts,
            float[] dtoaSums,
            float minDuration)
  {
  int prevSeqMark=seqMark;
  float priRatioThreshold=DTOA_RATIO*DTOA_RATIO;
  float exclusivePriLowerLimit=0.0f;
  for(;;)
    {
    if(exclusivePriLowerLimit==0.0f)
      {
      ge.dtoaHistogram(indices,pulseCount,toa64,
                       dtoaCounts,dtoaSums,dtoaBinCount,
                       minDtoa,DTOA_RATIO);
      step();
      }
    float estimatedPri=ge.estimateMajorPri(dtoaCounts,dtoaSums,dtoaBinCount,
                                           minDtoa,DTOA_RATIO,
                                           exclusivePriLowerLimit,minDuration);
    step();
    if(estimatedPri==0.0f) { break; }
    Console.Write("estimated PRI: {0} ({1} Hz)\n",
                  estimatedPri,1.0f/estimatedPri);
    float pri=ge.markPulses(indices,pulseCount,toa64,
                            marks,seqMark,estimatedPri,
                            priRatioThreshold,MIN_MATCH_COUNT);
    step();
    if(pri>0.0f)
      {
      int remaining=ge.filterUnmarked(indices,pulseCount,indices,marks);
      step();
      int extractedCount=pulseCount-remaining;
      pulseCount=remaining; // don't consider marked pulses anymore
      Console.Write("[{0}] ---> found {1} pulses for {2} Hz\n",
                    seqMark,extractedCount,1.0f/pri);
      ++seqMark;
      exclusivePriLowerLimit=0.0f;
      }
    else
      { exclusivePriLowerLimit=estimatedPri; }
    }
  return seqMark-prevSeqMark;
  }

void
compute(GpuElint ge,
        int pulseCount,
        double[] host_toa64,
        float[] host_valuesA,
        float[] host_valuesB)
  {
  Console.Write("#######################################"+
                "#######################################\n");
  Console.Write("## gpu analysis of {0} pulses\n",pulseCount);

  //-- send data to GPU --
  int toa64=ge.gpuAlloc((IntPtr)(pulseCount*sizeof(double)));
  ge.hostToGpu(toa64,host_toa64,
               IntPtr.Zero,IntPtr.Zero,(IntPtr)(pulseCount*sizeof(double)));
  int valuesA=ge.gpuAlloc((IntPtr)(pulseCount*sizeof(float)));
  ge.hostToGpu(valuesA,host_valuesA,
               IntPtr.Zero,IntPtr.Zero,(IntPtr)(pulseCount*sizeof(float)));
  int valuesB=ge.gpuAlloc((IntPtr)(pulseCount*sizeof(float)));
  ge.hostToGpu(valuesB,host_valuesB,
               IntPtr.Zero,IntPtr.Zero,(IntPtr)(pulseCount*sizeof(float)));
  step();

  //-- prepare ancillary data --
  float minDuration=((float)(host_toa64[pulseCount-1]-host_toa64[0]))*
                    DTOA_DETECTION_THRESHOLD;
  GpuElint.HistoMode2D[] modes=new GpuElint.HistoMode2D[MAX_HISTO_MODE_COUNT];
  int histo=ge.gpuAlloc((IntPtr)(BIN_COUNT_A*BIN_COUNT_B*sizeof(int)));
  float minDtoa=1.0f/HIGHEST_PRF;
  float maxDtoa=1.0f/LOWEST_PRF;
  int dtoaBinCount=(int)Math.Ceiling((Math.Log(maxDtoa)-Math.Log(minDtoa))/
                                     Math.Log(DTOA_RATIO));
  int[] dtoaCounts=new int[dtoaBinCount];
  float[] dtoaSums=new float[dtoaBinCount];
  int indices=ge.gpuAlloc((IntPtr)(pulseCount*sizeof(int)));
  int toaIndices=ge.gpuAlloc((IntPtr)(pulseCount*sizeof(int)));
  ge.countFill(indices,0,pulseCount,0,1);
  int marks=mark_vbo; // shared with rendering
  ge.countFill(marks,0,pulseCount,-1,0);
  ge.flush(true);
  step();
  int seqMark=0;
  for(;;)
    {
    Console.Write("---------------------------------------"+
                  "---------------------------------------\n");
    int prevSeqMark=seqMark;
    float minValueA,maxValueA;
    bool cyclicA=false;
    if(NAME_A.Equals("AZIMUT"))
      { minValueA=0.0f; maxValueA=360.0f; cyclicA=true; }
    else
      {
      int minId,maxId;
      ge.minMax(indices,pulseCount,valuesA,
                out minValueA,out minId,out maxValueA,out maxId);
      step();
      }
    float minValueB,maxValueB;
    bool cyclicB=false;
    if(NAME_B.Equals("AZIMUT"))
      { minValueB=0.0f; maxValueB=360.0f; cyclicB=true; }
    else
      {
      int minId,maxId;
      ge.minMax(indices,pulseCount,valuesB,
                out minValueB,out minId,out maxValueB,out maxId);
      step();
      }
    ge.histogram2D(indices,pulseCount,histo,
                   valuesA,minValueA,maxValueA,BIN_COUNT_A,
                   valuesB,minValueB,maxValueB,BIN_COUNT_B);
    step();
    int modeCount=ge.modes2D(modes,histo,HISTO_CUTOFF_RATIO,
                             BIN_COUNT_A,cyclicA,
                             BIN_COUNT_B,cyclicB);
    step();
    Console.Write("{0} modes found in {1}/{2} histogram\n",
                  modeCount,NAME_A,NAME_B);
    if(modeCount>modes.Length)
      {
      Console.Write("MAX_HISTO_MODE_COUNT exceeded ({0})\n",modeCount);
      modeCount=modes.Length;
      }
    step();
    for(;;)
      {
      bool lastAttempt=(modeCount<=1);
      int toaPulseCount;
      if(lastAttempt) // consider every remaining pulse
        {
        ge.gpuToGpu(toaIndices,indices,
                    IntPtr.Zero,IntPtr.Zero,(IntPtr)(pulseCount*sizeof(int)));
        step();
        toaPulseCount=pulseCount;
        Console.Write("analysing {0} remaining pulses\n",toaPulseCount);
        }
      else // consider only the best mode
        {
        int bestMode=0;
        for(int i=1;i<modeCount;++i)
          { if(modes[i].weight>modes[bestMode].weight) { bestMode=i; } }
        float histoStepA=(maxValueA-minValueA)/BIN_COUNT_A;
        float minFilterA=minValueA+(modes[bestMode].firstBinA+0.0f)*histoStepA;
        float maxFilterA=minValueA+(modes[bestMode].lastBinA+1.0f)*histoStepA;
        float histoStepB=(maxValueB-minValueB)/BIN_COUNT_B;
        float minFilterB=minValueB+(modes[bestMode].firstBinB+0.0f)*histoStepB;
        float maxFilterB=minValueB+(modes[bestMode].lastBinB+1.0f)*histoStepB;
        toaPulseCount=ge.filter2D(indices,pulseCount,toaIndices,
                                  valuesA,minFilterA,maxFilterA,
                                  valuesB,minFilterB,maxFilterB);
        step();
        Console.Write("analysing {0} pulses out of {1}"+
                      " ({2}/{3} mode from bin {4}/{5} to {6}/{7})\n",
                      toaPulseCount,pulseCount,NAME_A,NAME_B,
                      modes[bestMode].firstBinA,modes[bestMode].firstBinB,
                      modes[bestMode].lastBinA,modes[bestMode].lastBinB);
        modes[bestMode]=modes[--modeCount]; // forget this mode
        }
      int seqCount=toaAnalysis(ge,toaIndices,toaPulseCount,toa64,
                               marks,ref seqMark,minDtoa,dtoaBinCount,
                               dtoaCounts,dtoaSums,minDuration);
      if(seqCount!=0)
        {
        // don't consider marked pulses anymore
        pulseCount=ge.filterUnmarked(indices,pulseCount,indices,marks);
        break; // histogram will be rebuilt
        }
      if(lastAttempt) { break; }
      }
    if(prevSeqMark==seqMark)
      { break; } // nothing was found with this histogram
    }
  //-- cleanup --
  // marks is mark_vbo (shared with rendering)
  ge.gpuFree(toaIndices);
  ge.gpuFree(indices);
  ge.gpuFree(histo);
  ge.gpuFree(valuesB);
  ge.gpuFree(valuesA);
  ge.gpuFree(toa64);
  }

internal
void
redraw()
  {
  //-- wait for next computation step change --
  lock(this)
    {
    while(stepCount==redrawCount) { Monitor.Wait(this); }
    redrawCount=stepCount;
    }

  //-- animate drawing --
  double a=Math.PI*0.25*DateTime.Now.TimeOfDay.TotalSeconds;
  float dx=0.1f*(float)Math.Cos(a);
  float dy=0.1f*(float)Math.Sin(a);

  //-- reset framebuffer --
  GL.Clear(ClearBufferMask.ColorBufferBit);
  GL.Viewport(0,0,width,height);

  //-- draw vertex-array using shader-program --
  GL.UseProgram(program);
  GL.Uniform2(u_Delta,dx,dy);
  GL.BindVertexArray(vao);
  GL.DrawArrays(PrimitiveType.Points,0,vertexCount);
  GL.Finish();
  }

internal
void
keyboard(KeyPressEventArgs e)
  {
  }
}

class MyApplication
{
public static
void
Main(string[] args)
  {
  string[] argv=Environment.GetCommandLineArgs();
  if(argv.Length!=2)
    {
    Console.Write("usage: {0} pulse_file.raw\n",argv[0]);
    Environment.Exit(1);
    }
  Application application=new Application();
  application.fileName=argv[1];
  string title="C#/OpenTK GpuElint test application ("+
               application.fileName+")";
  using(GameWindow game=new GameWindow(800,600,GraphicsMode.Default,title))
    {
    game.Load+=(sender,e)=>
      {
      // postpone initialisation until initial size is known
      if((application.game==null)&&
         (application.width>0)&&
         (application.height>0))
        {
        application.game=game;
        application.init();
        if(!application.init()) { game.Exit(); }
        }
      };
    game.Resize+=(sender,e)=>
      {
      application.width=game.Width;
      application.height=game.Height;
      // initialisation may have been postponed
      if(application.game==null)
        {
        application.game=game;
        if(!application.init()) { game.Exit(); }
        }
      };
    game.RenderFrame+=(sender,e)=>
      {
      if(application.game!=null)
        {
        application.redraw();
        game.SwapBuffers();
        }
      };
    game.KeyPress+=(sender,e)=>
      {
      application.keyboard(e);
      };
    game.KeyDown+=(sender,e)=>
      {
      if(e.Key==Key.Escape) { game.Exit(); }
      };
    game.Run(25.0);
    }
  }
}

//----------------------------------------------------------------------------
