//----------------------------------------------------------------------------
// Copyright (c) 2015-2016 Fabrice HARROUET (ENIB)
//
// This file is part of GpuElint.
//   http://www.enib.fr/~harrouet/gpuelint.html
//
// GpuElint is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GpuElint is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with GpuElint.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "common.h"

static int gm[3]; // GPU markers
#define GPU_MARKER_COUNT ((int)(sizeof(gm)/sizeof(*gm)))
static double init_time;
static double minMax_time;
static double histogram_time;
static double histogram_gpuTime;
static double modes_time;
static double modes_gpuTime;
static double filter_time;
static double filter_gpuTime;
static double dtoaHistogram_time;
static double estimatePri_time;
static double markPulses_time;
static double filterUnmarked_time;
static double filterUnmarked_gpuTime;

// very ugly, for display purpose only, don't use in normal conditions!
static
int
markCount(GpuElint *ge,
          int gpu_marks,
          int pulseCount,
          int mark)
{
#if SHOW_TRACE
  int *t=Memory_alloc(pulseCount,int);
  GpuElint_gpuToHost(ge,t,gpu_marks,0,0,pulseCount*sizeof(int32_t));
  int count=0;
  for(int i=0;i<pulseCount;++i)
    {
    if(t[i]==mark) ++count;
    }
  Memory_free(t);
  return count;
#else
  (void)ge;
  (void)gpu_marks;
  (void)pulseCount;
  (void)mark;
  return 0;
#endif
}

static
bool // cyclic
detectBounds(GpuElint *ge,
             PropertyKind property,
             int bitmask,
             int pulseCount,
             int values,
             float *out_minValue,
             float *out_maxValue)
{
if(property==AZIMUT_PROP)
  {
  *out_minValue=0.0f;
  *out_maxValue=360.0f;
  return true;
  }
else
  {
  int minId,maxId;
  const float t0=GET_TIME(ge);
  GpuElint_BM_minMax(ge,bitmask,pulseCount,values,
                     out_minValue,&minId,out_maxValue,&maxId);
  minMax_time+=(GET_TIME(ge)-t0);
  return false;
  }
}

static
int // new marks detected
toaAnalysis(GpuElint *ge,
            int bitmask,
            int pulseCount,
            int toa64,
            int marks,
            int *inout_seqMark,
            Sequence *sequences,
            float minDtoa,
            int dtoaBinCount,
            int *dtoaCounts,
            float *dtoaSums,
            float minDuration)
{
int prevSeqMark=*inout_seqMark;
const float priRatioThreshold=DTOA_RATIO*DTOA_RATIO;
float exclusivePriLowerLimit=0.0f;
bool firstHisto=true;
for(;;)
  {
  if(!exclusivePriLowerLimit)
    {
    const float t0=GET_TIME(ge);
    GpuElint_BM_dtoaHistogram(ge,bitmask,pulseCount,toa64,
                              dtoaCounts,dtoaSums,dtoaBinCount,
                              minDtoa,DTOA_RATIO);
    dtoaHistogram_time+=(GET_TIME(ge)-t0);
    if(firstHisto)
      {
      firstHisto=false;
      showDtoaHistogram(dtoaCounts,dtoaSums,false,
                        dtoaBinCount,minDtoa,DTOA_RATIO);
      }
    }
  const float t0=GET_TIME(ge);
  const float estimatedPri=GpuElint_estimateMajorPri(ge,
                           dtoaCounts,dtoaSums,dtoaBinCount,
                           minDtoa,DTOA_RATIO,exclusivePriLowerLimit,
                           minDuration);
  estimatePri_time+=(GET_TIME(ge)-t0);
  if(!estimatedPri)
    { break; }
  trace("estimated PRI: %g (%.3f Hz)\n",
        estimatedPri,1.0f/estimatedPri);
  const float t1=GET_TIME(ge);
  const float pri=GpuElint_BM_markPulses(ge,bitmask,pulseCount,toa64,
                                         marks,*inout_seqMark,estimatedPri,
                                         priRatioThreshold,MIN_MATCH_COUNT);
  const float t2=GET_TIME(ge);
  markPulses_time+=(t2-t1);
  if(pri>0.0f)
    {
    SET_GPU_MARKER(ge,gm[0]);
    GpuElint_BM_filterUnmarked(ge,bitmask,pulseCount,
                               bitmask,marks);
    SET_GPU_MARKER(ge,gm[1]);
    filterUnmarked_gpuTime+=GPU_MARKER_INTERVAL(ge,gm[0],gm[1]);
    filterUnmarked_time+=(GET_TIME(ge)-t2);

    const int newMark=(*inout_seqMark)++;
    trace("[%3d] ---> found %d pulses for %.3f Hz\n",
          newMark,markCount(ge,marks,pulseCount,newMark),1.0f/pri);
    if(newMark<MAX_SEQUENCE_COUNT)
      {
      sequences[newMark].pri=pri;
      sequences[newMark].pulseCount=0; // retrieved later
      }
    exclusivePriLowerLimit=0.0f;
    }
  else
    {
    exclusivePriLowerLimit=estimatedPri;
    }
  }
return (*inout_seqMark)-prevSeqMark;
}

int // number of extracted sequences
BM_gpuAnalysis(GpuElint *ge,
               const char *fileName,
               int pulseCount,
               int *host_marks,
               Sequence *sequences)
{
for(int i=0;i<GPU_MARKER_COUNT;++i)
  { gm[i]=GpuElint_createMarker(ge); }
const float startTime=GpuElint_elapsed(ge);
printf("## bitmask-based GPU analysis of %d pulses\n",pulseCount);
int toa64=GpuElint_gpuAlloc(ge,pulseCount*sizeof(real64_t));
real64_t *map_toa64=(real64_t *)GpuElint_beginMapping(
                    ge,toa64,0,pulseCount*sizeof(real64_t),false,true);
loadPulseToa(fileName,map_toa64,pulseCount);
const float minDuration=((float)(map_toa64[pulseCount-1]-map_toa64[0]))*
                        DTOA_DETECTION_THRESHOLD;
GpuElint_endMapping(ge,toa64);
int valuesA=GpuElint_gpuAlloc(ge,pulseCount*sizeof(real32_t));
real32_t *map_valuesA=(real32_t *)GpuElint_beginMapping(
                      ge,valuesA,0,pulseCount*sizeof(real32_t),false,true);
loadPulseProperty(fileName,HISTO_A,map_valuesA,pulseCount);
GpuElint_endMapping(ge,valuesA);
#if defined HISTO_B
  int valuesB=GpuElint_gpuAlloc(ge,pulseCount*sizeof(real32_t));
  real32_t *map_valuesB=(real32_t *)GpuElint_beginMapping(
                        ge,valuesB,0,pulseCount*sizeof(real32_t),false,true);
  loadPulseProperty(fileName,HISTO_B,map_valuesB,pulseCount);
  GpuElint_endMapping(ge,valuesB);
#endif
#if defined HISTO_C
  int valuesC=GpuElint_gpuAlloc(ge,pulseCount*sizeof(real32_t));
  real32_t *map_valuesC=(real32_t *)GpuElint_beginMapping(
                        ge,valuesC,0,pulseCount*sizeof(real32_t),false,true);
  loadPulseProperty(fileName,HISTO_C,map_valuesC,pulseCount);
  GpuElint_endMapping(ge,valuesC);
#endif
#if defined HISTO_D
  int valuesD=GpuElint_gpuAlloc(ge,pulseCount*sizeof(real32_t));
  real32_t *map_valuesD=(real32_t *)GpuElint_beginMapping(
                        ge,valuesD,0,pulseCount*sizeof(real32_t),false,true);
  loadPulseProperty(fileName,HISTO_D,map_valuesD,pulseCount);
  GpuElint_endMapping(ge,valuesD);
#endif
HistoMode *modes=Memory_alloc(MAX_HISTO_MODE_COUNT,HistoMode);
int histo=GpuElint_gpuAlloc(ge,BIN_COUNT*sizeof(int32_t));
const float minDtoa=1.0f/HIGHEST_PRF;
const float maxDtoa=1.0f/LOWEST_PRF;
const int dtoaBinCount=(int)ceilf((logf(maxDtoa)-logf(minDtoa))/
                                  logf(DTOA_RATIO));
int *dtoaCounts=Memory_alloc(dtoaBinCount,int);
float *dtoaSums=Memory_alloc(dtoaBinCount,float);
const int bitmaskCount=(pulseCount+31)/32;
int bitmask=GpuElint_gpuAlloc(ge,bitmaskCount*sizeof(int32_t));
int toaBitmask=GpuElint_gpuAlloc(ge,bitmaskCount*sizeof(int32_t));
GpuElint_countFill(ge,bitmask,0,bitmaskCount,-1,0);
int marks=GpuElint_gpuAlloc(ge,pulseCount*sizeof(int32_t));
GpuElint_countFill(ge,marks,0,pulseCount,-1,0);
GPU_FLUSH(ge,true);
init_time+=(GET_TIME(ge)-startTime);
int seqMark=0;
for(;;)
  {
  trace("---------------------------------------"
        "---------------------------------------\n");
  const int prevSeqMark=seqMark;
  float minValueA,maxValueA;
  const bool cyclicA=detectBounds(ge,HISTO_A,bitmask,pulseCount,valuesA,
                                  &minValueA,&maxValueA);
#if defined HISTO_B
  float minValueB,maxValueB;
  const bool cyclicB=detectBounds(ge,HISTO_B,bitmask,pulseCount,valuesB,
                                  &minValueB,&maxValueB);
#endif
#if defined HISTO_C
  float minValueC,maxValueC;
  const bool cyclicC=detectBounds(ge,HISTO_C,bitmask,pulseCount,valuesC,
                                  &minValueC,&maxValueC);
#endif
#if defined HISTO_D
  float minValueD,maxValueD;
  const bool cyclicD=detectBounds(ge,HISTO_D,bitmask,pulseCount,valuesD,
                                  &minValueD,&maxValueD);
#endif
#if !defined HISTO_B
  const float t0=GET_TIME(ge);
  SET_GPU_MARKER(ge,gm[0]);
  GpuElint_BM_histogram1D(ge,bitmask,pulseCount,histo,
                          valuesA,minValueA,maxValueA,BIN_COUNT_A);
  const float t1=GET_TIME(ge);
  SET_GPU_MARKER(ge,gm[1]);
  int modeCount=GpuElint_modes1D(ge,modes,MAX_HISTO_MODE_COUNT,
                                 histo,HISTO_CUTOFF_RATIO,
                                 BIN_COUNT_A,cyclicA);
  const float t2=GET_TIME(ge);
  SET_GPU_MARKER(ge,gm[2]);
# if SHOW_TRACE
  // very ugly, for display purpose only, don't use in normal conditions!
  int32_t *host_histo=Memory_alloc(BIN_COUNT_A,int32_t);
  GpuElint_gpuToHost(ge,host_histo,histo,0,0,BIN_COUNT_A*sizeof(int32_t));
  showHistogram1D(NAME_A,host_histo,BIN_COUNT_A,minValueA,maxValueA);
  Memory_free(host_histo);
# endif
  trace("%d modes found in %s histogram\n",
        modeCount,NAME_A);
#elif !defined HISTO_C
  const float t0=GET_TIME(ge);
  SET_GPU_MARKER(ge,gm[0]);
  GpuElint_BM_histogram2D(ge,bitmask,pulseCount,histo,
                          valuesA,minValueA,maxValueA,BIN_COUNT_A,
                          valuesB,minValueB,maxValueB,BIN_COUNT_B);
  const float t1=GET_TIME(ge);
  SET_GPU_MARKER(ge,gm[1]);
  int modeCount=GpuElint_modes2D(ge,modes,MAX_HISTO_MODE_COUNT,
                                 histo,HISTO_CUTOFF_RATIO,
                                 BIN_COUNT_A,cyclicA,
                                 BIN_COUNT_B,cyclicB);
  const float t2=GET_TIME(ge);
  SET_GPU_MARKER(ge,gm[2]);
  trace("%d modes found in %s/%s histogram\n",
        modeCount,NAME_A,NAME_B);
#elif !defined HISTO_D
  const float t0=GET_TIME(ge);
  SET_GPU_MARKER(ge,gm[0]);
  GpuElint_BM_histogram3D(ge,bitmask,pulseCount,histo,
                          valuesA,minValueA,maxValueA,BIN_COUNT_A,
                          valuesB,minValueB,maxValueB,BIN_COUNT_B,
                          valuesC,minValueC,maxValueC,BIN_COUNT_C);
  const float t1=GET_TIME(ge);
  SET_GPU_MARKER(ge,gm[1]);
  int modeCount=GpuElint_modes3D(ge,modes,MAX_HISTO_MODE_COUNT,
                                 histo,HISTO_CUTOFF_RATIO,
                                 BIN_COUNT_A,cyclicA,
                                 BIN_COUNT_B,cyclicB,
                                 BIN_COUNT_C,cyclicC);
  const float t2=GET_TIME(ge);
  SET_GPU_MARKER(ge,gm[2]);
  trace("%d modes found in %s/%s/%s histogram\n",
        modeCount,NAME_A,NAME_B,NAME_C);
#else
  const float t0=GET_TIME(ge);
  SET_GPU_MARKER(ge,gm[0]);
  GpuElint_BM_histogram4D(ge,bitmask,pulseCount,histo,
                          valuesA,minValueA,maxValueA,BIN_COUNT_A,
                          valuesB,minValueB,maxValueB,BIN_COUNT_B,
                          valuesC,minValueC,maxValueC,BIN_COUNT_C,
                          valuesD,minValueD,maxValueD,BIN_COUNT_D);
  const float t1=GET_TIME(ge);
  SET_GPU_MARKER(ge,gm[1]);
  int modeCount=GpuElint_modes4D(ge,modes,MAX_HISTO_MODE_COUNT,
                                 histo,HISTO_CUTOFF_RATIO,
                                 BIN_COUNT_A,cyclicA,
                                 BIN_COUNT_B,cyclicB,
                                 BIN_COUNT_C,cyclicC,
                                 BIN_COUNT_D,cyclicD);
  const float t2=GET_TIME(ge);
  SET_GPU_MARKER(ge,gm[2]);
  trace("%d modes found in %s/%s/%s/%s histogram\n",
        modeCount,NAME_A,NAME_B,NAME_C,NAME_D);
#endif
  histogram_gpuTime+=GPU_MARKER_INTERVAL(ge,gm[0],gm[1]);
  histogram_time+=(t1-t0);
  modes_gpuTime+=GPU_MARKER_INTERVAL(ge,gm[1],gm[2]);
  modes_time+=(t2-t1);
  if(modeCount>MAX_HISTO_MODE_COUNT)
    {
    printf("MAX_HISTO_MODE_COUNT exceeded (%d)\n",modeCount);
    modeCount=MAX_HISTO_MODE_COUNT;
    }
  for(;;)
    {
    const bool lastAttempt=(modeCount<=1);
    if(lastAttempt) // consider every remaining pulse
      {
      GpuElint_gpuToGpu(ge,toaBitmask,bitmask,0,0,bitmaskCount*sizeof(int));
      trace("analysing remaining pulses\n");
      }
    else // consider only the best mode
      {
      int bestMode=0;
      for(int i=1;i<modeCount;++i)
        { if(modes[i].weight>modes[bestMode].weight) { bestMode=i; } }
      const float histoStepA=(maxValueA-minValueA)/BIN_COUNT_A;
      const float minFilterA=minValueA+(modes[bestMode].FIRST_BIN_A+0.0f)*
                                       histoStepA;
      const float maxFilterA=minValueA+(modes[bestMode].LAST_BIN_A+1.0f)*
                                       histoStepA;
#if defined HISTO_B
      const float histoStepB=(maxValueB-minValueB)/BIN_COUNT_B;
      const float minFilterB=minValueB+(modes[bestMode].firstBinB+0.0f)*
                                       histoStepB;
      const float maxFilterB=minValueB+(modes[bestMode].lastBinB+1.0f)*
                                       histoStepB;
# if defined HISTO_C
      const float histoStepC=(maxValueC-minValueC)/BIN_COUNT_C;
      const float minFilterC=minValueC+(modes[bestMode].firstBinC+0.0f)*
                                       histoStepC;
      const float maxFilterC=minValueC+(modes[bestMode].lastBinC+1.0f)*
                                       histoStepC;
#   if defined HISTO_D
      const float histoStepD=(maxValueD-minValueD)/BIN_COUNT_D;
      const float minFilterD=minValueC+(modes[bestMode].firstBinD+0.0f)*
                                       histoStepD;
      const float maxFilterD=minValueC+(modes[bestMode].lastBinD+1.0f)*
                                       histoStepD;
      const float t0=GET_TIME(ge);
      SET_GPU_MARKER(ge,gm[0]);
      GpuElint_BM_filter4D(ge,bitmask,pulseCount,toaBitmask,
                           valuesA,minFilterA,maxFilterA,
                           valuesB,minFilterB,maxFilterB,
                           valuesC,minFilterC,maxFilterC,
                           valuesD,minFilterD,maxFilterD);
      const float t1=GET_TIME(ge);
      SET_GPU_MARKER(ge,gm[1]);
      trace("analysing %s/%s/%s/%s mode from bin %d/%d/%d/%d to %d/%d/%d/%d\n",
            NAME_A,NAME_B,NAME_C,NAME_D,
            modes[bestMode].firstBinA,modes[bestMode].firstBinB,
            modes[bestMode].firstBinC,modes[bestMode].firstBinD,
            modes[bestMode].lastBinA,modes[bestMode].lastBinB,
            modes[bestMode].lastBinC,modes[bestMode].lastBinD);
#   else
      const float t0=GET_TIME(ge);
      SET_GPU_MARKER(ge,gm[0]);
      GpuElint_BM_filter3D(ge,bitmask,pulseCount,toaBitmask,
                           valuesA,minFilterA,maxFilterA,
                           valuesB,minFilterB,maxFilterB,
                           valuesC,minFilterC,maxFilterC);
      const float t1=GET_TIME(ge);
      SET_GPU_MARKER(ge,gm[1]);
      trace("analysing %s/%s/%s mode from bin %d/%d/%d to %d/%d/%d\n",
            NAME_A,NAME_B,NAME_C,
            modes[bestMode].firstBinA,modes[bestMode].firstBinB,
            modes[bestMode].firstBinC,
            modes[bestMode].lastBinA,modes[bestMode].lastBinB,
            modes[bestMode].lastBinC);
#   endif
# else
      const float t0=GET_TIME(ge);
      SET_GPU_MARKER(ge,gm[0]);
      GpuElint_BM_filter2D(ge,bitmask,pulseCount,toaBitmask,
                           valuesA,minFilterA,maxFilterA,
                           valuesB,minFilterB,maxFilterB);
      const float t1=GET_TIME(ge);
      SET_GPU_MARKER(ge,gm[1]);
      trace("analysing %s/%s mode from bin %d/%d to %d/%d\n",
            NAME_A,NAME_B,
            modes[bestMode].firstBinA,modes[bestMode].firstBinB,
            modes[bestMode].lastBinA,modes[bestMode].lastBinB);
# endif
#else
      const float t0=GET_TIME(ge);
      SET_GPU_MARKER(ge,gm[0]);
      GpuElint_BM_filter1D(ge,bitmask,pulseCount,toaBitmask,
                           valuesA,minFilterA,maxFilterA);
      const float t1=GET_TIME(ge);
      SET_GPU_MARKER(ge,gm[1]);
      trace("analysing %s mode from bin %d to %d\n",
            NAME_A,
            modes[bestMode].firstBin,modes[bestMode].lastBin);
#endif
      filter_gpuTime+=GPU_MARKER_INTERVAL(ge,gm[0],gm[1]);
      filter_time+=(t1-t0);
      modes[bestMode]=modes[--modeCount]; // forget this mode
      }
    const int seqCount=toaAnalysis(ge,toaBitmask,pulseCount,toa64,
                       marks,&seqMark,sequences,minDtoa,dtoaBinCount,
                       dtoaCounts,dtoaSums,minDuration);
    if(seqCount)
      {
      // don't consider marked pulses anymore
      const float t0=GET_TIME(ge);
      SET_GPU_MARKER(ge,gm[0]);
      GpuElint_BM_filterUnmarked(ge,bitmask,pulseCount,
                                 bitmask,marks);
      SET_GPU_MARKER(ge,gm[1]);
      filterUnmarked_gpuTime+=GPU_MARKER_INTERVAL(ge,gm[0],gm[1]);
      filterUnmarked_time+=(GET_TIME(ge)-t0);
      break; // histogram will be rebuilt
      }
    if(lastAttempt) break;
    }
  if(prevSeqMark==seqMark)
    {
    break; // nothing was found with this histogram
    }
  }
GpuElint_gpuToHost(ge,host_marks,marks,0,0,pulseCount*sizeof(int32_t));
for(int i=0;i<pulseCount;++i)
  {
  const int mark=host_marks[i];
  if((mark>=0)&&(mark<MAX_SEQUENCE_COUNT))
    { ++sequences[mark].pulseCount; }
  }
GpuElint_gpuFree(ge,marks);
GpuElint_gpuFree(ge,toaBitmask);
GpuElint_gpuFree(ge,bitmask);
Memory_free(dtoaSums);
Memory_free(dtoaCounts);
GpuElint_gpuFree(ge,histo);
Memory_free(modes);
#if defined HISTO_D
  GpuElint_gpuFree(ge,valuesD);
#endif
#if defined HISTO_C
  GpuElint_gpuFree(ge,valuesC);
#endif
#if defined HISTO_B
  GpuElint_gpuFree(ge,valuesB);
#endif
GpuElint_gpuFree(ge,valuesA);
GpuElint_gpuFree(ge,toa64);
const float endTime=GpuElint_elapsed(ge);
for(int i=0;i<GPU_MARKER_COUNT;++i)
  { GpuElint_destroyMarker(ge,gm[i]); }
#if SHOW_TIME
  printf("             init: %9.3f ms\n",1e3*init_time);
  printf("           minMax: %9.3f ms\n",1e3*minMax_time);
  printf("        histogram: %9.3f ms (%9.3f GPU)\n",
         1e3*histogram_time,
         1e3*histogram_gpuTime);
  printf("            modes: %9.3f ms (%9.3f GPU)\n",
         1e3*modes_time,
         1e3*modes_gpuTime);
  printf("           filter: %9.3f ms (%9.3f GPU)\n",
         1e3*filter_time,
         1e3*filter_gpuTime);
  printf("    dtoaHistogram: %9.3f ms\n",1e3*dtoaHistogram_time);
  printf("      estimatePri: %9.3f ms\n",1e3*estimatePri_time);
  printf("       markPulses: %9.3f ms\n",1e3*markPulses_time);
  printf("   filterUnmarked: %9.3f ms (%9.3f GPU)\n",
         1e3*filterUnmarked_time,
         1e3*filterUnmarked_gpuTime);
#endif
printf("## total duration: %9.3f ms\n",1e3f*(endTime-startTime));
return seqMark;
}

//----------------------------------------------------------------------------
