//----------------------------------------------------------------------------
// Copyright (c) 2015-2016 Fabrice HARROUET (ENIB)
//
// This file is part of GpuElint.
//   http://www.enib.fr/~harrouet/gpuelint.html
//
// GpuElint is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GpuElint is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with GpuElint.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "common.h"

void
trace(const char *format,
      ...)
{
#if SHOW_TRACE
  va_list args;
  va_start(args,format);
  vprintf(format,args);
  va_end(args);
#else
  (void)format; // avoid ``unused parameter'' warning
#endif
}

static
void
loadPulseData(const char *fileName,
              void *storage,
              intptr_t offset,
              intptr_t size)
{
FILE *f=fopen(fileName,"rb");
ON_ERROR(!f,"fopen() failure",{ exit(1); });
intptr_t r=fseek(f,(long)offset,SEEK_SET);
ON_ERROR(r==-1,"fseek() failure",{ exit(1); });
r=fread(storage,size,1,f);
ON_ERROR(r!=1,"fread() failure",{ exit(1); });
fclose(f);
}

int
readPulseCount(const char *fileName)
{
int32_t count;
const intptr_t offset=(0);
loadPulseData(fileName,&count,offset,sizeof(count));
return (int)count;
}

void
loadPulseToa(const char *fileName,
             real64_t *storage,
             int pulseCount)
{
loadPulseData(fileName,storage,
              sizeof(int32_t), // count
              pulseCount*sizeof(real64_t));
}

void
loadPulseProperty(const char *fileName,
                  PropertyKind property,
                  real32_t *storage,
                  int pulseCount)
{
loadPulseData(fileName,storage,
              sizeof(int32_t)+ // count
              pulseCount*sizeof(real64_t)+ // toa
              pulseCount*sizeof(real32_t)*property, // previous properties
              pulseCount*sizeof(real32_t));
}

void
showHistogram1D(const char *title,
                const int *counts,
                const int binCount,
                const float minValue,
                const float maxValue)
{
#if SHOW_TRACE
  printf("%s histogram from %f to %f in %d bins\n",
         title,
         minValue,
         maxValue,
         binCount);
  int maxCount=0;
  for(int i=0;i<binCount;++i)
    {
    const int count=counts[i];
    if(count>maxCount) maxCount=count;
    }
  const float factor=40.0f/maxCount;
  const float delta=(maxValue-minValue)/binCount;
  for(int i=0;i<binCount;++i)
    {
    const int count=counts[i];
    if(count)
      {
      printf("%3d|%11.3f|",i,minValue+(i+0.5f)*delta);
      for(int j=0,j_end=(int)ceilf(count*factor);j<j_end;++j)
        { printf("*"); }
      printf(" %d\n",count);
      }
    }
#else
  (void)title; // avoid ``unused parameter'' warning
  (void)counts;
  (void)binCount;
  (void)minValue;
  (void)maxValue;
#endif
}

void
showDtoaHistogram(const int *counts,
                  const void *sums,
                  bool sumsAre64,
                  const int binCount,
                  const float minDtoa,
                  const float dtoaRatio)
{
#if SHOW_TRACE
  const float *floatSums=(const float *)sums;
  const double *doubleSums=(const double *)sums;
  printf("DTOA histogram from %.3f Hz to %.3f Hz in %d bins\n",
         1.0f/minDtoa,
         1.0f/(minDtoa*powf(dtoaRatio,binCount)),
         binCount);
  int maxCount=0;
  for(int i=0;i<binCount;++i)
    {
    const int count=counts[i];
    if(count>maxCount) maxCount=count;
    }
  const float factor=40.0f/maxCount;
  int totalCount=0;
  float totalSum=0.0f;
  for(int i=0;i<binCount;++i)
    {
    const int count=counts[i];
    const float sum=sumsAre64 ? (float)doubleSums[i] : floatSums[i];
    totalCount+=count;
    totalSum+=sum;
    if(count)
      {
      const float freq=count/sum;
      printf("%3d|%11.3f Hz|",i,freq);
      for(int j=0,j_end=(int)ceilf(count*factor);j<j_end;++j)
        { printf("*"); }
      printf(" %d %.3f\n",count,sum);
      }
    }
  printf("totalCount=%d  totalSum=%g\n",totalCount,totalSum);
#else
  (void)counts; // avoid ``unused parameter'' warning
  (void)sums;
  (void)sumsAre64;
  (void)binCount;
  (void)minDtoa;
  (void)dtoaRatio;
#endif
}

float // pulse-repetition-interval or 0.0f
estimateMajorPri(const int *counts,
                 const void *sums,
                 bool sumsAre64,
                 const int binCount,
                 const float minDtoa,
                 const float dtoaRatio,
                 const float exclusivePriLowerLimit,
                 const float minDtoaSum)
{
const float *floatSums=(const float *)sums;
const double *doubleSums=(const double *)sums;
const float log_minDtoa=logf(minDtoa);
const float div_log_dtoaRatio=1.0f/logf(dtoaRatio);
int bestCount=0;
float bestPri=0.0f;
for(int i=0;i<binCount;++i)
  {
  const float sum=sumsAre64 ? (float)doubleSums[i] : floatSums[i];
  if(sum>=minDtoaSum)
    {
    int count=counts[i];
    float pri=sum/count;
    for(int skip=2;skip<=GPUELINT_DTOA_SKIP_COUNT;++skip)
      {
      const int i2=(int)floorf((logf(skip*pri)-log_minDtoa)*
                               div_log_dtoaRatio);
      if((i2<0)||(i2>=binCount)) { break; }
      const int count2=counts[i2];
      if(count2)
        {
        const float sum2=sumsAre64 ? (float)doubleSums[i2] : floatSums[i2];
        const float pri2=sum2/skip/count2;
        count+=count2;
        pri=(pri2+pri*(skip-1))/skip;
        }
      }
    if((count>bestCount)&&(pri>exclusivePriLowerLimit))
      {
      bestCount=count;
      bestPri=pri;
      }
    }
  }
return bestPri;
}

//----------------------------------------------------------------------------
