//----------------------------------------------------------------------------
// Copyright (c) 2015-2016 Fabrice HARROUET (ENIB)
//
// This file is part of GpuElint.
//   http://www.enib.fr/~harrouet/gpuelint.html
//
// GpuElint is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GpuElint is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with GpuElint.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#ifndef COMMON_H
#define COMMON_H 1

#include "gpuElint.h"

#define SHOW_TIME  1
#define SHOW_TRACE 0

#define USE_ID_SERIAL   1
#define USE_CP_SERIAL   1
#define USE_ID_PARALLEL 1
#define USE_CP_PARALLEL 1
#define USE_ID_GPU      1
#define USE_BM_GPU      1

// (un)comment and change properties to be used in the histogram
//   HISTO_A                            --> use 1D histogram
//   HISTO_A, HISTO_B                   --> use 2D histogram
//   HISTO_A, HISTO_B, HISTO_C          --> use 3D histogram
//   HISTO_A, HISTO_B, HISTO_C, HISTO_D --> use 4D histogram

#define HISTO_A DI_PROP
#define HISTO_B AZIMUT_PROP
// #define HISTO_C FN_PROP
// #define HISTO_D NIV_PROP

#define MAX_SEQUENCE_COUNT       128
#define MAX_HISTO_MODE_COUNT     128
#define HISTO_CUTOFF_RATIO       0.1f
#define HIGHEST_PRF              100000.0f
#define LOWEST_PRF               100.0f
#define DTOA_RATIO               1.025f
#define DTOA_DETECTION_THRESHOLD 0.01f
#define MIN_MATCH_COUNT          4

#define APPROXIMATE_LOGARITHM 1

typedef enum { DI_PROP, FN_PROP, NIV_PROP, AZIMUT_PROP } PropertyKind;

static const struct
{
const char * name;
int binCount;
} properties[]={ { "DI",     128 },
                 { "FN",     32  },
                 { "NIV",    32  },
                 { "AZIMUT", 128 } };

#if 1 // ease conditional compilation of test programs
# define NAME_A (properties[HISTO_A].name)
# define NAME_B (properties[HISTO_B].name)
# define NAME_C (properties[HISTO_C].name)
# define NAME_D (properties[HISTO_D].name)
# define BIN_COUNT_A (properties[HISTO_A].binCount)
# define BIN_COUNT_B (properties[HISTO_B].binCount)
# define BIN_COUNT_C (properties[HISTO_C].binCount)
# define BIN_COUNT_D (properties[HISTO_D].binCount)
# if defined HISTO_D
#   define BIN_COUNT (BIN_COUNT_A*BIN_COUNT_B*BIN_COUNT_C*BIN_COUNT_D)
    typedef HistoMode4D HistoMode;
#   define FIRST_BIN_A firstBinA
#   define LAST_BIN_A lastBinA
# elif defined HISTO_C
#   define BIN_COUNT (BIN_COUNT_A*BIN_COUNT_B*BIN_COUNT_C)
    typedef HistoMode3D HistoMode;
#   define FIRST_BIN_A firstBinA
#   define LAST_BIN_A lastBinA
# elif defined HISTO_B
#   define BIN_COUNT (BIN_COUNT_A*BIN_COUNT_B)
    typedef HistoMode2D HistoMode;
#   define FIRST_BIN_A firstBinA
#   define LAST_BIN_A lastBinA
# else
#   define BIN_COUNT (BIN_COUNT_A)
    typedef HistoMode1D HistoMode;
#   define FIRST_BIN_A firstBin
#   define LAST_BIN_A lastBin
# endif
#endif

#if SHOW_TIME
# define GET_TIME(ge) (GpuElint_elapsed((ge)))
# define GPU_FLUSH(ge,w) GpuElint_flush((ge),(w))
# define SET_GPU_MARKER(ge,m) GpuElint_setMarker((ge),(m))
# define GPU_MARKER_INTERVAL(ge,a,b) (GpuElint_markerInterval((ge),(a),(b)))
#else
# define GET_TIME(g) (0.0f)
# define GPU_FLUSH(g,w)
# define SET_GPU_MARKER(ge,m)
# define GPU_MARKER_INTERVAL(ge,m1,m2) (0.0f)
#endif

typedef struct _Sequence
{
float pri;
int pulseCount;
} Sequence;

PRINTF_FORMAT(1,2)
void
trace(const char *format,
      ...);

void
createDefaultGlContext(void);

void
destroyDefaultGlContext(void);

int
readPulseCount(const char *fileName);

void
loadPulseToa(const char *fileName,
             real64_t *storage,
             int pulseCount);

void
loadPulseProperty(const char *fileName,
                  PropertyKind property,
                  real32_t *storage,
                  int pulseCount);

INLINE
double
approxLog64(double value)
{
// inspired from SIMD-math-prims:
//   https://github.com/jhjourdan/SIMD-math-prims
//   Copyright (c) 2015 Jacques-Henri Jourdan <jourgun@gmail.com>
// FIXME: directly adapted from float to double, but Remez algorithm
//        should be rerun to obtain the polinomial terms with a better
//        accuracy.
union { real64_t r; int64_t i; } u;
u.r=value;
const double exp=(double)(u.i>>52);
const double cst_term=-1.941064434886954;
const double log2_x_1023=709.0895657128241; // 1023*log(2.0)
const double addcst=-(value>0.0 ? (log2_x_1023-cst_term) : INFINITY);
u.i=(u.i&0x000FFFFFFFFFFFFF)|0x3FF0000000000000;
const double x=u.r;
return x*(3.529304993+
          x*(-2.461222105+
             x*(1.130626167+
                x*(-0.288739945+
                   x*3.110401639e-2))))+(addcst+0.69314718055995*exp);
}

void
showHistogram1D(const char *title,
                const int *counts,
                const int binCount,
                const float minValue,
                const float maxValue);

void
showDtoaHistogram(const int *counts,
                  const void *sums,
                  bool sumsAre64,
                  const int binCount,
                  const float minDtoa,
                  const float dtoaRatio);

float // pulse-repetition-interval or 0.0f
estimateMajorPri(const int *counts,
                 const void *sums,
                 bool sumsAre64,
                 const int binCount,
                 const float minDtoa,
                 const float dtoaRatio,
                 const float exclusivePriLowerLimit,
                 const float minDtoaSum);

//----------------------------------------------------------------------------

int // number of extracted sequences
ID_serialAnalysis(GpuElint *ge,
                  const char *fileName,
                  int pulseCount,
                  int *marks,
                  Sequence *sequences);

int // number of extracted sequences
CP_serialAnalysis(GpuElint *ge,
                  const char *fileName,
                  int pulseCount,
                  int *marks,
                  Sequence *sequences);

int // number of extracted sequences
ID_parallelAnalysis(GpuElint *ge,
                    const char *fileName,
                    int pulseCount,
                    int *marks,
                    Sequence *sequences);

int // number of extracted sequences
CP_parallelAnalysis(GpuElint *ge,
                    const char *fileName,
                    int pulseCount,
                    int *marks,
                    Sequence *sequences);

int // number of extracted sequences
ID_gpuAnalysis(GpuElint *ge,
               const char *fileName,
               int pulseCount,
               int *marks,
               Sequence *sequences);

int // number of extracted sequences
BM_gpuAnalysis(GpuElint *ge,
               const char *fileName,
               int pulseCount,
               int *marks,
               Sequence *sequences);

#endif // COMMON_H

//----------------------------------------------------------------------------
