//----------------------------------------------------------------------------
// Copyright (c) 2015-2016 Fabrice HARROUET (ENIB)
//
// This file is part of GpuElint.
//   http://www.enib.fr/~harrouet/gpuelint.html
//
// GpuElint is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GpuElint is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with GpuElint.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "common.h"

static double init_time;
static double minMax_time;
static double histogram_time;
static double modes_time;
static double filter_time;
static double dtoaHistogram_time;
static double estimatePri_time;
static double markPulses_time;
static double filterUnmarked_time;

static int pulseCount[4];
static real64_t *toa64[4];
static real32_t *valuesA[4];
static real32_t *valuesB[4];
static real32_t *valuesC[4];
static real32_t *valuesD[4];
static int *indices[4];

static
void
serial_CP_minMax(int pulseCount,
                 const real32_t * RESTRICT values,
                 float * RESTRICT out_minValue,
                 int * RESTRICT out_minId,
                 float * RESTRICT out_maxValue,
                 int * RESTRICT out_maxId)
{
float minValue=FLT_MAX,maxValue=-FLT_MAX;
int minId=-1,maxId=-1;
TRY_TO_VECTORISE
for(int id=0;id<pulseCount;++id)
  {
  const float value=values[id];
  if(value<minValue)
    { minValue=value; minId=id; }
  if(value>maxValue)
    { maxValue=value; maxId=id; }
  }
*out_minValue=minValue;
*out_minId=minId;
*out_maxValue=maxValue;
*out_maxId=maxId;
}

static
void
serial_CP_histogram1D(int pulseCount,
                      int * RESTRICT counts,
                      const real32_t * RESTRICT values,
                      float minValue,
                      float maxValue,
                      int binCount)
{
memset(counts,0,binCount*sizeof(*counts));
const float factor=binCount/(maxValue-minValue);
const float increment=-minValue*factor;
const int lastBin=binCount-1;
for(int id=0;id<pulseCount;++id)
  {
  const float value=values[id];
  int bin=(int)floorf(value*factor+increment);
  bin=bin<0 ? 0 : bin>lastBin ? lastBin : bin;
  ++counts[bin];
  }
}

static
int // number of modes
serial_modes1D(HistoMode1D * RESTRICT out_modes,
               int modeCapacity,
               const int * RESTRICT counts,
               float cutoffRatio,
               int binCount,
               bool cyclic)
{
int modeCount=0;
for(int bin=0;bin<binCount;++bin)
  {
  int count=counts[bin];
  if(!count) { continue; }
  HistoMode1D m={ count, count, bin, bin };
  const int cutoff=(int)floorf(cutoffRatio*m.maxCount);
  bool giveUp=false;
  int stop=cyclic ? bin : binCount-1;
  int prevCount=m.maxCount;
  bool lowEnough=false;
  for(int i=bin;(i=(i+binCount-1)%binCount)!=stop;)
    {
    count=counts[i];
    if(count>m.maxCount) { giveUp=true; break; }
    if(count<=cutoff) { lowEnough=true; }
    if((count==0)||((count>=prevCount)&&lowEnough))
      { break; }
    m.weight+=count;
    m.firstBin=i;
    prevCount=count;
    }
  if(giveUp) { continue; }
  stop=cyclic ? bin : 0;
  prevCount=m.maxCount;
  lowEnough=false;
  for(int i=bin;(i=(i+1)%binCount)!=stop;)
    {
    count=counts[i];
    if(count>m.maxCount) { giveUp=true; break; }
    if(count<=cutoff) { lowEnough=true; }
    if((count==0)||((count>=prevCount)&&lowEnough))
      { break; }
    m.weight+=count;
    m.lastBin=i;
    prevCount=count;
    }
  if(giveUp) { continue; }
  const int pos=modeCount++;
  if(pos<modeCapacity) { out_modes[pos]=m; }
  }
return modeCount;
}

static
void
serial_CP_filter1D(int toLevel,
                   int fromLevel,
                   float minValue,
                   float maxValue)
{
const bool wrap=minValue>maxValue;
const real64_t * RESTRICT fromToa64=toa64[fromLevel];
real64_t * RESTRICT toToa64=toa64[toLevel];
const int * RESTRICT fromIndices=indices[fromLevel];
int * RESTRICT toIndices=indices[toLevel];
const real32_t * RESTRICT fromA=valuesA[fromLevel];
int remaining=0;
for(int id=0,id_end=pulseCount[fromLevel];id<id_end;++id)
  {
  const float value=fromA[id];
  const bool c1=value>=minValue;
  const bool c2=value<=maxValue;
  const bool keep=wrap ? (c1||c2) : (c1&&c2);
  if(keep)
    {
    const int index=remaining++;
    toToa64[index]=fromToa64[id];
    toIndices[index]=fromIndices[id];
    }
  }
pulseCount[toLevel]=remaining;
}

static
void
serial_CP_histogram2D(int pulseCount,
                      int * RESTRICT counts,
                      const real32_t * RESTRICT valuesA,
                      float minValueA,
                      float maxValueA,
                      int binCountA,
                      const real32_t * RESTRICT valuesB,
                      float minValueB,
                      float maxValueB,
                      int binCountB)
{
const int binCount=binCountA*binCountB;
memset(counts,0,binCount*sizeof(*counts));
const float factorA=binCountA/(maxValueA-minValueA);
const float factorB=binCountB/(maxValueB-minValueB);
const float incrementA=-minValueA*factorA;
const float incrementB=-minValueB*factorB;
const int lastBinA=binCountA-1,lastBinB=binCountB-1;
for(int id=0;id<pulseCount;++id)
  {
  const float valueA=valuesA[id];
  const float valueB=valuesB[id];
  int binA=(int)floorf(valueA*factorA+incrementA);
  int binB=(int)floorf(valueB*factorB+incrementB);
  binA=binA<0 ? 0 : binA>lastBinA ? lastBinA : binA;
  binB=binB<0 ? 0 : binB>lastBinB ? lastBinB : binB;
  const int binId=binB*binCountA+binA;
  ++counts[binId];
  }
}

static
int // number of modes
serial_modes2D(HistoMode2D * RESTRICT out_modes,
               int modeCapacity,
               const int * RESTRICT counts,
               float cutoffRatio,
               int binCountA,
               bool cyclicA,
               int binCountB,
               bool cyclicB)
{
int modeCount=0;
const int binCount=binCountA*binCountB;
for(int bin=0;bin<binCount;++bin)
  {
  const int binA=bin%binCountA;
  const int binB=bin/binCountA;
  int count=counts[bin];
  if(!count) { continue; }
  HistoMode2D m={ count, count, binA, binA, binB, binB };
  const int cutoff=(int)floorf(cutoffRatio*m.maxCount);
  bool giveUp=false;
  int stop=cyclicA ? binA : binCountA-1;
  int prevCount=m.maxCount;
  bool lowEnough=false;
  for(int i=binA;(i=(i+binCountA-1)%binCountA)!=stop;)
    {
    count=counts[binB*binCountA+i];
    if(count>m.maxCount) { giveUp=true; break; }
    if(count<=cutoff) { lowEnough=true; }
    if((count==0)||((count>=prevCount)&&lowEnough))
      { break; }
    m.weight+=count;
    m.firstBinA=i;
    prevCount=count;
    }
  if(giveUp) { continue; }
  stop=cyclicA ? binA : 0;
  prevCount=m.maxCount;
  lowEnough=false;
  for(int i=binA;(i=(i+1)%binCountA)!=stop;)
    {
    count=counts[binB*binCountA+i];
    if(count>m.maxCount) { giveUp=true; break; }
    if(count<=cutoff) { lowEnough=true; }
    if((count==0)||((count>=prevCount)&&lowEnough))
      { break; }
    m.weight+=count;
    m.lastBinA=i;
    prevCount=count;
    }
  if(giveUp) { continue; }
  stop=cyclicB ? binB : binCountB-1;
  prevCount=m.maxCount;
  lowEnough=false;
  for(int i=binB;(i=(i+binCountB-1)%binCountB)!=stop;)
    {
    count=counts[i*binCountA+binA];
    if(count>m.maxCount) { giveUp=true; break; }
    if(count<=cutoff) { lowEnough=true; }
    if((count==0)||((count>=prevCount)&&lowEnough))
      { break; }
    m.weight+=count;
    m.firstBinB=i;
    prevCount=count;
    }
  if(giveUp) { continue; }
  stop=cyclicB ? binB : 0;
  prevCount=m.maxCount;
  lowEnough=false;
  for(int i=binB;(i=(i+1)%binCountB)!=stop;)
    {
    count=counts[i*binCountA+binA];
    if(count>m.maxCount) { giveUp=true; break; }
    if(count<=cutoff) { lowEnough=true; }
    if((count==0)||((count>=prevCount)&&lowEnough))
      { break; }
    m.weight+=count;
    m.lastBinB=i;
    prevCount=count;
    }
  if(giveUp) { continue; }
  const int pos=modeCount++;
  if(pos<modeCapacity) { out_modes[pos]=m; }
  }
return modeCount;
}

static
void
serial_CP_filter2D(int toLevel,
                   int fromLevel,
                   float minValueA,
                   float maxValueA,
                   float minValueB,
                   float maxValueB)
{
const bool wrapA=minValueA>maxValueA;
const bool wrapB=minValueB>maxValueB;
const real64_t * RESTRICT fromToa64=toa64[fromLevel];
real64_t * RESTRICT toToa64=toa64[toLevel];
const int * RESTRICT fromIndices=indices[fromLevel];
int * RESTRICT toIndices=indices[toLevel];
const real32_t * RESTRICT fromA=valuesA[fromLevel];
const real32_t * RESTRICT fromB=valuesB[fromLevel];
int remaining=0;
for(int id=0,id_end=pulseCount[fromLevel];id<id_end;++id)
  {
  const float valueA=fromA[id];
  const float valueB=fromB[id];
  const bool c1A=valueA>=minValueA;
  const bool c2A=valueA<=maxValueA;
  const bool c1B=valueB>=minValueB;
  const bool c2B=valueB<=maxValueB;
  const bool keep=(wrapA ? (c1A||c2A) : (c1A&&c2A))&&
                  (wrapB ? (c1B||c2B) : (c1B&&c2B));
  if(keep)
    {
    const int index=remaining++;
    toToa64[index]=fromToa64[id];
    toIndices[index]=fromIndices[id];
    }
  }
pulseCount[toLevel]=remaining;
}

static
void
serial_CP_histogram3D(int pulseCount,
                      int * RESTRICT counts,
                      const real32_t * RESTRICT valuesA,
                      float minValueA,
                      float maxValueA,
                      int binCountA,
                      const real32_t * RESTRICT valuesB,
                      float minValueB,
                      float maxValueB,
                      int binCountB,
                      const real32_t * RESTRICT valuesC,
                      float minValueC,
                      float maxValueC,
                      int binCountC)
{
const int binCount=binCountA*binCountB*binCountC;
memset(counts,0,binCount*sizeof(*counts));
const float factorA=binCountA/(maxValueA-minValueA);
const float factorB=binCountB/(maxValueB-minValueB);
const float factorC=binCountC/(maxValueC-minValueC);
const float incrementA=-minValueA*factorA;
const float incrementB=-minValueB*factorB;
const float incrementC=-minValueC*factorC;
const int lastBinA=binCountA-1,lastBinB=binCountB-1,lastBinC=binCountC-1;
for(int id=0;id<pulseCount;++id)
  {
  const float valueA=valuesA[id];
  const float valueB=valuesB[id];
  const float valueC=valuesC[id];
  int binA=(int)floorf(valueA*factorA+incrementA);
  int binB=(int)floorf(valueB*factorB+incrementB);
  int binC=(int)floorf(valueC*factorC+incrementC);
  binA=binA<0 ? 0 : binA>lastBinA ? lastBinA : binA;
  binB=binB<0 ? 0 : binB>lastBinB ? lastBinB : binB;
  binC=binC<0 ? 0 : binC>lastBinC ? lastBinC : binC;
  const int binId=(binC*binCountB+binB)*binCountA+binA;
  ++counts[binId];
  }
}

static
int // number of modes
serial_modes3D(HistoMode3D * RESTRICT out_modes,
               int modeCapacity,
               const int * RESTRICT counts,
               float cutoffRatio,
               int binCountA,
               bool cyclicA,
               int binCountB,
               bool cyclicB,
               int binCountC,
               bool cyclicC)
{
int modeCount=0;
const int binCount=binCountA*binCountB*binCountC;
for(int bin=0;bin<binCount;++bin)
  {
  const int binA=bin%binCountA;
  const int binB=(bin/binCountA)%binCountB;
  const int binC=bin/(binCountA*binCountB);
  int count=counts[bin];
  if(!count) { continue; }
  HistoMode3D m={ count, count, binA, binA, binB, binB, binC, binC };
  const int cutoff=(int)floorf(cutoffRatio*m.maxCount);
  bool giveUp=false;
  int stop=cyclicA ? binA : binCountA-1;
  int prevCount=m.maxCount;
  bool lowEnough=false;
  for(int i=binA;(i=(i+binCountA-1)%binCountA)!=stop;)
    {
    count=counts[(binC*binCountB+binB)*binCountA+i];
    if(count>m.maxCount) { giveUp=true; break; }
    if(count<=cutoff) { lowEnough=true; }
    if((count==0)||((count>=prevCount)&&lowEnough))
      { break; }
    m.weight+=count;
    m.firstBinA=i;
    prevCount=count;
    }
  if(giveUp) { continue; }
  stop=cyclicA ? binA : 0;
  prevCount=m.maxCount;
  lowEnough=false;
  for(int i=binA;(i=(i+1)%binCountA)!=stop;)
    {
    count=counts[(binC*binCountB+binB)*binCountA+i];
    if(count>m.maxCount) { giveUp=true; break; }
    if(count<=cutoff) { lowEnough=true; }
    if((count==0)||((count>=prevCount)&&lowEnough))
      { break; }
    m.weight+=count;
    m.lastBinA=i;
    prevCount=count;
    }
  if(giveUp) { continue; }
  stop=cyclicB ? binB : binCountB-1;
  prevCount=m.maxCount;
  lowEnough=false;
  for(int i=binB;(i=(i+binCountB-1)%binCountB)!=stop;)
    {
    count=counts[(binC*binCountB+i)*binCountA+binA];
    if(count>m.maxCount) { giveUp=true; break; }
    if(count<=cutoff) { lowEnough=true; }
    if((count==0)||((count>=prevCount)&&lowEnough))
      { break; }
    m.weight+=count;
    m.firstBinB=i;
    prevCount=count;
    }
  if(giveUp) { continue; }
  stop=cyclicB ? binB : 0;
  prevCount=m.maxCount;
  lowEnough=false;
  for(int i=binB;(i=(i+1)%binCountB)!=stop;)
    {
    count=counts[(binC*binCountB+i)*binCountA+binA];
    if(count>m.maxCount) { giveUp=true; break; }
    if(count<=cutoff) { lowEnough=true; }
    if((count==0)||((count>=prevCount)&&lowEnough))
      { break; }
    m.weight+=count;
    m.lastBinB=i;
    prevCount=count;
    }
  if(giveUp) { continue; }
  stop=cyclicC ? binC : binCountC-1;
  prevCount=m.maxCount;
  lowEnough=false;
  for(int i=binC;(i=(i+binCountC-1)%binCountC)!=stop;)
    {
    count=counts[(i*binCountB+binB)*binCountA+binA];
    if(count>m.maxCount) { giveUp=true; break; }
    if(count<=cutoff) { lowEnough=true; }
    if((count==0)||((count>=prevCount)&&lowEnough))
      { break; }
    m.weight+=count;
    m.firstBinC=i;
    prevCount=count;
    }
  if(giveUp) { continue; }
  stop=cyclicC ? binC : 0;
  prevCount=m.maxCount;
  lowEnough=false;
  for(int i=binC;(i=(i+1)%binCountC)!=stop;)
    {
    count=counts[(i*binCountB+binB)*binCountA+binA];
    if(count>m.maxCount) { giveUp=true; break; }
    if(count<=cutoff) { lowEnough=true; }
    if((count==0)||((count>=prevCount)&&lowEnough))
      { break; }
    m.weight+=count;
    m.lastBinC=i;
    prevCount=count;
    }
  if(giveUp) { continue; }
  const int pos=modeCount++;
  if(pos<modeCapacity) { out_modes[pos]=m; }
  }
return modeCount;
}

static
void
serial_CP_filter3D(int toLevel,
                   int fromLevel,
                   float minValueA,
                   float maxValueA,
                   float minValueB,
                   float maxValueB,
                   float minValueC,
                   float maxValueC)
{
const bool wrapA=minValueA>maxValueA;
const bool wrapB=minValueB>maxValueB;
const bool wrapC=minValueC>maxValueC;
const real64_t * RESTRICT fromToa64=toa64[fromLevel];
real64_t * RESTRICT toToa64=toa64[toLevel];
const int * RESTRICT fromIndices=indices[fromLevel];
int * RESTRICT toIndices=indices[toLevel];
const real32_t * RESTRICT fromA=valuesA[fromLevel];
const real32_t * RESTRICT fromB=valuesB[fromLevel];
const real32_t * RESTRICT fromC=valuesC[fromLevel];
int remaining=0;
for(int id=0,id_end=pulseCount[fromLevel];id<id_end;++id)
  {
  const float valueA=fromA[id];
  const float valueB=fromB[id];
  const float valueC=fromC[id];
  const bool c1A=valueA>=minValueA;
  const bool c2A=valueA<=maxValueA;
  const bool c1B=valueB>=minValueB;
  const bool c2B=valueB<=maxValueB;
  const bool c1C=valueC>=minValueC;
  const bool c2C=valueC<=maxValueC;
  const bool keep=(wrapA ? (c1A||c2A) : (c1A&&c2A))&&
                  (wrapB ? (c1B||c2B) : (c1B&&c2B))&&
                  (wrapC ? (c1C||c2C) : (c1C&&c2C));
  if(keep)
    {
    const int index=remaining++;
    toToa64[index]=fromToa64[id];
    toIndices[index]=fromIndices[id];
    }
  }
pulseCount[toLevel]=remaining;
}

static
void
serial_CP_histogram4D(int pulseCount,
                      int * RESTRICT counts,
                      const real32_t * RESTRICT valuesA,
                      float minValueA,
                      float maxValueA,
                      int binCountA,
                      const real32_t * RESTRICT valuesB,
                      float minValueB,
                      float maxValueB,
                      int binCountB,
                      const real32_t * RESTRICT valuesC,
                      float minValueC,
                      float maxValueC,
                      int binCountC,
                      const real32_t * RESTRICT valuesD,
                      float minValueD,
                      float maxValueD,
                      int binCountD)
{
const int binCount=binCountA*binCountB*binCountC*binCountD;
memset(counts,0,binCount*sizeof(*counts));
const float factorA=binCountA/(maxValueA-minValueA);
const float factorB=binCountB/(maxValueB-minValueB);
const float factorC=binCountC/(maxValueC-minValueC);
const float factorD=binCountD/(maxValueD-minValueD);
const float incrementA=-minValueA*factorA;
const float incrementB=-minValueB*factorB;
const float incrementC=-minValueC*factorC;
const float incrementD=-minValueD*factorD;
const int lastBinA=binCountA-1,lastBinB=binCountB-1,
          lastBinC=binCountC-1,lastBinD=binCountD-1;
for(int id=0;id<pulseCount;++id)
  {
  const float valueA=valuesA[id];
  const float valueB=valuesB[id];
  const float valueC=valuesC[id];
  const float valueD=valuesD[id];
  int binA=(int)floorf(valueA*factorA+incrementA);
  int binB=(int)floorf(valueB*factorB+incrementB);
  int binC=(int)floorf(valueC*factorC+incrementC);
  int binD=(int)floorf(valueD*factorD+incrementD);
  binA=binA<0 ? 0 : binA>lastBinA ? lastBinA : binA;
  binB=binB<0 ? 0 : binB>lastBinB ? lastBinB : binB;
  binC=binC<0 ? 0 : binC>lastBinC ? lastBinC : binC;
  binD=binD<0 ? 0 : binD>lastBinD ? lastBinD : binD;
  const int binId=((binD*binCountC+binC)*binCountB+binB)*binCountA+binA;
  ++counts[binId];
  }
}

static
int // number of modes
serial_modes4D(HistoMode4D * RESTRICT out_modes,
               int modeCapacity,
               const int * RESTRICT counts,
               float cutoffRatio,
               int binCountA,
               bool cyclicA,
               int binCountB,
               bool cyclicB,
               int binCountC,
               bool cyclicC,
               int binCountD,
               bool cyclicD)
{
int modeCount=0;
const int binCount=binCountA*binCountB*binCountC*binCountD;
for(int bin=0;bin<binCount;++bin)
  {
  const int binA=bin%binCountA;
  const int binB=(bin/binCountA)%binCountB;
  const int binC=(bin/(binCountA*binCountB))%binCountC;
  const int binD=bin/(binCountA*binCountB*binCountC);
  int count=counts[bin];
  if(!count) { continue; }
  HistoMode4D m={ count, count,
                  binA, binA, binB, binB, binC, binC, binD, binD };
  const int cutoff=(int)floorf(cutoffRatio*m.maxCount);
  bool giveUp=false;
  int stop=cyclicA ? binA : binCountA-1;
  int prevCount=m.maxCount;
  bool lowEnough=false;
  for(int i=binA;(i=(i+binCountA-1)%binCountA)!=stop;)
    {
    count=counts[((binD*binCountC+binC)*binCountB+binB)*binCountA+i];
    if(count>m.maxCount) { giveUp=true; break; }
    if(count<=cutoff) { lowEnough=true; }
    if((count==0)||((count>=prevCount)&&lowEnough))
      { break; }
    m.weight+=count;
    m.firstBinA=i;
    prevCount=count;
    }
  if(giveUp) { continue; }
  stop=cyclicA ? binA : 0;
  prevCount=m.maxCount;
  lowEnough=false;
  for(int i=binA;(i=(i+1)%binCountA)!=stop;)
    {
    count=counts[((binD*binCountC+binC)*binCountB+binB)*binCountA+i];
    if(count>m.maxCount) { giveUp=true; break; }
    if(count<=cutoff) { lowEnough=true; }
    if((count==0)||((count>=prevCount)&&lowEnough))
      { break; }
    m.weight+=count;
    m.lastBinA=i;
    prevCount=count;
    }
  if(giveUp) { continue; }
  stop=cyclicB ? binB : binCountB-1;
  prevCount=m.maxCount;
  lowEnough=false;
  for(int i=binB;(i=(i+binCountB-1)%binCountB)!=stop;)
    {
    count=counts[((binD*binCountC+binC)*binCountB+i)*binCountA+binA];
    if(count>m.maxCount) { giveUp=true; break; }
    if(count<=cutoff) { lowEnough=true; }
    if((count==0)||((count>=prevCount)&&lowEnough))
      { break; }
    m.weight+=count;
    m.firstBinB=i;
    prevCount=count;
    }
  if(giveUp) { continue; }
  stop=cyclicB ? binB : 0;
  prevCount=m.maxCount;
  lowEnough=false;
  for(int i=binB;(i=(i+1)%binCountB)!=stop;)
    {
    count=counts[((binD*binCountC+binC)*binCountB+i)*binCountA+binA];
    if(count>m.maxCount) { giveUp=true; break; }
    if(count<=cutoff) { lowEnough=true; }
    if((count==0)||((count>=prevCount)&&lowEnough))
      { break; }
    m.weight+=count;
    m.lastBinB=i;
    prevCount=count;
    }
  if(giveUp) { continue; }
  stop=cyclicC ? binC : binCountC-1;
  prevCount=m.maxCount;
  lowEnough=false;
  for(int i=binC;(i=(i+binCountC-1)%binCountC)!=stop;)
    {
    count=counts[((binD*binCountC+i)*binCountB+binB)*binCountA+binA];
    if(count>m.maxCount) { giveUp=true; break; }
    if(count<=cutoff) { lowEnough=true; }
    if((count==0)||((count>=prevCount)&&lowEnough))
      { break; }
    m.weight+=count;
    m.firstBinC=i;
    prevCount=count;
    }
  if(giveUp) { continue; }
  stop=cyclicC ? binC : 0;
  prevCount=m.maxCount;
  lowEnough=false;
  for(int i=binC;(i=(i+1)%binCountC)!=stop;)
    {
    count=counts[((binD*binCountC+i)*binCountB+binB)*binCountA+binA];
    if(count>m.maxCount) { giveUp=true; break; }
    if(count<=cutoff) { lowEnough=true; }
    if((count==0)||((count>=prevCount)&&lowEnough))
      { break; }
    m.weight+=count;
    m.lastBinC=i;
    prevCount=count;
    }
  if(giveUp) { continue; }
  stop=cyclicD ? binD : binCountD-1;
  prevCount=m.maxCount;
  lowEnough=false;
  for(int i=binD;(i=(i+binCountD-1)%binCountD)!=stop;)
    {
    count=counts[((i*binCountC+binC)*binCountB+binB)*binCountA+binA];
    if(count>m.maxCount) { giveUp=true; break; }
    if(count<=cutoff) { lowEnough=true; }
    if((count==0)||((count>=prevCount)&&lowEnough))
      { break; }
    m.weight+=count;
    m.firstBinD=i;
    prevCount=count;
    }
  if(giveUp) { continue; }
  stop=cyclicD ? binD : 0;
  prevCount=m.maxCount;
  lowEnough=false;
  for(int i=binD;(i=(i+1)%binCountD)!=stop;)
    {
    count=counts[((i*binCountC+binC)*binCountB+binB)*binCountA+binA];
    if(count>m.maxCount) { giveUp=true; break; }
    if(count<=cutoff) { lowEnough=true; }
    if((count==0)||((count>=prevCount)&&lowEnough))
      { break; }
    m.weight+=count;
    m.lastBinD=i;
    prevCount=count;
    }
  if(giveUp) { continue; }
  const int pos=modeCount++;
  if(pos<modeCapacity) { out_modes[pos]=m; }
  }
return modeCount;
}

static
void
serial_CP_filter4D(int toLevel,
                   int fromLevel,
                   float minValueA,
                   float maxValueA,
                   float minValueB,
                   float maxValueB,
                   float minValueC,
                   float maxValueC,
                   float minValueD,
                   float maxValueD)
{
const bool wrapA=minValueA>maxValueA;
const bool wrapB=minValueB>maxValueB;
const bool wrapC=minValueC>maxValueC;
const bool wrapD=minValueD>maxValueD;
const real64_t * RESTRICT fromToa64=toa64[fromLevel];
real64_t * RESTRICT toToa64=toa64[toLevel];
const int * RESTRICT fromIndices=indices[fromLevel];
int * RESTRICT toIndices=indices[toLevel];
const real32_t * RESTRICT fromA=valuesA[fromLevel];
const real32_t * RESTRICT fromB=valuesB[fromLevel];
const real32_t * RESTRICT fromC=valuesC[fromLevel];
const real32_t * RESTRICT fromD=valuesD[fromLevel];
int remaining=0;
for(int id=0,id_end=pulseCount[fromLevel];id<id_end;++id)
  {
  const float valueA=fromA[id];
  const float valueB=fromB[id];
  const float valueC=fromC[id];
  const float valueD=fromD[id];
  const bool c1A=valueA>=minValueA;
  const bool c2A=valueA<=maxValueA;
  const bool c1B=valueB>=minValueB;
  const bool c2B=valueB<=maxValueB;
  const bool c1C=valueC>=minValueC;
  const bool c2C=valueC<=maxValueC;
  const bool c1D=valueD>=minValueD;
  const bool c2D=valueD<=maxValueD;
  const bool keep=(wrapA ? (c1A||c2A) : (c1A&&c2A))&&
                  (wrapB ? (c1B||c2B) : (c1B&&c2B))&&
                  (wrapC ? (c1C||c2C) : (c1C&&c2C))&&
                  (wrapD ? (c1D||c2D) : (c1D&&c2D));
  if(keep)
    {
    const int index=remaining++;
    toToa64[index]=fromToa64[id];
    toIndices[index]=fromIndices[id];
    }
  }
pulseCount[toLevel]=remaining;
}

static
void
serial_CP_dtoaHistogram(const int pulseCount,
                        const real64_t * RESTRICT toa64,
                        int * RESTRICT counts,
                        double * RESTRICT sums,
                        const int binCount,
                        const float minDtoa,
                        const float dtoaRatio)
{
memset(counts,0,binCount*sizeof(*counts));
memset(sums,0,binCount*sizeof(*sums));
const double log_minDtoa=log(minDtoa);
const double log_dtoaRatio=log(dtoaRatio);
const double div_log_dtoaRatio=1.0/log_dtoaRatio;
for(int id=GPUELINT_DTOA_SKIP_COUNT+1;id<pulseCount;++id)
  {
  const double toa=toa64[id];
  for(int skip=1;skip<=GPUELINT_DTOA_SKIP_COUNT;++skip)
    {
    const double nextToa=toa64[id-skip];
    const double dtoa=fabs(nextToa-toa);
#if APPROXIMATE_LOGARITHM
    const double log_dtoa=approxLog64(dtoa);
#else
    const double log_dtoa=log(dtoa);
#endif
    const double pos=(log_dtoa-log_minDtoa)*div_log_dtoaRatio;
    const int bin=(int)floor(pos);
    if((bin>=0)&&(bin<binCount))
      {
      ++counts[bin];
#define ACCUM_LOG 0
#if ACCUM_LOG
      sums[bin]+=pos;
#else
      sums[bin]+=dtoa;
#endif
      }
    }
  }
#if ACCUM_LOG
 for(int bin=0;bin<binCount;++bin)
   {
   const int count=counts[bin];
   const double sum=sums[bin];
   sums[bin]=count ? count*exp((log_dtoaRatio*sum/count)+log_minDtoa)
                   : 0.0;
   }
#endif
}

static
float // actual pulse-repetition-interval or 0.0f
serial_CP_markPulses(int pulseCount,
                     const real64_t * RESTRICT toa64,
                     const int * RESTRICT indices,
                     int * RESTRICT marks,
                     int mark,
                     float pri,
                     float priRatioThreshold,
                     int minMatchCount)
{
int matchMem[GPUELINT_MAX_MATCH_COUNT];
int priCount=0;
double priSum=0.0;
float expectedPri=pri;
for(int id=0;id<pulseCount;++id)
  {
  double toa=toa64[id];
  double expectedToa=toa+expectedPri;
  float localPriSum=0.0f;
  int matchCount=0,last=id;
  double lastToa=toa;
  float lastDtoa=FLT_MAX;
  for(int next=id;++next<=pulseCount;)
    {
    bool useLast=true;
    if(next<pulseCount)
      {
      const double nextToa=toa64[next];
      const float nextDtoa=fabsf((float)(expectedToa-nextToa));
      if(nextDtoa<lastDtoa)
        {
        last=next;
        lastToa=nextToa;
        lastDtoa=nextDtoa;
        useLast=false;
        }
      }
    if(useLast)
      {
      const float pri=(float)(lastToa-toa);
      if((expectedPri>pri*priRatioThreshold)||
         (pri>expectedPri*priRatioThreshold))
        { break; }
      localPriSum+=pri;
      matchMem[matchCount]=last;
      if(++matchCount==minMatchCount) { break; }
      next=last;
      lastDtoa=FLT_MAX;
      toa=0.5*(lastToa+expectedToa);
      expectedToa=toa+expectedPri;
      }
    }
  if(matchCount==minMatchCount)
    {
    priCount+=minMatchCount;
    priSum+=localPriSum;
    expectedPri=(float)(priSum/priCount);
    marks[indices[id]]=mark;
    for(int m=0;m<minMatchCount;++m)
      { marks[indices[matchMem[m]]]=mark; }
    }
  }
return priCount ? expectedPri : 0.0f;
}

static
void
serial_CP_filterUnmarked(int toLevel,
                         int fromLevel,
                         bool copyValues,
                         const int * RESTRICT marks)
{
const real64_t * RESTRICT fromToa64=toa64[fromLevel];
real64_t * RESTRICT toToa64=toa64[toLevel];
const int * RESTRICT fromIndices=indices[fromLevel];
int * RESTRICT toIndices=indices[toLevel];
int remaining=0;
if(copyValues)
  {
  const real32_t * RESTRICT fromA=valuesA[fromLevel];
  real32_t * RESTRICT toA=valuesA[toLevel];
#if defined HISTO_B
  const real32_t * RESTRICT fromB=valuesB[fromLevel];
  real32_t * RESTRICT toB=valuesB[toLevel];
#endif
#if defined HISTO_C
  const real32_t * RESTRICT fromC=valuesC[fromLevel];
  real32_t * RESTRICT toC=valuesC[toLevel];
#endif
#if defined HISTO_D
  const real32_t * RESTRICT fromD=valuesD[fromLevel];
  real32_t * RESTRICT toD=valuesD[toLevel];
#endif
  for(int id=0,id_end=pulseCount[fromLevel];id<id_end;++id)
    {
    const int index=fromIndices[id];
    const bool keep=marks[index]<0;
    if(keep)
      {
      const int idx=remaining++;
      toToa64[idx]=fromToa64[id];
      toIndices[idx]=index;
      toA[idx]=fromA[id];
#if defined HISTO_B
      toB[idx]=fromB[id];
#endif
#if defined HISTO_C
      toC[idx]=fromC[id];
#endif
#if defined HISTO_D
      toD[idx]=fromD[id];
#endif
      }
    }
  }
else
  {
  for(int id=0,id_end=pulseCount[fromLevel];id<id_end;++id)
    {
    const int index=fromIndices[id];
    const bool keep=marks[index]<0;
    if(keep)
      {
      const int idx=remaining++;
      toToa64[idx]=fromToa64[id];
      toIndices[idx]=index;
      }
    }
  }
pulseCount[toLevel]=remaining;
}

static
bool // cyclic
detectBounds(GpuElint *ge,
             PropertyKind property,
             int pulseCount,
             const float *values,
             float *out_minValue,
             float *out_maxValue)
{
(void)ge; // avoid ``unused parameter'' warning
if(property==AZIMUT_PROP)
  {
  *out_minValue=0.0f;
  *out_maxValue=360.0f;
  return true;
  }
else
  {
  int minId,maxId;
  const float t0=GET_TIME(ge);
  serial_CP_minMax(pulseCount,values,
                   out_minValue,&minId,out_maxValue,&maxId);
  minMax_time+=(GET_TIME(ge)-t0);
  return false;
  }
}

static
void
swapPulseLevels(int levelA,
                int levelB,
                bool swapValues)
{
int tmpC=pulseCount[levelA];
pulseCount[levelA]=pulseCount[levelB]; 
pulseCount[levelB]=tmpC;
real64_t *tmp64=toa64[levelA];
toa64[levelA]=toa64[levelB]; 
toa64[levelB]=tmp64;
int *tmpI=indices[levelA];
indices[levelA]=indices[levelB]; 
indices[levelB]=tmpI;
if(swapValues)
  {
  real32_t *tmp32=valuesA[levelA];
  valuesA[levelA]=valuesA[levelB]; 
  valuesA[levelB]=tmp32;
#if defined HISTO_B
  tmp32=valuesB[levelA];
  valuesB[levelA]=valuesB[levelB]; 
  valuesB[levelB]=tmp32;
#endif
#if defined HISTO_C
  tmp32=valuesC[levelA];
  valuesC[levelA]=valuesC[levelB]; 
  valuesC[levelB]=tmp32;
#endif
#if defined HISTO_D
  tmp32=valuesD[levelA];
  valuesD[levelA]=valuesD[levelB]; 
  valuesD[levelB]=tmp32;
#endif
  }
}

static
int // new marks detected
toaAnalysis(GpuElint *ge,
            int toaLevel,
            int *marks,
            int *inout_seqMark,
            Sequence *sequences,
            float minDtoa,
            int dtoaBinCount,
            int *dtoaCounts,
            double *dtoaSums,
            float minDuration)
{
(void)ge; // avoid ``unused parameter'' warning
int prevSeqMark=*inout_seqMark;
const float priRatioThreshold=DTOA_RATIO*DTOA_RATIO;
float exclusivePriLowerLimit=0.0f;
bool firstHisto=true;
for(;;)
  {
  if(!exclusivePriLowerLimit)
    {
    const float t0=GET_TIME(ge);
    serial_CP_dtoaHistogram(pulseCount[toaLevel],toa64[toaLevel],
                         dtoaCounts,dtoaSums,dtoaBinCount,
                         minDtoa,DTOA_RATIO);
    dtoaHistogram_time+=(GET_TIME(ge)-t0);
    if(firstHisto)
      {
      firstHisto=false;
      showDtoaHistogram(dtoaCounts,dtoaSums,true,
                        dtoaBinCount,minDtoa,DTOA_RATIO);
      }
    }
  const float t0=GET_TIME(ge);
  const float estimatedPri=estimateMajorPri(
                           dtoaCounts,dtoaSums,true,dtoaBinCount,
                           minDtoa,DTOA_RATIO,exclusivePriLowerLimit,
                           minDuration);
  estimatePri_time+=(GET_TIME(ge)-t0);
  if(!estimatedPri)
    { break; }
  trace("estimated PRI: %g (%.3f Hz)\n",
        estimatedPri,1.0f/estimatedPri);
  const float t1=GET_TIME(ge);
  const float pri=serial_CP_markPulses(pulseCount[toaLevel],toa64[toaLevel],
                                       indices[toaLevel],marks,*inout_seqMark,
                                       estimatedPri,priRatioThreshold,
                                       MIN_MATCH_COUNT);
  const float t2=GET_TIME(ge);
  markPulses_time+=(t2-t1);
  if(pri>0.0f)
    {
    const int tmpLevel=3;
    serial_CP_filterUnmarked(tmpLevel,toaLevel,false,marks);
    const int extractedCount=pulseCount[toaLevel]-pulseCount[tmpLevel];
    toaLevel=2;
    swapPulseLevels(tmpLevel,toaLevel,false);
    filterUnmarked_time+=(GET_TIME(ge)-t2);
    const int newMark=(*inout_seqMark)++;
    trace("[%3d] ---> found %d pulses for %.3f Hz\n",
          newMark,extractedCount,1.0f/pri);
    if(newMark<MAX_SEQUENCE_COUNT)
      {
      sequences[newMark].pri=pri;
      sequences[newMark].pulseCount=extractedCount;
      }
    exclusivePriLowerLimit=0.0f;
    }
  else
    {
    exclusivePriLowerLimit=estimatedPri;
    }
  }
return (*inout_seqMark)-prevSeqMark;
}

int // number of extracted sequences
CP_serialAnalysis(GpuElint *ge,
                  const char *fileName,
                  int initialPulseCount,
                  int *marks,
                  Sequence *sequences)
{
(void)&serial_CP_histogram1D; // avoid ``unused function'' warnings
(void)&serial_modes1D;
(void)&serial_CP_filter1D;
(void)&serial_CP_histogram2D;
(void)&serial_modes2D;
(void)&serial_CP_filter2D;
(void)&serial_CP_histogram3D;
(void)&serial_modes3D;
(void)&serial_CP_filter3D;
(void)&serial_CP_histogram4D;
(void)&serial_modes4D;
(void)&serial_CP_filter4D;
const float startTime=GpuElint_elapsed(ge);
printf("## copy-based serial analysis of %d pulses\n",initialPulseCount);
pulseCount[0]=initialPulseCount;
for(int i=0;i<4;++i)
  { toa64[i]=Memory_alloc(initialPulseCount,real64_t); }
loadPulseToa(fileName,toa64[0],initialPulseCount);
const float minDuration=((float)(toa64[0][initialPulseCount-1]-toa64[0][0]))*
                        DTOA_DETECTION_THRESHOLD;
for(int i=0;i<4;++i)
  { valuesA[i]=Memory_alloc(initialPulseCount,real32_t); }
loadPulseProperty(fileName,HISTO_A,valuesA[0],initialPulseCount);
#if defined HISTO_B
  for(int i=0;i<4;++i)
    { valuesB[i]=Memory_alloc(initialPulseCount,real32_t); }
  loadPulseProperty(fileName,HISTO_B,valuesB[0],initialPulseCount);
#endif
#if defined HISTO_C
  for(int i=0;i<4;++i)
    { valuesC[i]=Memory_alloc(initialPulseCount,real32_t); }
  loadPulseProperty(fileName,HISTO_C,valuesC[0],initialPulseCount);
#endif
#if defined HISTO_D
  for(int i=0;i<4;++i)
    { valuesD[i]=Memory_alloc(initialPulseCount,real32_t); }
  loadPulseProperty(fileName,HISTO_D,valuesD[0],initialPulseCount);
#endif
HistoMode *modes=Memory_alloc(MAX_HISTO_MODE_COUNT,HistoMode);
int *histo=Memory_alloc(BIN_COUNT,int);
const float minDtoa=1.0f/HIGHEST_PRF;
const float maxDtoa=1.0f/LOWEST_PRF;
const int dtoaBinCount=(int)ceilf((logf(maxDtoa)-logf(minDtoa))/
                                  logf(DTOA_RATIO));
int *dtoaCounts=Memory_alloc(dtoaBinCount,int);
double *dtoaSums=Memory_alloc(dtoaBinCount,double);
for(int i=0;i<4;++i)
  { indices[i]=Memory_alloc(initialPulseCount,int); }
int * RESTRICT ind0=indices[0];
TRY_TO_VECTORISE
for(int i=0;i<initialPulseCount;++i)
  { ind0[i]=i; marks[i]=-1; }
init_time+=(GET_TIME(ge)-startTime);
int seqMark=0;
int histoLevel=0;
for(;;)
  {
  trace("---------------------------------------"
        "---------------------------------------\n");
  const int prevSeqMark=seqMark;
  float minValueA,maxValueA;
  const bool cyclicA=detectBounds(ge,HISTO_A,
                                  pulseCount[histoLevel],valuesA[histoLevel],
                                  &minValueA,&maxValueA);
#if defined HISTO_B
  float minValueB,maxValueB;
  const bool cyclicB=detectBounds(ge,HISTO_B,
                                  pulseCount[histoLevel],valuesB[histoLevel],
                                  &minValueB,&maxValueB);
#endif
#if defined HISTO_C
  float minValueC,maxValueC;
  const bool cyclicC=detectBounds(ge,HISTO_C,
                                  pulseCount[histoLevel],valuesC[histoLevel],
                                  &minValueC,&maxValueC);
#endif
#if defined HISTO_D
  float minValueD,maxValueD;
  const bool cyclicD=detectBounds(ge,HISTO_D,
                                  pulseCount[histoLevel],valuesD[histoLevel],
                                  &minValueD,&maxValueD);
#endif
#if !defined HISTO_B
  const float t0=GET_TIME(ge);
  serial_CP_histogram1D(pulseCount[histoLevel],histo,
                        valuesA[histoLevel],minValueA,maxValueA,BIN_COUNT_A);
  const float t1=GET_TIME(ge);
  int modeCount=serial_modes1D(modes,MAX_HISTO_MODE_COUNT,
                               histo,HISTO_CUTOFF_RATIO,
                               BIN_COUNT_A,cyclicA);
  const float t2=GET_TIME(ge);
  showHistogram1D(NAME_A,histo,BIN_COUNT_A,minValueA,maxValueA);
  trace("%d modes found in %s histogram\n",
        modeCount,NAME_A);
#elif !defined HISTO_C
  const float t0=GET_TIME(ge);
  serial_CP_histogram2D(pulseCount[histoLevel],histo,
                        valuesA[histoLevel],minValueA,maxValueA,BIN_COUNT_A,
                        valuesB[histoLevel],minValueB,maxValueB,BIN_COUNT_B);
  const float t1=GET_TIME(ge);
  int modeCount=serial_modes2D(modes,MAX_HISTO_MODE_COUNT,
                               histo,HISTO_CUTOFF_RATIO,
                               BIN_COUNT_A,cyclicA,
                               BIN_COUNT_B,cyclicB);
  const float t2=GET_TIME(ge);
  trace("%d modes found in %s/%s histogram\n",
        modeCount,NAME_A,NAME_B);
#elif !defined HISTO_D
  const float t0=GET_TIME(ge);
  serial_CP_histogram3D(pulseCount[histoLevel],histo,
                        valuesA[histoLevel],minValueA,maxValueA,BIN_COUNT_A,
                        valuesB[histoLevel],minValueB,maxValueB,BIN_COUNT_B,
                        valuesC[histoLevel],minValueC,maxValueC,BIN_COUNT_C);
  const float t1=GET_TIME(ge);
  int modeCount=serial_modes3D(modes,MAX_HISTO_MODE_COUNT,
                               histo,HISTO_CUTOFF_RATIO,
                               BIN_COUNT_A,cyclicA,
                               BIN_COUNT_B,cyclicB,
                               BIN_COUNT_C,cyclicC);
  const float t2=GET_TIME(ge);
  trace("%d modes found in %s/%s/%s histogram\n",
        modeCount,NAME_A,NAME_B,NAME_C);
#else
  const float t0=GET_TIME(ge);
  serial_CP_histogram4D(pulseCount[histoLevel],histo,
                        valuesA[histoLevel],minValueA,maxValueA,BIN_COUNT_A,
                        valuesB[histoLevel],minValueB,maxValueB,BIN_COUNT_B,
                        valuesC[histoLevel],minValueC,maxValueC,BIN_COUNT_C,
                        valuesD[histoLevel],minValueD,maxValueD,BIN_COUNT_D);
  const float t1=GET_TIME(ge);
  int modeCount=serial_modes4D(modes,MAX_HISTO_MODE_COUNT,
                               histo,HISTO_CUTOFF_RATIO,
                               BIN_COUNT_A,cyclicA,
                               BIN_COUNT_B,cyclicB,
                               BIN_COUNT_C,cyclicC,
                               BIN_COUNT_D,cyclicD);
  const float t2=GET_TIME(ge);
  trace("%d modes found in %s/%s/%s/%s histogram\n",
        modeCount,NAME_A,NAME_B,NAME_C,NAME_D);
#endif
  histogram_time+=(t1-t0);
  modes_time+=(t2-t1);
  if(modeCount>MAX_HISTO_MODE_COUNT)
    {
    printf("MAX_HISTO_MODE_COUNT exceeded (%d)\n",modeCount);
    modeCount=MAX_HISTO_MODE_COUNT;
    }
  for(;;)
    {
    const bool lastAttempt=(modeCount<=1);
    int toaLevel;
    if(lastAttempt) // consider every remaining pulse
      {
      toaLevel=histoLevel;
      trace("analysing %d remaining pulses\n",pulseCount[toaLevel]);
      }
    else // consider only the best mode
      {
      toaLevel=2;
      int bestMode=0;
      for(int i=1;i<modeCount;++i)
        { if(modes[i].weight>modes[bestMode].weight) { bestMode=i; } }
      const float histoStepA=(maxValueA-minValueA)/BIN_COUNT_A;
      const float minFilterA=minValueA+(modes[bestMode].FIRST_BIN_A+0.0f)*
                                       histoStepA;
      const float maxFilterA=minValueA+(modes[bestMode].LAST_BIN_A+1.0f)*
                                       histoStepA;
#if defined HISTO_B
      const float histoStepB=(maxValueB-minValueB)/BIN_COUNT_B;
      const float minFilterB=minValueB+(modes[bestMode].firstBinB+0.0f)*
                                       histoStepB;
      const float maxFilterB=minValueB+(modes[bestMode].lastBinB+1.0f)*
                                       histoStepB;
# if defined HISTO_C
      const float histoStepC=(maxValueC-minValueC)/BIN_COUNT_C;
      const float minFilterC=minValueC+(modes[bestMode].firstBinC+0.0f)*
                                       histoStepC;
      const float maxFilterC=minValueC+(modes[bestMode].lastBinC+1.0f)*
                                       histoStepC;
#   if defined HISTO_D
      const float histoStepD=(maxValueD-minValueD)/BIN_COUNT_D;
      const float minFilterD=minValueC+(modes[bestMode].firstBinD+0.0f)*
                                       histoStepD;
      const float maxFilterD=minValueC+(modes[bestMode].lastBinD+1.0f)*
                                       histoStepD;
      const float t0=GET_TIME(ge);
      serial_CP_filter4D(toaLevel,histoLevel,
                         minFilterA,maxFilterA,
                         minFilterB,maxFilterB,
                         minFilterC,maxFilterC,
                         minFilterD,maxFilterD);
      const float t1=GET_TIME(ge);
      trace("analysing %d pulses out of %d"
            " (%s/%s/%s/%s mode from bin %d/%d/%d/%d to %d/%d/%d/%d)\n",
            pulseCount[toaLevel],pulseCount[histoLevel],
            NAME_A,NAME_B,NAME_C,NAME_D,
            modes[bestMode].firstBinA,modes[bestMode].firstBinB,
            modes[bestMode].firstBinC,modes[bestMode].firstBinD,
            modes[bestMode].lastBinA,modes[bestMode].lastBinB,
            modes[bestMode].lastBinC,modes[bestMode].lastBinD);
#   else
      const float t0=GET_TIME(ge);
      serial_CP_filter3D(toaLevel,histoLevel,
                         minFilterA,maxFilterA,
                         minFilterB,maxFilterB,
                         minFilterC,maxFilterC);
      const float t1=GET_TIME(ge);
      trace("analysing %d pulses out of %d"
            " (%s/%s/%s mode from bin %d/%d/%d to %d/%d/%d)\n",
            pulseCount[toaLevel],pulseCount[histoLevel],
            NAME_A,NAME_B,NAME_C,
            modes[bestMode].firstBinA,modes[bestMode].firstBinB,
            modes[bestMode].firstBinC,
            modes[bestMode].lastBinA,modes[bestMode].lastBinB,
            modes[bestMode].lastBinC);
#   endif
# else
      const float t0=GET_TIME(ge);
      serial_CP_filter2D(toaLevel,histoLevel,
                         minFilterA,maxFilterA,
                         minFilterB,maxFilterB);
      const float t1=GET_TIME(ge);
      trace("analysing %d pulses out of %d"
            " (%s/%s mode from bin %d/%d to %d/%d)\n",
            pulseCount[toaLevel],pulseCount[histoLevel],
            NAME_A,NAME_B,
            modes[bestMode].firstBinA,modes[bestMode].firstBinB,
            modes[bestMode].lastBinA,modes[bestMode].lastBinB);
# endif
#else
      const float t0=GET_TIME(ge);
      serial_CP_filter1D(toaLevel,histoLevel,
                         minFilterA,maxFilterA);
      const float t1=GET_TIME(ge);
      trace("analysing %d pulses out of %d"
            " (%s mode from bin %d to %d)\n",
            pulseCount[toaLevel],pulseCount[histoLevel],
            NAME_A,
            modes[bestMode].firstBin,modes[bestMode].lastBin);
#endif
      filter_time+=(t1-t0);
      modes[bestMode]=modes[--modeCount]; // forget this mode
      }
    const int seqCount=toaAnalysis(ge,toaLevel,marks,&seqMark,sequences,
                                   minDtoa,dtoaBinCount,
                                   dtoaCounts,dtoaSums,minDuration);
    if(seqCount)
      {
      // don't consider marked pulses anymore
      const float t0=GET_TIME(ge);
      const int tmpLevel=3;
      serial_CP_filterUnmarked(tmpLevel,histoLevel,true,marks);
      histoLevel=1;
      swapPulseLevels(tmpLevel,histoLevel,true);
      filterUnmarked_time+=(GET_TIME(ge)-t0);
      break; // histogram will be rebuilt
      }
    if(lastAttempt) break;
    }
  if(prevSeqMark==seqMark)
    {
    break; // nothing was found with this histogram
    }
  }
for(int i=0;i<4;++i)
  { Memory_free(indices[i]); }
Memory_free(dtoaSums);
Memory_free(dtoaCounts);
Memory_free(histo);
Memory_free(modes);
#if defined HISTO_D
  for(int i=0;i<4;++i)
    { Memory_free(valuesD[i]); }
#endif
#if defined HISTO_C
  for(int i=0;i<4;++i)
    { Memory_free(valuesC[i]); }
#endif
#if defined HISTO_B
  for(int i=0;i<4;++i)
    { Memory_free(valuesB[i]); }
#endif
for(int i=0;i<4;++i)
  { Memory_free(valuesA[i]); }
for(int i=0;i<4;++i)
  { Memory_free(toa64[i]); }
const float endTime=GpuElint_elapsed(ge);
#if SHOW_TIME
  printf("             init: %9.3f ms\n",1e3*init_time);
  printf("           minMax: %9.3f ms\n",1e3*minMax_time);
  printf("        histogram: %9.3f ms\n",1e3*histogram_time);
  printf("            modes: %9.3f ms\n",1e3*modes_time);
  printf("           filter: %9.3f ms\n",1e3*filter_time);
  printf("    dtoaHistogram: %9.3f ms\n",1e3*dtoaHistogram_time);
  printf("      estimatePri: %9.3f ms\n",1e3*estimatePri_time);
  printf("       markPulses: %9.3f ms\n",1e3*markPulses_time);
  printf("   filterUnmarked: %9.3f ms\n",1e3*filterUnmarked_time);
#endif
printf("## total duration: %9.3f ms\n",1e3f*(endTime-startTime));
return seqMark;
}

//----------------------------------------------------------------------------
