#!/usr/bin/env python
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (c) 2015-2016 Fabrice HARROUET (ENIB)
#
# This file is part of GpuElint.
#   http://www.enib.fr/~harrouet/gpuelint.html
#
# GpuElint is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GpuElint is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with GpuElint.  If not, see <http://www.gnu.org/licenses/>.
#-----------------------------------------------------------------------------

import sys
import traceback
import time
import math
import threading
import ctypes
import struct

sys.path.insert(0,'../GpuElint')
sys.dont_write_bytecode=True
import PythonGpuElint

from OpenGL.GL import *
from OpenGL.GLUT import *

sizeof_double=ctypes.sizeof(ctypes.c_double)
sizeof_float=ctypes.sizeof(ctypes.c_float)
sizeof_int=ctypes.sizeof(ctypes.c_int)

#-----------------------------------------------------------------------------
# Application-specific code
#-----------------------------------------------------------------------------

HISTO_A=0
HISTO_B=3

properties=[ ('DI',128), ('FN',32), ('NIV',32), ('AZIMUT',128) ]
NAME_A=properties[HISTO_A][0]
NAME_B=properties[HISTO_B][0]
BIN_COUNT_A=properties[HISTO_A][1]
BIN_COUNT_B=properties[HISTO_B][1]

MAX_HISTO_MODE_COUNT=128
HISTO_CUTOFF_RATIO=0.1
HIGHEST_PRF=100000.0
LOWEST_PRF=100.0
DTOA_RATIO=1.025
DTOA_DETECTION_THRESHOLD=0.01
MIN_MATCH_COUNT=4

class Application:

  def __init__(self,fileName,width,height):
    #-- properties set by execution environment --
    self.width=width
    self.height=height
    self.fileName=fileName
    #
    #-- simple pass-through vertex-shader source code --
    vertexSrc='''
      #version 430
      in vec2 v_Vertex;
      in int v_Mark;
      flat out int f_Mark;
      uniform vec2 u_Delta;
      void
      main(void)
      {
      f_Mark=v_Mark;
      gl_Position=vec4(v_Vertex+u_Delta,0.0,1.0);
      }
      '''
    #
    #-- simple unicolour fragment-shader source code --
    fragmentSrc='''
      #version 430
      layout(location=0) out vec4 defaultFragColour;
      #define gl_FragColor defaultFragColour
      flat in int f_Mark;
      void
      main(void)
      {
      gl_FragColor=f_Mark<0 ? vec4(0.2,0.8,0.2,1.0)
                            : vec4(0.8,0.2,0.2,1.0);
      }
      '''
    #
    #
    #-- shader-program and parameters --
    self.program=None
    self.v_Vertex=None
    self.v_Mark=None
    self.u_Delta=None
    #
    #-- informations required to draw a simple shape --
    self.vertexCount=None
    self.vertex_vbo=None
    self.mark_vbo=None
    self.vao=None
    self.stepCount=None
    self.redrawCount=None
    self.condVar=None
    #
    #-- GPU execution environment --
    self.ge=None
    #
    try:
      self.ge=PythonGpuElint.GpuElint()
    except:
      sys.stderr.write('~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n')
      sys.stderr.write('~~~~ GpuElint initialisation failure ~~~~\n')
      traceback.print_exc()
      sys.stderr.write('~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n')
      return
    properties=self.ge.properties()
    sys.stdout.write('DTOA_SKIP_COUNT=%d\n'%
                     properties.DTOA_SKIP_COUNT)
    sys.stdout.write('MAX_MATCH_COUNT=%d\n'%
                     properties.MAX_MATCH_COUNT)
    sys.stdout.write('glVersionMajor=%d\n'%
                     properties.glVersionMajor)
    sys.stdout.write('glVersionMinor=%d\n'%
                     properties.glVersionMinor)
    sys.stdout.write('glslVersion=%d\n'%
                     properties.glslVersion)
    sys.stdout.write('maxWorkGroupCountX=%d\n'%
                     properties.maxWorkGroupCountX)
    sys.stdout.write('maxWorkGroupCountY=%d\n'%
                     properties.maxWorkGroupCountY)
    sys.stdout.write('maxWorkGroupCountZ=%d\n'%
                     properties.maxWorkGroupCountZ)
    sys.stdout.write('maxWorkGroupSizeX=%d\n'%
                     properties.maxWorkGroupSizeX)
    sys.stdout.write('maxWorkGroupSizeY=%d\n'%
                     properties.maxWorkGroupSizeY)
    sys.stdout.write('maxWorkGroupSizeZ=%d\n'%
                     properties.maxWorkGroupSizeZ)
    sys.stdout.write('maxWorkGroupInvocations=%d\n'%
                     properties.maxWorkGroupInvocations)
    sys.stdout.write('maxUniformBlocks=%d\n'%
                     properties.maxUniformBlocks)
    sys.stdout.write('maxTextureImageUnits=%d\n'%
                     properties.maxTextureImageUnits)
    sys.stdout.write('maxAtomicCounterBuffers=%d\n'%
                     properties.maxAtomicCounterBuffers)
    sys.stdout.write('maxAtomicCounters=%d\n'%
                     properties.maxAtomicCounters)
    sys.stdout.write('maxSharedMemorySize=%d\n'%
                     properties.maxSharedMemorySize)
    sys.stdout.write('maxUniformComponents=%d\n'%
                     properties.maxUniformComponents)
    sys.stdout.write('maxImageUniforms=%d\n'%
                     properties.maxImageUniforms)
    sys.stdout.write('maxCombinedUniformComponents=%d\n'%
                     properties.maxCombinedUniformComponents)
    sys.stdout.write('maxShaderStorageBlocks=%d\n'%
                     properties.maxShaderStorageBlocks)
    sys.stdout.write('warpSize=%d\n'%
                     properties.warpSize)
    sys.stdout.write('warpsPerStreamingMultiprocessor=%d\n'%
                     properties.warpsPerStreamingMultiprocessor)
    sys.stdout.write('streamingMultiprocessorCount=%d\n'%
                     properties.streamingMultiprocessorCount)
    sys.stdout.write('hasBallot=%d\n'%
                     properties.hasBallot)
    sys.stdout.write('hasShuffle=%d\n'%
                     properties.hasShuffle)
    #
    #-- detect platform-specific features --
    sys.stdout.write('OpenGL version: %s\n'%
                     glGetString(GL_VERSION).decode())
    sys.stdout.write('          GLSL: %s\n'%
                     glGetString(GL_SHADING_LANGUAGE_VERSION).decode())
    sys.stdout.write('        vendor: %s\n'%
                     glGetString(GL_VENDOR).decode())
    sys.stdout.write('      renderer: %s\n'%
                     glGetString(GL_RENDERER).decode())
    #
    #-- load data from file --
    f=open(self.fileName,'rb')
    pulseCount=struct.unpack('@i',f.read(sizeof_int))[0]
    host_toa64=(ctypes.c_double*pulseCount)()
    f.readinto(host_toa64)
    host_valuesA=(ctypes.c_float*pulseCount)()
    f.seek(sizeof_int+pulseCount*(sizeof_double+HISTO_A*sizeof_float))
    f.readinto(host_valuesA)
    host_valuesB=(ctypes.c_float*pulseCount)()
    f.seek(sizeof_int+pulseCount*(sizeof_double+HISTO_B*sizeof_float))
    f.readinto(host_valuesB)
    f.close()
    #
    #-- change initial value for some OpenGL settings --
    glClearColor(0.3,0.3,0.3,1.0)
    #
    #-- create shader-program and access its parameters --
    vs=glCreateShader(GL_VERTEX_SHADER)
    glShaderSource(vs,[vertexSrc])
    glCompileShader(vs)
    fs=glCreateShader(GL_FRAGMENT_SHADER)
    glShaderSource(fs,[fragmentSrc])
    glCompileShader(fs)
    self.program=glCreateProgram()
    glAttachShader(self.program,vs)
    glAttachShader(self.program,fs)
    glLinkProgram(self.program)
    glDeleteShader(vs)
    glDeleteShader(fs)
    self.v_Vertex=glGetAttribLocation(self.program,'v_Vertex')
    self.v_Mark=glGetAttribLocation(self.program,'v_Mark')
    self.u_Delta=glGetUniformLocation(self.program,'u_Delta')
    #
    #-- compute vertices --
    self.vertexCount=pulseCount
    tmpV=(self.vertexCount*2*ctypes.c_float)()
    side=int(math.sqrt(self.vertexCount+1.0))
    factor=1.8/side
    increment=-0.9
    for i in range(self.vertexCount):
      tmpV[2*i+0]=factor*(i%side)+increment;
      tmpV[2*i+1]=factor*(i/side)+increment;
    #
    #-- create and fill buffers with vertices --
    self.vertex_vbo=glGenBuffers(1)
    glBindBuffer(GL_ARRAY_BUFFER,self.vertex_vbo)
    glBufferData(GL_ARRAY_BUFFER,ctypes.sizeof(tmpV),tmpV,GL_STATIC_DRAW)
    self.mark_vbo=self.ge.gpuAlloc(self.vertexCount*sizeof_int)
    self.ge.countFill(self.mark_vbo,0,self.vertexCount,-1,0)
    #
    #-- create and set vertex-array up --
    self.vao=glGenVertexArrays(1)
    glBindVertexArray(self.vao)
    #-- associate vertex-array attributes and buffers --
    glEnableVertexAttribArray(self.v_Vertex)
    glBindBuffer(GL_ARRAY_BUFFER,self.vertex_vbo)
    glVertexAttribPointer(self.v_Vertex,2,GL_FLOAT,GL_FALSE,0,None)
    glEnableVertexAttribArray(self.v_Mark)
    glBindBuffer(GL_ARRAY_BUFFER,self.mark_vbo)
    glVertexAttribPointer(self.v_Mark,1,GL_INT,GL_FALSE,0,None)
    glBindVertexArray(0) # avoid accidental change
    #
    #-- start an asynchronous computation thread --
    glFinish() # ensure graphics initialisation is done
    self.stepCount=0
    self.redrawCount=0
    self.condVar=threading.Condition()
    sys.stdout.write('----> launching an asynchronous computation thread\n')
    th=threading.Thread(target=self.computeThread,
                        args=(pulseCount,host_toa64,
                              host_valuesA,host_valuesB))
    th.setDaemon(True)
    th.start()

  def computeThread(self,pulseCount,host_toa64,host_valuesA,host_valuesB):
    ge=None
    try:
      ge=PythonGpuElint.GpuElint(self.ge)
    except:
      sys.stderr.write('~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n')
      sys.stderr.write('~~~~ shared GpuElint initialisation failure ~~~~\n')
      traceback.print_exc()
      sys.stderr.write('~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n')
    while True:
      for i in range(100):
        time.sleep(0.01)
        self.step()
      if ge:
        self.compute(ge,pulseCount,host_toa64,host_valuesA,host_valuesB)

  def step(self):
    with self.condVar:
      self.stepCount+=1
      self.condVar.notifyAll()
    glutPostRedisplay()

  def toaAnalysis(self,ge,indices,pulseCount,toa64,
                  marks,seqMark,minDtoa,dtoaBinCount,
                  dtoaCounts,dtoaSums,minDuration):
    prevSeqMark=seqMark
    priRatioThreshold=DTOA_RATIO*DTOA_RATIO
    exclusivePriLowerLimit=0.0
    while True:
      if not exclusivePriLowerLimit:
        ge.dtoaHistogram(indices,pulseCount,toa64,
                         dtoaCounts,dtoaSums,dtoaBinCount,
                         minDtoa,DTOA_RATIO)
        self.step()
      estimatedPri=ge.estimateMajorPri(dtoaCounts,dtoaSums,dtoaBinCount,
                                       minDtoa,DTOA_RATIO,
                                       exclusivePriLowerLimit,minDuration)
      self.step()
      if not estimatedPri: break
      sys.stdout.write('estimated PRI: %g (%.3f Hz)\n'%
                       (estimatedPri,1.0/estimatedPri))
      pri=ge.markPulses(indices,pulseCount,toa64,
                        marks,seqMark,estimatedPri,
                        priRatioThreshold,MIN_MATCH_COUNT)
      self.step()
      if pri>0.0:
        remaining=ge.filterUnmarked(indices,pulseCount,indices,marks)
        self.step()
        extractedCount=pulseCount-remaining
        pulseCount=remaining # don't consider marked pulses anymore
        sys.stdout.write('[%3d] ---> found %d pulses for %.3f Hz\n'%
                         (seqMark,extractedCount,1.0/pri))
        seqMark+=1
        exclusivePriLowerLimit=0.0
      else:
        exclusivePriLowerLimit=estimatedPri
    return (seqMark-prevSeqMark,seqMark)

  def compute(self,ge,pulseCount,host_toa64,host_valuesA,host_valuesB):
    sys.stdout.write('#######################################'+
                     '#######################################\n')
    sys.stdout.write('## gpu analysis of %d pulses\n'%pulseCount)
    #
    #-- send data to GPU --
    toa64=ge.gpuAlloc(ctypes.sizeof(host_toa64))
    ge.hostToGpu(toa64,host_toa64,0,0,ctypes.sizeof(host_toa64))
    valuesA=ge.gpuAlloc(ctypes.sizeof(host_valuesA))
    ge.hostToGpu(valuesA,host_valuesA,0,0,ctypes.sizeof(host_valuesA))
    valuesB=ge.gpuAlloc(ctypes.sizeof(host_valuesB))
    ge.hostToGpu(valuesB,host_valuesB,0,0,ctypes.sizeof(host_valuesB))
    self.step()
    #
    #-- prepare ancillary data --
    minDuration=(float(host_toa64[pulseCount-1]-host_toa64[0])*
                 DTOA_DETECTION_THRESHOLD)
    modes=(ge.HistoMode2D*MAX_HISTO_MODE_COUNT)()
    histo=ge.gpuAlloc(BIN_COUNT_A*BIN_COUNT_B*sizeof_int)
    minDtoa=1.0/HIGHEST_PRF
    maxDtoa=1.0/LOWEST_PRF
    dtoaBinCount=int(math.ceil((math.log(maxDtoa)-math.log(minDtoa))/
                               math.log(DTOA_RATIO)))
    dtoaCounts=(ctypes.c_int*dtoaBinCount)()
    dtoaSums=(ctypes.c_float*dtoaBinCount)()
    indices=ge.gpuAlloc(pulseCount*sizeof_int)
    toaIndices=ge.gpuAlloc(pulseCount*sizeof_int)
    ge.countFill(indices,0,pulseCount,0,1)
    marks=self.mark_vbo; # shared with rendering
    ge.countFill(marks,0,pulseCount,-1,0)
    ge.flush(True)
    self.step()
    seqMark=0
    while True:
      sys.stdout.write('---------------------------------------'+
                       '---------------------------------------\n')
      prevSeqMark=seqMark
      if NAME_A=='AZIMUT':
        (minValueA,maxValueA,cyclicA)=(0.0,360.0,True)
      else:
        (minValueA,minId,
         maxValueA,maxId)=ge.minMax(indices,pulseCount,valuesA)
        self.step()
        cyclicA=False
      if NAME_B=='AZIMUT':
        (minValueB,maxValueB,cyclicB)=(0.0,360.0,True)
      else:
        (minValueB,minId,
         maxValueB,maxId)=ge.minMax(indices,pulseCount,valuesB)
        self.step()
        cyclicB=False
      ge.histogram2D(indices,pulseCount,histo,
                     valuesA,minValueA,maxValueA,BIN_COUNT_A,
                     valuesB,minValueB,maxValueB,BIN_COUNT_B)
      self.step()
      modeCount=ge.modes2D(modes,histo,HISTO_CUTOFF_RATIO,
                           BIN_COUNT_A,cyclicA,
                           BIN_COUNT_B,cyclicB)
      self.step()
      sys.stdout.write('%d modes found in %s/%s histogram\n'%
                       (modeCount,NAME_A,NAME_B))
      if modeCount>len(modes):
        sys.stderr.write('MAX_HISTO_MODE_COUNT exceeded (%d)\n'%modeCount)
        modeCount=len(modes)
      self.step()
      while True:
        lastAttempt=(modeCount<=1)
        if lastAttempt: # consider every remaining pulse
          ge.gpuToGpu(toaIndices,indices,0,0,pulseCount*sizeof_int)
          self.step()
          toaPulseCount=pulseCount
          sys.stdout.write('analysing %d remaining pulses\n'%toaPulseCount)
        else: # consider only the best mode
          bestMode=0
          for i in range(1,modeCount):
            if modes[i].weight>modes[bestMode].weight: bestMode=i
          histoStepA=(maxValueA-minValueA)/BIN_COUNT_A
          minFilterA=minValueA+(modes[bestMode].firstBinA+0.0)*histoStepA
          maxFilterA=minValueA+(modes[bestMode].lastBinA+1.0)*histoStepA
          histoStepB=(maxValueB-minValueB)/BIN_COUNT_B
          minFilterB=minValueB+(modes[bestMode].firstBinB+0.0)*histoStepB
          maxFilterB=minValueB+(modes[bestMode].lastBinB+1.0)*histoStepB
          toaPulseCount=ge.filter2D(indices,pulseCount,toaIndices,
                                    valuesA,minFilterA,maxFilterA,
                                    valuesB,minFilterB,maxFilterB)
          self.step()
          sys.stdout.write(('analysing %d pulses out of %d'+
                            ' (%s/%s mode from bin %d/%d to %d/%d)\n')%
                           (toaPulseCount,pulseCount,NAME_A,NAME_B,
                            modes[bestMode].firstBinA,
                            modes[bestMode].firstBinB,
                            modes[bestMode].lastBinA,
                            modes[bestMode].lastBinB))
          modeCount-=1; modes[bestMode]=modes[modeCount] # forget this mode
        (seqCount,seqMark)=self.toaAnalysis(ge,toaIndices,toaPulseCount,
                                            toa64,marks,seqMark,
                                            minDtoa,dtoaBinCount,
                                            dtoaCounts,dtoaSums,
                                            minDuration)
        if seqCount:
          # don't consider marked pulses anymore
          pulseCount=ge.filterUnmarked(indices,pulseCount,indices,marks)
          break # histogram will be rebuilt
        if lastAttempt: break
      if prevSeqMark==seqMark: break # nothing was found with this histogram
    #
    #-- cleanup --
    # marks is self.mark_vbo (shared with rendering)
    ge.gpuFree(toaIndices)
    ge.gpuFree(indices)
    ge.gpuFree(histo)
    ge.gpuFree(valuesB)
    ge.gpuFree(valuesA)
    ge.gpuFree(toa64)

  def redraw(self):
    #-- wait for next computation step change --
    with self.condVar:
      while self.stepCount==self.redrawCount: self.condVar.wait()
      self.redrawCount=self.stepCount
    #
    #-- animate drawing --
    a=math.pi*0.25*time.time()
    dx=0.1*math.cos(a)
    dy=0.1*math.sin(a)
    #
    #-- reset framebuffer --
    glClear(GL_COLOR_BUFFER_BIT)
    glViewport(0,0,self.width,self.height)
    #
    #-- draw vertex-array using shader-program --
    glUseProgram(self.program)
    glUniform2f(self.u_Delta,dx,dy);
    glBindVertexArray(self.vao)
    glDrawArrays(GL_POINTS,0,self.vertexCount)
    glFinish()

  def keyboard(self,key):
    pass

#-----------------------------------------------------------------------------
# Execution environment
#-----------------------------------------------------------------------------

application=None
fileName=None
width=-1
height=-1

def tryInit():
  global application,width,height
  if not application and width>0 and height>0:
    application=Application(fileName,width,height)
    if not application.ge:
      sys.exit(0)
  return application is not None

def display():
  if tryInit():
    application.redraw()
  glutSwapBuffers()

def timer(value):
  glutPostRedisplay()
  glutTimerFunc(40,timer,value)

def reshape(w,h):
  global width,height
  width=w
  height=h
  if tryInit():
    application.width=width
    application.height=height

def keyboard(key,x,y):
  if key==b'\x1b':
    sys.exit(0)
  if tryInit():
    application.keyboard(key)

if __name__=='__main__':
  if len(sys.argv)!=2:
    sys.stderr.write('usage: %s pulse_file.raw\n'%sys.argv[0])
    sys.exit(1)
  fileName=sys.argv[1]
  title='Python/OpenGL GpuElint test application (%s)'%fileName
  glutInit()
  glutInitDisplayMode(GLUT_DOUBLE|GLUT_RGBA|GLUT_3_2_CORE_PROFILE)
  glutCreateWindow(title.encode(sys.stdout.encoding))
  glutReshapeWindow(800,600)
  glutReshapeFunc(reshape)
  glutDisplayFunc(display)
  glutKeyboardFunc(keyboard)
  glutTimerFunc(0,timer,0)
  glutMainLoop()

#-----------------------------------------------------------------------------
